/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.generator;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.ListIterator;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.plaf.basic.BasicFileChooserUI;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureForItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureIfItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureMergeItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureSelectItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureWhileItem;
import fr.greyc.pantheon.ariane.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.IScriptArgument;
import fr.greyc.pantheon.ariane.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.graph.slot.InSlot;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotResult;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.ui.Workspace;

/**
 * The <code>CodeGenerator</code> is the base class for 
 * all the language generators.
 * A <code>CodeGenerator</code> generates a script file
 * that contains code of a specified language
 * for encoding a specified graph.
 * 
 * It implements the <b>Template Method Design Pattern</b>.
 * 
 * @version Régis CLOUARD - Dec 04, 2007
 * @version Régis CLOUARD - Aug 25, 2012 (add system + script items)
 */
public abstract class CodeGenerator {
	/** Correspondence table : index i -> item identifier. */
	private GraphItem _correspondenceTable[];

	/** The list of all available languages. */
	private static final String[][] _langages = { {"Shell","sh"}, {"Python","py"}, {"Perl","pl"} };

	/** The code file */
	private File _chosenFile;
	/** The graph to save. */
	private Graph _graph;

	/**
	 * Creates a new <code>CodeGenerator></code> for the 
	 * specified language and the related file extension.
	 *
	 * @param chosenFile the file.
	 * @param graph the graph to save.
	 */
	protected CodeGenerator( File chosenFile, Graph graph ) {
		_chosenFile = chosenFile;
		_graph = graph;
		_correspondenceTable=null;
	}

	/**
	 * @param graph the graph to save.
	 * @return the selected generator from the file extension.
	 */
	public static CodeGenerator selectCodeGenerator( Graph graph ) {
		CodeGenerator generator=null;
		String[] chosenFile = chooseFilename(graph.getName());
		if (chosenFile!=null) {
			String filename = chosenFile[0];
			String language = chosenFile[1];
			File file = new File(filename);
			if (language.equals(_langages[0][0])) {
				generator = new BashGenerator(file, graph);
			} else if (language.equals(_langages[1][0])) {
				generator = new PythonGenerator(file, graph);
			} else if (language.equals(_langages[2][0])) {
				generator = new PerlGenerator(file, graph);
			}
			if (file.exists() && JOptionPane.showConfirmDialog(Ariane.workspace,
					Ariane.lang.getString("Generator.dialog.message.part1")
					+filename
					+Ariane.lang.getString("Generator.dialog.message.part2"),
					Ariane.lang.getString("Generator.dialog.title"),
					JOptionPane.YES_NO_OPTION)==JOptionPane.NO_OPTION) {
				Ariane.workspace.printErrorMessage(Ariane.lang.getString("Generator.dialog.error1"));
				return null;
			}
		}
		return generator;
	}

	/**
	 * Launches the code generation for the specified graph
	 * and related subgraphs.
	 * Creates the file named: graph name +extension. If this file already
	 * exists, asks for overwriting permission.
	 * 
	 * @return true if the code has been successfully generated.
	 */
	public final boolean generate( ) {
		if (_chosenFile != null) {
			FileOutputCode fp = null;
			try {
				fp = new FileOutputCode(_chosenFile);
				generateCode(_graph, fp, _chosenFile.getName());
				Ariane.workspace.printSystemMessage(Ariane.lang.getString("Generator.success")+" "+_chosenFile);
				return true;
			} catch(Exception e) {
//				e.printStackTrace();
				_chosenFile.delete();
				Ariane.workspace.printErrorMessage(Ariane.lang.getString("Generator.failure")+Ariane.lang.getString("Generator.failure.text"));
			} finally {
				if (fp!=null) {
					fp.close();
				}
			}
		}
		return false;
	}

	/**
	 * Generates the code for the specified graph.
	 * Implements the <b>Template Method Design Pattern</b>.
	 * 
	 * @param graph the graph.
	 * @param fp the file descriptor where to save the code.
	 * @param filename the name of the file (for comments)
	 * @throws Exception in case of malformed code.
	 */
	private void generateCode( Graph graph, FileOutputCode fp, String filename ) throws Exception {
		ArrayList<GraphItem> litems=buildItemsSequence(graph);
		ArrayList<GraphItem> args=getArguments(litems);
		ArrayList<GraphItem> inputs=getLoaderItems(litems);
		ArrayList<GraphItem> outputs=getSaverItems(litems);

		generateScriptHeader(fp, filename, args, inputs, outputs);
		generateSubroutinesDefinition(litems,fp);
		generateUtilitySubroutinesDefinition(fp);
		generateMainPrologue(graph,fp, args, inputs, outputs);
		generateSubroutineBody(litems,1,fp,1);
		generateMainEpilogue(graph,fp);
		generateScriptFooter(fp);
		fp.close();
	}

	/**
	 * Returns the list of all the loaders in the graph.
	 * The list keeps the topological order.
	 * 
	 * @param litems the list of graph items.
	 * @return the list of input images.
	 */
	private ArrayList<GraphItem> getLoaderItems( ArrayList<GraphItem> litems ) {
		ArrayList<GraphItem> items = new ArrayList<GraphItem>();
		// In practice: it is better to return a reverse list. 
		ListIterator<GraphItem> iterator = litems.listIterator(litems.size());
		while (iterator.hasPrevious() ){
			GraphItem item = iterator.previous();
			if (item instanceof LoaderItem) {
				items.add(item);
			}
		}
		return items;
	}

	/**
	 * Returns the list of all the loaders in the graph.
	 * The list keeps the topological order.
	 * 
	 * @param litems the list of graph items.
	 * @return the list of input images.
	 */
	private ArrayList<GraphItem> getArguments( ArrayList<GraphItem> litems ) {
		ArrayList<GraphItem> items = new ArrayList<GraphItem>();
		// In practice: it is better to return a reverse list.
		ListIterator<GraphItem> iterator = litems.listIterator(litems.size());
		while (iterator.hasPrevious()){
			GraphItem item = iterator.previous();
			if (item instanceof IScriptArgument && ((IScriptArgument) item).isArgument()) {
				items.add(item);
			}
		}
		return items;
	}

	/**
	 * Returns the list of all the savers in the graph.
	 * The list keeps the topological order.
	 * 
	 * @param litems the list of graph items.
	 * @return the list of output images.
	 */
	private ArrayList<GraphItem> getSaverItems( ArrayList<GraphItem> litems ) {
		ArrayList<GraphItem> items = new ArrayList<GraphItem>();
		// In practice: it is better to return a reverse list.
		ListIterator<GraphItem> iterator = litems.listIterator(litems.size());
		while (iterator.hasPrevious()){
			GraphItem item = iterator.previous();
			if (item instanceof SaverItem) {
				items.add(item);
			}
		}
		return items;
	}

	/**
	 * Generates the header of the script file and
	 * the code to check the argument line consistency.
	 * 
	 * @param fp the output stream.
	 * @param filename the name of the code file.
	 * @param arguments the list of arguments defined in the graph.
	 * @param inputs the list of loaders defined in the graph.
	 * @param outputs the list of savers defined in the graph.
	 */
	protected abstract void generateScriptHeader( FileOutputCode fp, String filename, ArrayList<GraphItem> arguments, ArrayList<GraphItem> inputs, ArrayList<GraphItem> outputs );

	/**
	 * Generates the footer of the script file.
	 * 
	 * @param fp the output stream.
	 */
	protected abstract void generateScriptFooter( FileOutputCode fp );

	/**
	 * Generates utility subroutines required to execute the script.
	 * 
	 * @param fp the output stream.
	 */
	protected abstract void generateUtilitySubroutinesDefinition( FileOutputCode fp );

	/**
	 * Generates code for the main subroutine.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream.
	 * @param arguments the list of argument.
	 * @param inputs the list of inputs.
	 */
	protected abstract void generateMainPrologue( Graph graph, FileOutputCode fp, ArrayList<GraphItem> arguments, ArrayList<GraphItem> inputs, ArrayList<GraphItem> outputs );

	/**
	 * Generates code to conclude the program code.
	 * At least, erases all temporary files.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream. Generates the prologue and epilogue code.
	 */
	protected abstract void generateMainEpilogue( Graph graph, FileOutputCode fp );

	/**
	 * Opens a dialog box for selecting the output filename.
	 * 
	 * @param filename the presupposed filename.
	 * @return the chosen file name or null when canceled.
	 */
	private static String[] chooseFilename( String filename ) {
		File currentFileName = new File(filename);
		File currentDirName = new File(Workspace.getCurrentDir());
		final JFileChooser fileChooser = new JFileChooser();
		fileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
		fileChooser.setCurrentDirectory(currentDirName);
		fileChooser.setSelectedFile(currentFileName);
		fileChooser.setDialogTitle(Ariane.lang.getString("Generator.dialogchooser.title"));
		fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);

		fileChooser.addPropertyChangeListener(new PropertyChangeListener() {
			@Override
			public void propertyChange(PropertyChangeEvent evt) {
				if (JFileChooser.FILE_FILTER_CHANGED_PROPERTY.equals(evt.getPropertyName())) {
					JFileChooser chooser = (JFileChooser)evt.getSource();
					FileNameExtensionFilter f = ((FileNameExtensionFilter)chooser.getFileFilter());
					if (f != null) {
						String currentName = ((BasicFileChooserUI)fileChooser.getUI()).getFileName();
						if (currentName.lastIndexOf('.')>0) {
							currentName = currentName.substring(0, currentName .lastIndexOf("."));
						}
						String extension = ((FileNameExtensionFilter)fileChooser.getFileFilter()).getExtensions()[0];
						String fullFileName = currentName+"."+extension;
						fileChooser.setSelectedFile(new File(fullFileName));
					}
				}
			}
		});

		// Use file type preference
		FileNameExtensionFilter defaultFilter=null;
		int typePreference = PreferenceChooser.getInt("CodeGenerator.Preference");
		fileChooser.setAcceptAllFileFilterUsed(false);
		for (int i=0; i< _langages.length; i++) {
			FileNameExtensionFilter filter = new FileNameExtensionFilter(_langages[i][0], _langages[i][1]);
			fileChooser.addChoosableFileFilter(filter);
			if (i==typePreference ) { defaultFilter=filter; }
		}
		fileChooser.setFileFilter(defaultFilter);

		if (fileChooser.showSaveDialog(Ariane.workspace) == JFileChooser.APPROVE_OPTION){
			// Update file type preference.
			for (int i=0; i<_langages.length; i++ ) {
				if (_langages[i][0].equals(fileChooser.getFileFilter().getDescription())) {
					PreferenceChooser.putInt("CodeGenerator.Preference",i);
				}
			}

			File theFile = fileChooser.getSelectedFile();
			if (theFile != null) {
				String fullFileName = fileChooser.getSelectedFile().getPath();
				Workspace.setCurrentDir(fileChooser.getSelectedFile().getParent());
				String description = ((FileNameExtensionFilter)fileChooser.getFileFilter()).getDescription();
				return new String[]{ fullFileName, description};
			}
		}
		return null;
	}

	/**
	 * Builds a representation of the graph as a sequence
	 * of items sorted according to the topological order.
	 * 
	 * @param graph the graph to be transformed.
	 * @return the ordered list of items.
	 */
	protected ArrayList<GraphItem> buildItemsSequence( Graph graph ) {
		return topologicalSort(buildCorrespondenceTable(graph));
	}

	/**
	 * Builds a representation of the graph as a representation
	 * that is use for the topological sort.<br>
	 * Builds the correspondence table: item i -> item identifier,
	 * and the list of output slots for each item.
	 * Caution: keep also the order defined of the output rank!
	 * 
	 * @param graph the graph to be transformed.
	 * @return the list of adjacency between items in the graph.
	 */
	private ArrayList<Integer>[] buildCorrespondenceTable( Graph graph ) {
		LinkedList<GraphItem> items = graph.getGraphItems();
		final int n = items.size();

		// The node table.
		@SuppressWarnings("unchecked")
		ArrayList<Integer>[] adjacenceList = new ArrayList[n];
		for (int i = 0; i < n; i++) {
			adjacenceList[i] = new ArrayList<Integer>();
		}

		_correspondenceTable = new GraphItem[n]; // The correspondence table: [num] -> id
		int max = 0; // The maximum id
		for (int i=0; i<n; i++) {
			if ((_correspondenceTable[i] = items.get(i)).getId() > max)
				max = _correspondenceTable[i].getId();
		}

		int tab_anti[] = new int[max+1]; // Inverse correspondence table: id -> num
		for (int i=0; i<n; i++) {
			tab_anti[items.get(i).getId()] = i;
		}

		// For each item -> get the output slots (images + results).
		for (int i = 0; i< n; i++) {
			LinkedList<OutSlotImage> out = items.get(i).getOutSlotImages();
			ArrayList<InSlot> next;
			// Keep this inverse loop because we use add item at the end of the list!
			for (int j = 0; j<out.size(); j++) {
				next = out.get(j).getConnectedSlots();
				for (int k = 0; k < next.size(); k++) {
					GraphItem gi = next.get(k).getGraphItem();
					adjacenceList[i].add(new Integer(tab_anti[gi.getId()]));
				}
			}
		}
		for (int i = 0; i< n; i++) {
			OutSlotResult out = items.get(i).getOutSlotResult();
			if (out == null) continue;
			ArrayList<InSlot> next;
			// Keep this inverse loop because we use add item in the head of the list!
			next = out.getConnectedSlots();
			for (int k = 0; k < next.size(); k++) {
				GraphItem gi = next.get(k).getGraphItem();
				adjacenceList[i].add(new Integer(tab_anti[gi.getId()]));
			}
		}
		return adjacenceList;
	}

	/**
	 * Sorts the items according to inputs and outputs
	 * order since later items use inputs of the former ones.
	 * Uses the the topological sort on the adjacency list
	 * of items.
	 * Note, all virtual operator are removed from the list.
	 * 
	 * @param adjacenceList the list of adjacent items.
	 * @return the ordered list of items.
	 */
	protected ArrayList<GraphItem> topologicalSort( ArrayList<Integer>[] adjacenceList) {
		int n = adjacenceList.length;
		int d[] = new int[n];
		ArrayList<GraphItem> ordered = new ArrayList<GraphItem>();

		// Creates the "inside half-degree table".
		ArrayList<Integer> succ;
		for (int i=0; i<n; i++) {
			succ = adjacenceList[i];
			for (int j=0; j<succ.size(); j++) {
				d[succ.get(j).intValue()]++;
			}
		}

		// Linking index i such as d[i] = 0.
		int som_pile = 0;
		for (int i=0; i<n; i++) {
			if (d[i] == 0) {
				d[i] = som_pile;
				som_pile = i;
			}
		}

		// Build the result list.
		int li, j;
		for (int i=0; i<n; i++) {
			ordered.add(_correspondenceTable[som_pile]);
			li = som_pile;
			som_pile = d[som_pile];
			d[li] = -1;
			succ = adjacenceList[li];
			for (int successor = succ.size()-1; successor>=0; successor--) {
				j = (succ.get(successor)).intValue();
				d[j]--;
				if (d[j] == 0) {
					d[j] = som_pile;
					som_pile = j;
				}
			}
		}
		return ordered;
	}

	/**
	 * Generates the body of the specified graph.
	 * 
	 * @param litems the list of items in the body.
	 * @param num_graph the identifier of the graph.
	 * @param fp the output stream.
	 * @param indent the current code indentation
	 * @throws Exception in case of IO exception or malformed graph.
	 */
	protected void generateSubroutineBody( ArrayList<GraphItem> litems, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		// 'IF' generation is more complex that is seems
		// when there is item that are shared by operators in
		// both IF and ELSE  parts.
		ArrayList<GraphItem> ifItems=new ArrayList<GraphItem>();
		ArrayList<ArrayList<GraphItem>> elseItems=new ArrayList<ArrayList<GraphItem>>();
		ArrayList<GraphItem> generatedItems = new ArrayList<GraphItem>();
		for (GraphItem gi: litems) {
			generatedItems.add(gi);
			if (ifItems.size()>0 && elseItems.size()>0) {
				ArrayList<GraphItem> items = elseItems.get(0);
				if (items.contains(gi)) {
					generateElsePart((ControlStructureIfItem)ifItems.get(0), num_graph, fp, indent-1);
					ifItems.remove(0);
				}
			}
			if (gi instanceof LoaderItem) {
				generateItem((LoaderItem)gi, num_graph, fp, indent);
			} else if (gi instanceof SaverItem) {
				generateItem((SaverItem)gi, num_graph, fp, indent);
			} else if (gi instanceof ViewerItem) {
				generateItem((ViewerItem)gi, num_graph, fp, indent);
			} else if (gi instanceof EvalItem) {
				generateItem((EvalItem)gi, num_graph, fp, indent);
			} else if (gi instanceof FileNameItem) {
				generateItem((FileNameItem)gi, num_graph, fp, indent);
			} else if (gi instanceof FormatItem) {
				generateItem((FormatItem)gi, num_graph, fp, indent);
			} else if (gi instanceof ConstantItem) {
				generateItem((ConstantItem)gi, num_graph, fp, indent);
			} else if (gi instanceof ScriptItem) {
				generateItem((ScriptItem)gi, num_graph, fp, indent);
			} else if (gi instanceof SystemItem) {
				generateItem((SystemItem)gi, num_graph, fp, indent);
			} else if (gi instanceof NoteItem) {
				generateItem((NoteItem)gi, num_graph, fp, indent);
			} else if (gi instanceof OperatorItem) {
				generateItem((OperatorItem)gi, num_graph, fp, indent);
			} else if (gi instanceof ControlStructureSelectItem) {
				generateItem((ControlStructureSelectItem)gi, num_graph, fp, indent);
			} else if (gi instanceof ControlStructureIfItem) {
				ifItems.add(0, gi);
				generateItem((ControlStructureIfItem)gi, num_graph, fp, indent);
				generateThenPart((ControlStructureIfItem)gi, num_graph, fp, indent++);
				elseItems.add(0, searchForElsePartFirstItem(gi, generatedItems));
			} else if (gi instanceof ControlStructureMergeItem) {
				if (ifItems.size()>0) { ifItems.remove(0); }
				if (elseItems.size()>0) { elseItems.remove(0); }
				indent--;
				generateFiPart(null, num_graph, fp, indent);
				generateItem((ControlStructureMergeItem)gi, num_graph, fp, indent);
			} else if (gi instanceof MacroOperatorItem) {
				generateItem((MacroOperatorItem) gi, num_graph, fp, indent);
			} else if (gi instanceof ControlStructureWhileItem) {
				generateItem((ControlStructureWhileItem)gi, num_graph, fp, indent);
			} else if (gi instanceof ControlStructureForItem) {
				generateItem((ControlStructureForItem)gi, num_graph, fp, indent);
			}
		}
		int size= ifItems.size();
		for (int i=0; i<size; i++) {
			if (ifItems.size()>0) { ifItems.remove(0); }
			if (elseItems.size()>0) { elseItems.remove(0); }
			generateFiPart(null, num_graph, fp, --indent);
		}
	}

	private ArrayList <GraphItem> searchForElsePartFirstItem( GraphItem ifItem, ArrayList<GraphItem> generatedItems ) {
		OutSlotImage elseSlot = ifItem.getOutSlotImages().get(1);
		ArrayList<InSlot> connectedSlots = elseSlot.getConnectedSlots();
		ArrayList <GraphItem> graphItems = new ArrayList <GraphItem>();

		// slots connected to the else part.
		for ( InSlot s : connectedSlots ) {
			graphItems.add(s.getGraphItem());
		}

		// add operators that computes the parameters
		// that are not already generated.
		ArrayList <GraphItem> graphItemsBis = new ArrayList <GraphItem>();
		for ( GraphItem item : graphItems ) {
			for ( InSlotParameter parameter : item.getInSlotParameters()) {
				if (parameter.isConnected()) {
					GraphItem connectedItem = parameter.getConnectedSlot().getGraphItem();
					if (!generatedItems.contains(connectedItem)) {
						graphItemsBis.add(connectedItem);
					}
				}
			}
		}

		// add operators that computes the input images
		// that are not already generated.
		for ( GraphItem item : graphItems ) {
			for ( InSlotImage slot : item.getInSlotImages()) {
				GraphItem connectedItem = slot.getConnectedSlot().getGraphItem();
				if (slot.isConnected() && !generatedItems.contains(connectedItem)) {
					graphItemsBis.add(connectedItem);
				}
			}
		}
		graphItems.addAll(graphItemsBis);
		return graphItems;
	}

	/**
	 * Generates the code of sub-functions called in the graph body.
	 * 
	 * @param litems the list of item in the body.
	 * @param fp the output stream.
	 * @throws Exception incomplete graph.
	 */
	protected final void generateSubroutinesDefinition( ArrayList<GraphItem> litems, FileOutputCode fp ) throws Exception {
		for (int i = 0; i < litems.size(); i++) {
			GraphItem gi = litems.get(i);
			if (gi instanceof ControlStructureWhileItem)
				generateSubGraph((ControlStructureWhileItem)gi,fp);
			else if (gi instanceof ControlStructureForItem)
				generateSubGraph((ControlStructureForItem)gi,fp);
			else if (gi instanceof MacroOperatorItem)
				generateSubGraph((MacroOperatorItem)gi,fp);
		}
	}

	protected Object getVirtualOperator( ArrayList<GraphItem> litems ) {
		// Get and discard the virtual operator.
		Object object=null;
		for ( GraphItem gi : litems ) {
			if (gi instanceof VirtualItem) {
				object = gi;
				break;
			}
		}
		return object;
	}

	/**
	 * Generates the code for loading an image.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( LoaderItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for saving an image.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( SaverItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for viewer item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ViewerItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for executing an operator.
	 * 
	 * @param item the operator to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( OperatorItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for an Eval item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( EvalItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for a <code>FilenameItem</code> item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( FileNameItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for a <code>FormatItem</code> item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( FormatItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for an <code>ConstantItem</code>item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ConstantItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for an <code>ScriptItem</code>item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ScriptItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for an <code>SystemtItem</code>item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( SystemItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for the control structure Select.
	 * 
	 * @param item the select to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ControlStructureSelectItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for the 'then' part of a if structure.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateThenPart( ControlStructureIfItem item, int num_graph, FileOutputCode fp, int indent) throws Exception;

	/**
	 * Generates the code for the 'else' part of a if structure.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateElsePart( ControlStructureIfItem item, int num_graph, FileOutputCode fp, int indent) throws Exception;

	/**
	 * Generates the code for the 'end' part of a if structure.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateFiPart( ControlStructureIfItem item, int num_graph, FileOutputCode fp, int indent) throws Exception;

	/**
	 * Generates the code for the control structure If.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ControlStructureIfItem item, int num_graph,FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for the control structure merge.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateItem( ControlStructureMergeItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception;

	/**
	 * Generates the code for a note. Nothing to do.
	 * 
	 * @param item the note to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 */
	protected abstract void generateItem( NoteItem item, int num_graph, FileOutputCode fp, int indent );

	/**
	 * Generates the code for the specified MacroOperator,
	 * for, or while call.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception in case of incomplete graph.
	 */
	protected abstract void generateItem( CompositeItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception ;

	/**
	 * Generates the code for the specified MacroOperator.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception In case of incomplete graph.
	 */
	protected abstract void generateSubGraph( MacroOperatorItem item, FileOutputCode fp ) throws Exception;

	/**
	 * Generates the code for the specified control structure For.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateSubGraph( ControlStructureForItem item, FileOutputCode fp ) throws Exception;

	/**
	 * Generates the code for the specified control structure While.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	protected abstract void generateSubGraph( ControlStructureWhileItem item, FileOutputCode fp ) throws Exception ;
}
