/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.generator;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.graph.item.ConstantItem;
import fr.greyc.pantheon.ariane.graph.item.EvalItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureForItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureIfItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureMergeItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureSelectItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureWhileItem;
import fr.greyc.pantheon.ariane.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.graph.item.FormatItem;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.graph.item.NoteItem;
import fr.greyc.pantheon.ariane.graph.item.OperatorItem;
import fr.greyc.pantheon.ariane.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.graph.item.ScriptItem;
import fr.greyc.pantheon.ariane.graph.item.SystemItem;
import fr.greyc.pantheon.ariane.graph.item.ViewerItem;
import fr.greyc.pantheon.ariane.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.graph.slot.InSlot;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotMask;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;

import java.io.File;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedList;
import java.util.StringTokenizer;

/**
 * The <code>PerlGenerator</code> generates Python code
 * for the specified graph.
 * It operates in three steps:<ol>
 * <li>transforms the graph into a list of operators usable for the topological sort;
 * <li>executes the topological sort;
 * <li>generates the related code.
 * </ol>
 *
 * @version Régis CLOUARD - Dec 28, 2011
 * @version Régis CLOUARD - Aug 25, 2012 (add system + script items)
 */ 
public final class PythonGenerator extends CodeGenerator {
	/** The list of generated output images. */
	private final ArrayList<String> _outputImages = new ArrayList<String>();

	/**
	 * Creates the Python generator.
	 * 
	 * @param chosenFile the name of the file where to store the script.
	 * @param graph the graph to be saved.
	 */
	public PythonGenerator( File chosenFile, Graph graph) {
		super(chosenFile,graph);
	}

	/**
	 * Generates the header code of the Python file and
	 * the code that checks the argument line consistency.
	 * 
	 * @param fp the output stream.
	 * @param filename the name of the code file.
	 * @param arguments the list of arguments defined in the graph.
	 * @param inputs the list of loaders defined in the graph.
	 * @param outputs the list of savers defined in the graph.
	 */
	@Override
	protected void generateScriptHeader( FileOutputCode fp, String filename, ArrayList<GraphItem> arguments, ArrayList<GraphItem> inputs, ArrayList<GraphItem> outputs) {
		// The current date.
		Date creationDate=new Date();
		DateFormat dateFormat=DateFormat.getDateInstance(DateFormat.MEDIUM);

		fp.println("# -*- coding: utf-8 -*-");
		fp.println("#");
		fp.println("# @(#)"+filename+"				"+dateFormat.format(creationDate));
		fp.println("#");
		fp.println("# This file was automatically generated by Ariane.");
		fp.println("# Ariane is part of the Pantheon Project.");
		fp.println("#\n# GREYC - ENSICAEN & University of Caen, France");

		fp.println("import sys, getopt");
		fp.println("import subprocess");
		fp.println("import os");
		fp.println("import os.path");
		fp.println("import shutil");
		fp.println("import glob");
		fp.println("from math import *\n");

		fp.println("verbose=False\n",0);
		fp.println("def printUsage():",0);
		fp.print("print \"Usage: $0 [-hv]",1);
		for (int i=0;i<arguments.size();i++) {
			fp.print(" "+arguments.get(i).getNickName());
		}
		for (int i=0;i<inputs.size();) {
			fp.print(" input-file"+(++i));
		}
		for (int i=0;i<outputs.size();) {
			fp.print(" output-file"+(++i));
		}
		fp.println("\"");
		fp.println("print \"  -h help\";",1);
		fp.println("print \"  -v verbose\";",1);
		fp.print("print \"  example: $0",1);
		for (int i=0;i<arguments.size();i++) {
			fp.print(" "+arguments.get(i).getNickName());
		}
		for (int i=0;i<inputs.size();i++) {
			fp.print(" "+inputs.get(i).getNickName());
		}
		for (int i=0;i<outputs.size();i++) {
			String name = outputs.get(i).getNickName();
			if (name.equals("saver")) name+=outputs.get(i).getId();
			fp.print(" "+name);
		}
		fp.println("\"");

		fp.println("sys.exit(1)\n",1);
	}

	/**
	 * Generates the footer of the script file.
	 * 
	 * @param fp the output stream.
	 */
	@Override
	protected void generateScriptFooter( FileOutputCode fp ) { }

	/**
	 * Generates utility functions required to execute the script.
	 * 
	 * @param fp the output stream.
	 */
	@Override
	protected void generateUtilitySubroutinesDefinition( FileOutputCode fp ) {
		fp.println("def execute(*command):");
		fp.println("''' Executes an operator. '''",1);
		fp.println("if (verbose): print command",1);
		fp.println("print subprocess.Popen(command, stdout=subprocess.PIPE).communicate()[0]",1);
		fp.println("return subprocess.Popen(\""+GraphItem.getCommandStatus()+"\", stdout=subprocess.PIPE).communicate()[0]",1);

		fp.println("\ndef link(src, dest):");
		fp.println("''' Creates a symbolic link between the two files. '''",1);
		fp.println("try:",1);
		fp.println("os.symlink(src,dest)",2);
		fp.println("except OSError:",1);
		fp.println("pass",2);
		fp.println("",1);
}

	/**
	 * Creates the name of the image from the specified slot number,
	 * item and graph identifier.
	 * 
	 * @param slotIndex the identifier of the slot.
	 * @param itemIndex the identifier of the item.
	 * @param num_graph the graph identifier
	 * @return a name that is supposed unique.
	 */
	private String createImageName( int slotIndex, int itemIndex, int num_graph ){
		String name= PreferenceChooser.createTemporaryName(num_graph, itemIndex, slotIndex);
		_outputImages.add(name);
		return name;
	}

	/**
	 * Returns the command line for getting the
	 * input images for the specified item slot of the specified
	 * graph identifier.
	 * 
	 * @param slot the related slot.
	 * @param num_graph the graph identifier
	 * @return a name that is supposed to unique.
	 */
	private String getInputImage( InSlot slot, int num_graph ){
		return PreferenceChooser.createTemporaryName(num_graph, slot.getConnectedSlot().getGraphItem().getId(), slot.getConnectedSlot().getRank());
	}

	/**
	 * Creates a variable name with the specified arguments.
	 * 
	 * @param graph the corresponding graph.
	 * @param item the corresponding item.
	 * @param slot the corresponding slot.
	 * @return the name of the variable.
	 */
	private String variable( int graph, int item, int slot ) {
		return "var"+graph+"_"+item+"_"+slot;
	}

	/**
	 * Returns the parameters line command of the specified item.
	 * 
	 * @param item the related item.
	 * @param slotIndex the slot index.
	 * @param graphId the current graph identifier.
	 * @return a value or a command to get the value.
	 */
	private String getParameter( GraphItem item, int slotIndex, int graphId ) {
		InSlotParameter isp = (InSlotParameter) item.getSlot("InSlotParameter", slotIndex);
		if (isp.getConnectedSlot() == null)
			return "\""+((InSlotParameter)item.getSlot("InSlotParameter", slotIndex)).getValue()+"\"";
		else
			return variable(graphId,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank());
	}

	/**
	 * Generates code for the main.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream.
	 * @param arguments the list of argument.
	 * @param inputs the list of inputs.
	 */
	@Override
	protected void generateMainPrologue( Graph graph, FileOutputCode fp, ArrayList<GraphItem> arguments, ArrayList<GraphItem> inputs, ArrayList<GraphItem> outputs ) {
		fp.println("\n### The main program"); 
		fp.println("def main(args):");

		// Comments
		fp.println("''' main ",1);
		for (GraphItem note : graph.getGraphItems()) {
			if (note instanceof NoteItem) {
				fp.println(parseComment(note.getFormula()),1);
			}
		}
		fp.println("'''",1);

		fp.println("if (len(args) != "+(arguments.size()+inputs.size()+outputs.size())+" ):",1);
		fp.println("printUsage()", 2);

		fp.println("fileList= glob.glob(\""+PreferenceChooser.TEMPORARY_FILE_PREFIX+"*"+ PreferenceChooser.TEMPORARY_FILE_EXT+"\")",1);		
		fp.println("for file in fileList: os.remove(file)",1);
		fp.println("argc = 0",1);
		for ( GraphItem item : arguments ) {
			fp.println(variable(1,item.getId(),0)+" = args[argc]; argc += 1",1);
		}
		for ( GraphItem item : inputs ) {
			fp.println("link(args[argc], \""+createImageName(0,item.getId(),1)+"\"); argc += 1",1);
		}
	}

	/**
	 * Generates code to erase all temporary files.
	 * 
	 * @param graph the graph to encode.
	 * @param fp the output stream.
	 */
	@Override
	protected void generateMainEpilogue( Graph graph, FileOutputCode fp ) {
		fp.println("\nif __name__ == \"__main__\":",0);
		fp.println("try:",1);
		fp.println("opts, args = getopt.getopt(sys.argv[1:], \"hv\")",2);
		fp.println("except getopt.GetoptError:",1);
		fp.println("printUsage()",2);
		fp.println("for opt, arg in opts:",1);
		fp.println("if opt=='-h':",2);
		fp.println("printUsage()",3);
		fp.println("if opt=='-v':",2);
		fp.println("verbose=True",3);
		fp.println("main(args)",1);
//		fp.println("unlink(<"+PreferenceChooser.TEMPORARY_FILE_PREFIX+"*"+PreferenceChooser.TEMPORARY_FILE_EXT+">);",1);
	}

	/**
	 * Generates the code for loading an image.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( LoaderItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception { }

	/**
	 * Generates the code for saving an image.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( SaverItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		if ((((item.getInSlotImages()).get(0)).getConnectedSlot()) != null) {
			fp.println("shutil.copy(\""+getInputImage(item.getInSlotImages().get(0),num_graph)
					+"\", args[argc]); argc+=1",indent);
			fp.println("link(\""
					+ getInputImage(item.getInSlotImages().get(0),num_graph)
					+ "\", \""
					+ createImageName(0,item.getId(),num_graph)
					+ "\")",indent);
		}
	}

	/**
	 * Generates the code for viewer item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ViewerItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String commandShell = item.getViewerCommandName();
		if ((item.getInSlotImages().get(0).getConnectedSlot()) != null) {
			fp.println("subprocess.call([\""+commandShell+"\", \""+getInputImage(item.getInSlotImages().get(0),num_graph)+"\"]);", indent);
			fp.println("link (\""
					+ getInputImage(item.getInSlotImages().get(0),num_graph)
					+ "\", \""
					+ createImageName(0,item.getId(),num_graph)
					+ "\");",indent);
		}
	}

	/**
	 * Generates the code for executing an operator.
	 * 
	 * @param item the operator to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( OperatorItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		LinkedList<String> maskLine = new LinkedList<String>();
		LinkedList<String> parameterLine = new LinkedList<String>();
		LinkedList<String> inputLine = new LinkedList<String>();
		LinkedList<String> outputLine = new LinkedList<String>();

		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName==null) {
			throw new Exception(Ariane.lang.getString("Generator.error2")+item.getName());
		}

		// Get parameters
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			parameterLine.add(getParameter(item,i,num_graph));
		}

		// Get mask
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,num_graph))) {
			maskLine.add(getInputImage(mask,num_graph));
		}

		// Get input images.
		for (InSlotImage im : item.getInSlotImages() ) {
			if (im.getConnectedSlot() == null)
				throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
			inputLine.add("\""+getInputImage(im,num_graph)+"\"");
		}


		// Get output images.
		LinkedList<OutSlotImage> im_out = item.getOutSlotImages();
		for (int i = 0; i < im_out.size(); i++) {
			outputLine.add("\""+createImageName(i,item.getId(),num_graph)+"\"");
		}

		String code=null;
		LinkedList<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		for ( String command : commandLine) {
			if (code == null) {
				code = "\""+command+"\"";
			} else {
				code += ", "+command;
			}
		}

		// Get output result.
		fp.println(variable(num_graph,item.getId(),0)+" = " + "execute("+code+")", indent);
	}

	/**
	 * Generates the code for executing an operating system command.
	 * 
	 * @param item the operator to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( SystemItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		LinkedList<String> maskLine = new LinkedList<String>();
		LinkedList<String> parameterLine = new LinkedList<String>();
		LinkedList<String> inputLine = new LinkedList<String>();
		LinkedList<String> outputLine = new LinkedList<String>();

		String codeName = OperatorItem.getCommandPart(item.getBinary());
		if (codeName==null) {
			throw new Exception(Ariane.lang.getString("Generator.error2")+item.getName());
		}

		// Get parameters
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			parameterLine.add(getParameter(item,i,num_graph));
		}

		// Get mask
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,num_graph))) {
			maskLine.add(getInputImage(mask,num_graph));
		}

		// Get input images.
		for (InSlotImage im : item.getInSlotImages() ) {
			if (im.getConnectedSlot() == null)
				throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
			inputLine.add("\""+getInputImage(im,num_graph)+"\"");
		}


		// Get output images.
		LinkedList<OutSlotImage> im_out = item.getOutSlotImages();
		for (int i = 0; i < im_out.size(); i++) {
			outputLine.add("\""+createImageName(i,item.getId(),num_graph)+"\"");
		}

		LinkedList<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
		String code=null;
		for ( String command : commandLine) {
			if (code==null) {
				code = "\""+command+"\"";
			} else if (command.startsWith("var")) {
				code += ", "+command;
			} else {
				code += ", \""+command+"\"";
			}
		}

		// Get output result.
		fp.println(variable(num_graph,item.getId(),0)+" = " + "execute("+code+")", indent);
	}

	/**
	 * Parses the give formula into Perl code.
	 * 
	 * @param formula the function.
	 * @param parameters the list of parameters.
	 * @return the Perl code.
	 */
	private String parseFormula( String formula, String[] parameters ) {
		// Parse the formula.
		StringTokenizer st = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
		String etape;
		String code = "";
		while (st.hasMoreTokens()) {
			etape = st.nextToken();
			if (etape.equals("^")) code += "**";
			else if (etape.equals("ln")) code+="log";
			else if (etape.equals("atan")) code+="a";
			else if (etape.equals("$")) {
				int n=Integer.parseInt(st.nextToken());
				code+= "float("+parameters[n]+")";
			} else
				code += etape;
		}
		return code;
	}

	/**
	 * Generates the code for an Eval item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( EvalItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String parameters[] = new String[item.getInSlotParameters().size()];

		// Get parameters
		for(int i = 0; i < item.getInSlotParameters().size(); i++) {
			parameters[i]=getParameter(item,i,num_graph);
		}

		fp.println(variable(num_graph,item.getId(),0)+" = str("
				+ parseFormula(item.getFormula(),parameters)
				+ ")",indent);
	}

	@Override
	protected void generateItem( FormatItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String parameters[] = new String[item.getInSlotParameters().size()];
		// Get parameters
		for(int i = 0; i < item.getInSlotParameters().size(); i++) {
			parameters[i]=getParameter(item,i,num_graph);
		}

		String code="";
		StringTokenizer st = new StringTokenizer(item.getFormula(), " $()+'\"", true);
		String etape;
		while (st.hasMoreTokens()) {
			etape = st.nextToken();
			if (etape.equals("+")){
				code+="+";
			} else if (etape.equals("quote")) {
			} else if (etape.equals("\"")) {
			} else if (etape.equals("(")) {
				code+="\"'\"+";
			} else if (etape.equals(")")) {
				code+="+\"'\"";
			} else if (etape.equals("$")) {
				int n=Integer.parseInt(st.nextToken());
				code += parameters[n];
			} else
				code += etape;
		}
		fp.println(variable(num_graph,item.getId(),0)+" = "+code, indent);
	}

	/**
	 * Generates the code for the control structure Select.
	 * 
	 * @param item the select to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureSelectItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception{
		String value=getParameter(item,0,num_graph);

		fp.println("# Select", indent);
		for(int i = 0; i < item.getInSlotImages().size(); i++) {
			fp.println("if (int(" + value +") == " + i +"):", indent);
			fp.println("link(\""
					+ getInputImage((item.getInSlotImages()).get(i),num_graph) +"\", \""
					+ createImageName(0,item.getId(),num_graph)+"\")", indent+1);
			fp.println(variable(num_graph,item.getId(),0)+" = \"" + i +"\"", indent+1);
		}
	}

	/**
	 * Generates the code for the control structure If.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureIfItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception{
		String parameters[] = new String[2];
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			parameters[i]=getParameter(item, i, num_graph);
		}

		// Parse the boolean expression.
		StringTokenizer st = new StringTokenizer(item.getFormula(),"&|$^!0123456789.E><=() +*-/", true);
		String etape;
		String code = "";
		while (st.hasMoreTokens()) {
			etape = st.nextToken();
			if (etape.equals("&")) {
				code += " and ";
			} else if (etape.equals("|")) {
				code += " or ";
			} else if (etape.equals("$")) {
				if (st.nextToken().equals("0")) {
					code += "float("+parameters[0]+")";
				} else {
					code += "float("+parameters[0]+")";
				}
			} else {
				code += etape;
			}
		}

		fp.println("if (" + code + "):", indent);
	}

	/**
	 * Generates the code for the 'then' part.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateThenPart( ControlStructureIfItem item, int num_graph, FileOutputCode fp, int indent) throws Exception {
		fp.println("link(\""
				+getInputImage((item.getInSlotImages()).get(0),num_graph)+"\", \""
				+ createImageName(0,item.getId(),num_graph)+"\")", indent+1);
		fp.println(variable(num_graph,item.getId(),0)+" = 0", indent+1);
	}

	/**
	 * Generates the code for the 'else' part.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateElsePart( ControlStructureIfItem item, int num_graph, FileOutputCode fp, int indent) throws Exception {
		fp.println("else:", indent);
		fp.println("link(\""
				+ getInputImage((item.getInSlotImages()).get(0),num_graph)+"\", \""
				+ createImageName(1,item.getId(),num_graph)+"\")", indent+1);
		fp.println(variable(num_graph,item.getId(),0)+" = 1", indent+1);
	}

	/**
	 * Generates the code for the end of a if control structure.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateFiPart( ControlStructureIfItem item, int num_graph, FileOutputCode fp, int indent) throws Exception {
	}


	/**
	 * Generates the code for the control structure merge.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ControlStructureMergeItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception{
		InSlotImage in0; //, in1;
		in0 = ((item.getInSlotImages()).get(0));

		int id0;//, id1;
		id0 = in0.getConnectedSlot().getGraphItem().getId();
		
		fp.println("# Merge" ,indent);
		fp.println("try: "+variable(num_graph,id0,0)+"; input=0;", indent);
		fp.println("except UnboundLocalError: input=1" ,indent);
		fp.print("if (input==0) and ", indent);
		fp.println("os.path.isfile(\""+getInputImage((item.getInSlotImages()).get(0),num_graph)+"\"):");
		fp.println("link(\""
				+ getInputImage((item.getInSlotImages()).get(0),num_graph)
				+ "\", \""
				+ createImageName(0,item.getId(),num_graph)+"\")", indent+1);
		fp.println(variable(num_graph,item.getId(),0)+" = 0", indent+1);
		fp.println( "else:", indent);
		fp.println("link(\""
				+ getInputImage((item.getInSlotImages()).get(1),num_graph) + "\", \""
				+ createImageName(0,item.getId(),num_graph)+"\")", indent+1);
		fp.println(variable(num_graph,item.getId(),0)+" = 1", indent+1);
	}

	/**
	 * Generates the code for a note. Nothing to do.
	 * 
	 * @param item the note to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 */
	@Override
	protected void generateItem( NoteItem item, int num_graph, FileOutputCode fp, int indent ) { }

	/**
	 * Generates the code for a <code>Filename</code> item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( FileNameItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		if (!item.isArgument()) {
			fp.println(variable(num_graph,item.getId(),0)+" = '"+item.getFormula()+"'", indent);
		}
	}

	/**
	 * Generates the code for an <code>Constant</code>item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ConstantItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		if (!item.isArgument() ) {
			String value = item.getFormula().replaceAll("\"","\\\\\"");
			fp.println(variable(num_graph,item.getId(),0)+" = \""+value+"\"",indent);
		}
	}

	/**
	 * Generates the code for an <code>Script</code>item.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateItem( ScriptItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String value = item.getFormula().replaceAll("\"","\\\\\"").replaceAll("\n", "\\\\\n");
		fp.println(variable(num_graph,item.getId(),0)+" = \""+value+"\"",indent);
	}

	/**
	 * Generates the code for the specified MacroOperator,
	 * FFor, or While call.
	 * 
	 * @param item the item to encode.
	 * @param num_graph the identifier number of the graph.
	 * @param fp the output stream.
	 * @param indent the indentation level.
	 * @throws Exception in case of incomplete graph.
	 */
	@Override
	protected void generateItem( CompositeItem item, int num_graph, FileOutputCode fp, int indent ) throws Exception {
		String code = "macro"+item.getId()+"_"+item.getGraph().getIdentifier()+"(";
		boolean comma = false;

		// Get parameters
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			if (!comma)
				comma = true;
			else
				code +=",";
			code +=getParameter(item,i,num_graph);
		}

		// Get mask
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,num_graph))) {
			if (!comma)
				comma = true;
			else
				code +=",";
			code += "\"" + getInputImage(mask,num_graph)+"\"";
		}

		// Get input images.
		for( InSlotImage im : item.getInSlotImages() ) {
			if (im.getConnectedSlot() == null)
				throw new Exception(Ariane.lang.getString("Generator.error.incomplete"));
			if (!comma)
				comma = true;
			else
				code +=",";
			code += "\"" + getInputImage(im,num_graph)+"\"";
		}

		// Get output images.
		Collection<OutSlotImage> im_out = item.getOutSlotImages();
		for(int i = 0; i < im_out.size(); i++) {
			if (!comma)
				comma = true;
			else
				code +=",";
			code += "\"" +createImageName(i,item.getId(),num_graph)+"\"";
		}
		// Get output result.
		fp.println(variable(num_graph,item.getId(),0)+" = " + code+")", indent);
	}

	/**
	 * Parses the comment in order to add a sharp after each newlines.
	 * 
	 * @param comment a string that constrains the comment.
	 * @return a new string.
	 */
	private String parseComment( String comment ) {
		return comment.replaceAll("\n", "\n    ");
	}

	/**
	 * Generates the code for the specified control structure For.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateSubGraph( ControlStructureForItem item, FileOutputCode fp ) throws Exception {
		int rank;
		SubGraph sb = item.getSubGraph();
		int num_subgraph=sb.getIdentifier();
		ArrayList<GraphItem> litems = buildItemsSequence(sb);

		// Get and discard the virtual operator.
		Object virtualOperator=getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.size() == 0) {
			throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
		}

		fp.println("def macro"+item.getId()+"_"+item.getGraph().getIdentifier()+"(*args):");
		// Comments
		fp.println("''' "+item.getNickName()+"\n",1);
		for (GraphItem note : sb.getGraphItems()) {
			if (note instanceof NoteItem) {
				fp.println(note.getFormula(),1);
			}
		}
		fp.println("'''",1);
		fp.println("argc = 0",1);

		// Get parameters
		fp.println("# Input parameters",1);
		fp.println("begin = int(args[argc]); argc += 1",1);
		fp.println("end = int(args[argc]); argc += 1",1);
		fp.println("step = int(args[argc]); argc += 1",1);
		for(int i = 3; i < item.getInSlotParameters().size(); i++) {
			fp.println(variable(num_subgraph,0,i-2)+" = args[argc]; argc += 1",1);
		}
		fp.println(variable(num_subgraph,0,0)+" = str(begin)",1);

		// Get mask.
		fp.println("# Input data files",1);
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,item.getGraph().getIdentifier()))) {
			int rz = sb.getVirtualOperator().getVirtualInSlotMask().getRank();
			fp.println("link(args[argc], \"" + createImageName(rz,0,num_subgraph)+"\"); argc += 1",1);
		}

		// Get input files
		rank = 0;
		for (InSlotImage isi : item.getInSlotImages() ) {
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new Exception("Generator.error.incomplete");
			}
			fp.println("shutil.copy(args[argc], \""+createImageName(rank++,0,num_subgraph)+"\"); argc += 1",1);
		}

		// Create the output images in case of no iteration
		VirtualItem vo = (VirtualItem)virtualOperator;
		for (int i=0; i<item.getOutSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualOutSlotImages().get(i);
			fp.println("f = open(\""+getInputImage(isi,num_subgraph)+"\",\"w\"); f.close()",1);
		}

		// Initialize the return variable in case of no iteration
		{
			InSlotParameter isp = (InSlotParameter)vo.getVirtualOutSlotResult();
			fp.println(variable(num_subgraph,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank())+"=\"SUCCESS\"",1);
		}

		fp.println();
		fp.println("for i in range(begin, end+1, step):",1);
		fp.println(variable(num_subgraph,0,0)+" = str(i)",2);

		generateSubroutineBody(litems,num_subgraph,fp,2);

		// Get intermediate images
		for (int i=0; i<item.getInSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualInterSlotImage(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new Exception("Generator.error.incomplete");
			}
			String input = getInputImage(isi,num_subgraph);
			String output = createImageName(i,0,num_subgraph);
			if (!input.equals(output)) {
				fp.println("shutil.copy(\""+input+"\",\""+output+"\");",2);
			}
		}

		// Get output images.
		fp.println();
		fp.println("# Output images",1);
		for (int i=0; i<item.getOutSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualOutSlotImages().get(i);
			fp.println("link(\""+getInputImage(isi,num_subgraph)+"\",args[argc]); argc += 1",1);
		}

		// Output result
		InSlotParameter isp = (InSlotParameter)vo.getVirtualOutSlotResult();
		if (isp.getConnectedSlot() == null) {
			fp.println("return 0;",1);
		} else {
			fp.println("return "+variable(num_subgraph,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank()),1);
		}
		fp.println("");

		generateSubroutinesDefinition(litems,fp);
	}

	/**
	 * Generates the code for the specified control structure While.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception Incomplete graph.
	 */
	@Override
	protected void generateSubGraph( ControlStructureWhileItem item, FileOutputCode fp ) throws Exception {
		int rank;
		SubGraph sb = item.getSubGraph();
		int num_subgraph=sb.getIdentifier();
		ArrayList<GraphItem> litems = buildItemsSequence(sb);

		// Get and discard the virtual operator.
		Object virtualOperator=getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.size() == 0) {
			throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
		}

		fp.println("def macro"+item.getId()+"_"+item.getGraph().getIdentifier()+"(*args):");
		// Comments
		fp.println("''' "+item.getNickName()+"\n",1);
		for (GraphItem note : sb.getGraphItems()) {
			if (note instanceof NoteItem) {
				fp.println(note.getFormula(),1);
			}
		}
		fp.println("'''",1);
		fp.println("argc = 0",1);

		// Get parameters
		String[] parameters = new String[item.getInSlotParameters().size()];
		fp.println("# Input parameters",1);
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			fp.println(variable(num_subgraph,0,i)+" = args[argc]; argc += 1",1);
			parameters[i]=variable(num_subgraph,0,i);
		}

		// Get mask.
		fp.println("# Input data files",1);
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,item.getGraph().getIdentifier()))) {
			int rz = sb.getVirtualOperator().getVirtualInSlotMask().getRank();
			fp.println("link(args[argc], \"" + createImageName(rz,0,num_subgraph)+"\"); argc += 1",1);
		}

		// Get input files
		rank = 0;
		for (InSlotImage isi : item.getInSlotImages() ) {
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new Exception("Generator.error.incomplete");
			}
			fp.println("shutil.copy(args[argc], \""+createImageName(rank++,0,num_subgraph)+"\"); argc += 1",1);
		}

		// Parse the boolean expression.
		StringTokenizer st = new StringTokenizer(item.getFormula(),"&|$^!0123456789.E><=() +*-/", true);
		String etape;
		String testFormula = "";
		while (st.hasMoreTokens()) {
			etape = st.nextToken();
			if (etape.equals("&")) testFormula += " and ";
			else if (etape.equals("|")) testFormula += " or ";
			else if (etape.equals("$")) {
				if (st.nextToken().equals("0")) {
					testFormula += "float("+parameters[0]+")";
				} else {
					testFormula += "float("+parameters[1]+")";
				}
			} else {
				testFormula += etape;
			}
		}

		VirtualItem vo = (VirtualItem)virtualOperator;
		fp.println();
		fp.println("while ("+testFormula+"):",1);

		generateSubroutineBody(litems,num_subgraph,fp,2);

		// Get intermediate images
		for (int i=0; i<item.getInSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualInterSlotImage(i);
			// Unlinked input!
			if (isi.getConnectedSlot() == null)
				throw new Exception("Generator.error.incomplete");
			String input = getInputImage(isi,num_subgraph);
			String output = createImageName(i,0,num_subgraph);
			if (!input.equals(output)) {
				fp.println("shutil.copy(\""+input+"\",\""+output+"\");",2);
			}
		}

		// Get intermediate parameters
		for (int i=0;i<2;i++) {
			InSlotParameter isp = vo.getVirtualInterSlotResult(i);
			if (isp.getConnectedSlot() != null) {
				fp.println(parameters[i]+" = "+variable(num_subgraph,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank()),2);
			}
		}

		// Get output images.
		fp.println();
		fp.println("# Output images",1);
		for (int i=0; i<item.getOutSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualOutSlotImages().get(i);
			fp.println("link(\""+getInputImage(isi,num_subgraph)+"\",args[argc]); argc += 1",1);
		}

		// Output result
		InSlotParameter isp = (InSlotParameter)vo.getVirtualOutSlotResult();
		if (isp.getConnectedSlot() == null) {
			fp.println("return 0;",1);
		} else {
			fp.println("return "+variable(num_subgraph,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank()),1);
		}
		fp.println("");

		generateSubroutinesDefinition(litems,fp);
	}

	/**
	 * Generates the code for the specified MacroOperator body.
	 * 
	 * @param item the item to encode.
	 * @param fp the output stream.
	 * @throws Exception In case of incomplete graph.
	 */
	@Override
	protected void generateSubGraph( MacroOperatorItem item, FileOutputCode fp ) throws Exception {
		int rank;
		SubGraph sb = item.getSubGraph();
		int num_subgraph=sb.getIdentifier();
		ArrayList<GraphItem> litems = buildItemsSequence(sb);

		// Get and discard the virtual operator.
		Object virtualOperator=getVirtualOperator(litems);
		litems.remove(virtualOperator);
		if (litems.size() == 0) {
			throw new Exception(item.getNickName()+": "+Ariane.lang.getString("Generator.error.incomplete"));
		}

		fp.println("def macro"+item.getId()+"_"+item.getGraph().getIdentifier()+"(*args):");
		// Comments
		fp.println("''' "+item.getNickName()+"\n",1);
		for (GraphItem note : sb.getGraphItems()) {
			if (note instanceof NoteItem) {
				fp.println(note.getFormula(),1);
			}
		}
		fp.println("'''",1);
		fp.println("argc = 0",1);
		// Get parameters.
		fp.println("# Input parameters",1);
		for (int i = 0; i < item.getInSlotParameters().size(); i++) {
			fp.println(variable(num_subgraph,0,i)+" = args[argc]; argc += 1",1);
		}

		// Get mask.
		fp.println("# Input data files",1);
		InSlotMask mask = item.getInSlotMask();
		if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask,item.getGraph().getIdentifier()))) {
			int rz = sb.getVirtualOperator().getVirtualInSlotMask().getRank();
			fp.println("link(args[argc], \"" + createImageName(rz,0,num_subgraph)+"\"); argc += 1",1);
		}

		// Get input images.
		rank=0;
		for (InSlotImage isi : item.getInSlotImages() ) {
			// Unlinked input!
			if (isi.getConnectedSlot() == null) {
				throw new Exception("Generator.error.incomplete");
			}
			fp.println("link(args[argc], \""+createImageName(rank++,0,num_subgraph)+"\"); argc += 1",1);
		}

		generateSubroutineBody(litems,num_subgraph,fp,1);

		// Get output images.
		VirtualItem vo = (VirtualItem)virtualOperator;
		fp.println();
		fp.println("# Output images",1);
		for (int i=0; i<item.getOutSlotImages().size(); i++) {
			InSlotImage isi = vo.getVirtualOutSlotImages().get(i);
			fp.println("link(\""+getInputImage(isi,num_subgraph)+"\",args[argc]); argc += 1",1);
		}

		// Output result.
		InSlotParameter isp = (InSlotParameter)vo.getVirtualOutSlotResult();
		if (isp.getConnectedSlot() == null) {
			fp.println("return 0;",1);
		} else {
			fp.println("return "+variable(num_subgraph,isp.getConnectedSlot().getGraphItem().getId(),isp.getConnectedSlot().getRank()),1);
		}
		fp.println("");

		generateSubroutinesDefinition(litems,fp);
	}
}