/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.command.AddItemEdit;
import fr.greyc.pantheon.ariane.graph.command.MoveItemEdit;
import fr.greyc.pantheon.ariane.graph.command.MoveLinkEdit;
import fr.greyc.pantheon.ariane.graph.command.RemoveItemEdit;
import fr.greyc.pantheon.ariane.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.graph.item.FileNameItem;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.LoaderItem;
import fr.greyc.pantheon.ariane.graph.item.SaverItem;
import fr.greyc.pantheon.ariane.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.graph.slot.InSlot;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotMask;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.Slot;
import fr.greyc.pantheon.ariane.graph.slot.Link;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.SubroutineTreeItem;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;
import fr.greyc.pantheon.ariane.xml.XMLFileOutputStream;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.Autoscroll;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetAdapter;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.JFileChooser;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JViewport;
import javax.swing.undo.CompoundEdit;
import javax.swing.undo.UndoManager;
import javax.swing.undo.UndoableEdit;

/**
 * A <i>Graph</i> holds an executable program as a list of linked items.
 * Items are included using Drag&Drop from the explorer, and linked
 * from a Drag&Drop on their slots.
 * <p>Copy/Paste uses XML version of each selected item.</p>
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 22, 2004
 * @version Frédéric LAIR - May 24, 2007
 * @version Régis CLOUARD - Nov 25, 2007 (restructuration)
 */
public class Graph extends JPanel implements Printable, Autoscroll {
	/** The maximum distance of the mouse pointer to the object to select it */
	private static double MOUSE_PRECISION=8.0F;

	/** The name of the graph. */
	private String _name;

	/** True if the graph is a root graph, false if it's is a subgraph. */
	private final boolean _isRoot;

	/** The counter of created graphs for the session. */
	private static int _graphCount=0;

	/** The counter of created items for the graph. */
	private int _graphItemCount=0;

	/** The identifier number of the graph. */
	private final int _identifier;

	/** The absolute file name. */
	private File _filename;

	/** The state of the root graph between modified or not (from the last saving). */
	private boolean _isModified=false;

	/** Flag used to determine if the graph is already alive. */
	private boolean _isAlive=true;

	/** The current state of execution : true when it is currently running. */
	static private boolean _executionState;

	/** The list of items defined in the graph. */
	private final LinkedList<GraphItem> _graphItems=new LinkedList<GraphItem>();

	/** A graph is a drop target site. */
	@SuppressWarnings("unused")
	private DropTarget _dropTarget = null;

	/**
	 * Creates a new graph with the given name.
	 * This graph can be either a root graph or a subgraph.
	 * 
	 * @param name the name of the graph.
	 * @param isRoot true if the graph is the root graph,
	 *               false if its is a subgraph?
	 */
	public Graph( String name, boolean isRoot ) {
		_name = name;
		_isRoot = isRoot;
		_identifier = ++_graphCount;
		_filename = null;
		setLayout(new XYLayout());
		setBackground(_backgroundColor);
		setAutoscrolls(true);

		setFocusCycleRoot(true);
		setFocusTraversalPolicyProvider(true);
		setFocusTraversalPolicy(new GraphFocusTraversalPolicy ());

		addListeners();
	}

	/**
	 * Creates a new graph from the given file.
	 * This graph is a root graph.
	 * 
	 * @param file a regular Ariane file.
	 */
	public Graph( File file ) {
		this(Graph.getAbsoluteName(file.getName()),true);
		_filename= file.getAbsoluteFile();
	}

	/**
	 * Adds the mouse and the windows listeners.
	 */
	private final void addListeners( ) {
		_dropTarget = new DropTarget(this, new DropTargetAdapter() {
			/**
			 * Called while a drag operation is ongoing, when the mouse pointer
			 * enters the operable part of the drop site for the DropTarget
			 * registered with this listener.
			 * Accepts the drop action for a move action.
			 * 
			 * @param e the DropTargetDragEvent.
			 */
			@Override
			public final void dragEnter( DropTargetDragEvent e ) {
				e.acceptDrag(DnDConstants.ACTION_MOVE);
			}

			/**
			 * The DropTarget intercepts drop() notifications before the
			 * registered DropTargetListener gets them.
			 * Decides if this drop is a valid action.
			 * Valid drop action can be:
			 * <ul>
			 * <li>Disconnection of two slots.
			 * <li>Addition of a new item if the graph.
			 * </ul>
			 * 
			 * @param e the DropTargetDropEvent.
			 */
			@Override
			public final void drop( DropTargetDropEvent e ) {
				try {
					if (e.isDataFlavorSupported(OutSlot.flavors[0])) { // Disconnection.
						OutSlot incoming=(OutSlot)e.getTransferable().getTransferData(OutSlot.flavors[0]);
						startActionEdition();
						incoming.disconnectAll();
						endActionEdition();
						unSelectAll();
						e.acceptDrop(DnDConstants.ACTION_MOVE);
						e.dropComplete(true);
					} else if (e.isDataFlavorSupported(InSlot.flavors[0])) { // Disconnection.
						InSlot incoming=(InSlot)e.getTransferable().getTransferData(InSlot.flavors[0]);
						startActionEdition();
						incoming.disconnect();
						endActionEdition();
						unSelectAll();
						e.acceptDrop(DnDConstants.ACTION_MOVE);
						e.dropComplete(true);
					} else if (e.isDataFlavorSupported(SubroutineTreeItem.flavors[0])) { // insert subroutine
						TreeItem incoming=(SubroutineTreeItem)e.getTransferable().getTransferData(SubroutineTreeItem.flavors[0]);
						GraphItem newItem=incoming.createGraphItem(Graph.this);
						if (newItem == null ) {
							e.acceptDrop(DnDConstants.ACTION_MOVE);
							e.dropComplete(true);
							return;
						}
						startActionEdition();
						addItem(newItem,e.getLocation(),true);
						endActionEdition();
						validate(); // force to display the new component. 
						unSelectAll();
						addItemInSelection(newItem,false);
						newItem.requestFocusInWindow();
						e.acceptDrop(DnDConstants.ACTION_MOVE);
						e.dropComplete(true);
					} else if (e.isDataFlavorSupported(TreeItem.flavors[0])) { // Insert Item
						TreeItem incoming=(TreeItem)e.getTransferable().getTransferData(TreeItem.flavors[0]);
						GraphItem newItem=incoming.createGraphItem(Graph.this);
						if (newItem == null ) {
							e.acceptDrop(DnDConstants.ACTION_MOVE);
							e.dropComplete(true);
							return;
						}
						startActionEdition();
						addItem(newItem,e.getLocation(),true);
						endActionEdition();
						validate(); // force to display the new component.
						unSelectAll();
						addItemInSelection(newItem,false);
						e.acceptDrop(DnDConstants.ACTION_MOVE);
						newItem.requestFocusInWindow();
						e.dropComplete(true);
					} else { // Unknown drop action.
						e.rejectDrop();
						e.dropComplete(false);
					}
				} catch(UnsupportedFlavorException exc) {
					e.rejectDrop();
					e.dropComplete(false);
				} catch(IOException exc) {
					e.rejectDrop();
					e.dropComplete(false);
				} catch(NullPointerException exc) {
					exc.printStackTrace();
					e.rejectDrop();
					e.dropComplete(false);
				}
			}
		});
		addMouseListener( new MouseAdapter() {
			@Override
			public final void mousePressed(MouseEvent e) {
				requestFocusInWindow();
				onMousePressed(e);
			}
			@Override
			public final void mouseReleased(MouseEvent e) {
				onMouseReleased(e);
			}
		});
		addMouseMotionListener( new MouseMotionAdapter() {
			@Override
			public final void mouseDragged(MouseEvent e) {
				onMouseDragged(e);
			}
		});
		addKeyListener( new KeyAdapter() {
			/**
			 * Moves the current selection of 1 pixel.
			 * 
			 * @param e the key event.
			 */
			@Override
			public final void keyPressed( KeyEvent e ) {
				onKeyPressed(e);
			}
		});
		addFocusListener(new FocusListener() {
			@Override
			public final void focusLost(FocusEvent e) {
			}

			/**
			 * Checks for each file item if the associated file
			 * has changed on the file system.
			 * If so reset the item.
			 */
			@Override
			public final void focusGained(FocusEvent e){ 
				for ( GraphItem item : _graphItems ) {
					if ( item instanceof LoaderItem ) {
						((LoaderItem)item).synchronizeWithFileSystem();
					} else if ( item instanceof FileNameItem ){
						((FileNameItem)item).synchronizeWithFileSystem();
					}
				}
			}
		});
	}

	/**
	 * Closes the current graph. Asks the user to confirm the closing.
	 * 
	 * @return true if the graph has been closed by the user.
	 */
	public final boolean close( ) {
		if (!_isRoot) {
			closeFrame();
			return true;
		} else {
			if (deleteConfirmed()) {
				closeFrame();
				dispose();
				Ariane.workspace.buildRecentFilesListMenu();
				return true;
			} else {
				return false;
			}
		}
	}

	/**
	 * Disposes definitively the graph and all its
	 * components from the workspace.
	 * Tries to avoid memory leaks.
	 */
	public void dispose( ) {
		_isAlive=false;
		while (!_graphItems.isEmpty()) {
			_graphItems.removeLast().dispose();
		}
		_undoRedoManager.discardAllEdits();
		unSelectAll();
		_graphItems.clear(); // !! do not set it to null !
		_dropTarget=null;
		setLayout(null);
	}

	///// 
	///// F R A M E S
	///// 

	/**
	 * Closes the related frame and all attached frames,
	 * those of the composite items.
	 * Deletes them from the workspace list of frames.
	 */
	public final void closeFrame( ) {
		for (GraphItem item : _graphItems) {
			if (item instanceof CompositeItem)
				((CompositeItem)item).getSubGraph().closeFrame();
		}
		Ariane.workspace.removeGraph(this);
	}

	/**
	 * Renames the related tab, and also renames
	 * all its subgraph tabs.
	 */
	public final void renameFrame( ) {
		for (GraphItem item : _graphItems) {
			if (item instanceof CompositeItem)
				((CompositeItem)item).getSubGraph().renameFrame();
		}
		// Use _name and not getName() to avoid overriding
		Ariane.workspace.renameGraph(this, _name);
	}

	///// 
	///// M E N U
	///// 

	/**
	 * Creates the contextual menu.
	 * 
	 * @param x the x-coordinate of the contextual menu.
	 * @param y the y-coordinate of the contextual menu.
	 */
	private final void displayPopupMenu( int x, int y ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item = new JMenuItem(Ariane.lang.getString("Graph.Menu.redraw"));
		item.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) { validate(); repaint(); }
		});
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("Graph.Menu.execute"));
		item.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) { execute(); }
		});
		item.setEnabled((_selectedItems.size() > 0));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("Graph.Menu.reset"));
		item.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) { resetAll(); }
		});
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("Graph.Menu.copy"));
		item.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) { Ariane.workspace.copy(); }
		});
		item.setEnabled((_selectedItems.size() > 0));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("Graph.Menu.cut"));
		item.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) { Ariane.workspace.cut(); }
		});
		item.setEnabled((_selectedItems.size() > 0));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("Graph.Menu.paste"));
		item.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) { Ariane.workspace.paste(); }
		});
		item.setEnabled(Ariane.workspace.isPastable());
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("Graph.Menu.delete"));
		item.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) { Ariane.workspace.erase(); }
		});
		item.setEnabled((_selectedItems.size() > 0));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("Graph.Menu.close"));
		item.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) { if (deleteConfirmed()) close(); }
		});
		item.setEnabled((_selectedItems.size() > 0));
		popupMenu.add(item);

		popupMenu.show (this,x,y);
	}

	/**
	 * Returns true if the graph deletion is confirmed.
	 * 
	 * @return true if the deletion is confirmed.
	 */
	private final boolean deleteConfirmed( ) {
		if (!_isModified)
			return true;
		int answer = JOptionPane.showConfirmDialog(Ariane.workspace,
				Ariane.lang.getString("Graph.quit.message.part1") + getName()+Ariane.lang.getString("Graph.quit.message.part2"),
				Ariane.lang.getString("Graph.quit.title"),
				JOptionPane.YES_NO_CANCEL_OPTION);
		if (answer == JOptionPane.YES_OPTION) {
			save();
			return true;
		} else if (answer == JOptionPane.NO_OPTION) {
			return true;
		} else { // cancel
			return false;
		}
	}

	///// 
	///// A C C E S S O R S
	///// 

	/**
	 * Returns the super graph item that contains
	 * this graph.
	 * 
	 * @return the graph item that encloses this graph.
	 */
	public GraphItem getSuperGraphItem( ) {
		return null;
	}

	/**
	 * Creates a string representation of the graph
	 * made with the name and the properties of the graph.
	 * 
	 * @return a string representation of the graph identification.
	 */
	@Override
	public final String toString( ) {
		return _name + "[" + _graphItems.size() + " éléments]";
	}

	/**
	 * Returns the name of the file without the extension and
	 * including super graphs (separated by ::).
	 * 
	 * @param filename a regular file name.
	 * @return the absolute name.
	 */
	private static final String getAbsoluteName( String filename ) {
		int i= filename.lastIndexOf('.');
		return filename.substring(0,i);
	}

	/**
	 * Changes the current name.
	 * 
	 * @param name the new name.
	 */
	@Override
	public final void setName( String name ) {
		_name=name;
	}

	/**
	 * Returns the last name of the graph.
	 * 
	 * @return the name.
	 */
	@Override
	public String getName( ) {
		return _name;
	}

	/**
	 * Returns the identification number of the graph.
	 * 
	 * @return the identification number.
	 */
	public final int getIdentifier( ) {
		return _identifier;
	}

	/**
	 * Returns the identifier number of the next graph to be created.
	 * 
	 * @return the identifier number.
	 */
	public final int getGraphItemCount( ) {
		return _graphItemCount;
	}

	/**
	 * Sets the graph to modified idf the value is true
	 * or not modified if the value is false.
	 * 
	 * @param value if true the graph is set to modified.
	 */
	public void setModified( boolean value ) {
		if (getParent()!=null && _isAlive) {
			Ariane.workspace.setModifiedSaveB(value);
			_isModified = value;
		}
	}

	/**
	 * Returns true if the graph is a root graph.
	 * 
	 * @return true if it is a root graph.
	 */
	public final boolean isRoot( ) {
		return _isRoot;
	}

	/**
	 * Returns the root graph of this graph.
	 * 
	 * @return the root graph.
	 */
	public Graph getRoot( ) {
		return this;
	}

	/**
	 * Returns the list of all items in the graph.
	 * 
	 * @return the LinkedList.
	 */
	public final LinkedList<GraphItem> getGraphItems( ) {
		return _graphItems;
	}

	/**
	 * Returns the item specified by the given identifier.
	 * Takes into account the current read session, it means
	 * the identifier depends on the current read session identifier.
	 * 
	 * @param identifier the identifier number of the item
	 * @return the item or null.
	 */
	public final GraphItem getGraphItem( int identifier ) {
		for (int i = _graphItems.size()-1; i>=0; i--) {
			GraphItem item = _graphItems.get(i);
			if (item.getCurrentReadPasteSession() == Workspace.getCurrentReadPasteSession()
					&& item.getReadIdentifier() == identifier) {
				return item;
			}
		}
		return null;
	}

	///// 
	///// A C T I O N S
	/////

	/**
	 * Deletes the given list of items.
	 * 
	 * @param litems the list of items to be deleted.
	 */
	private final void deleteItems( LinkedList<GraphItem> litems ) {
		startActionEdition();
		while (!litems.isEmpty()) {
			GraphItem item = litems.removeFirst();
			item.remove();
			removeItem(item,true);
		}
		endActionEdition();
	}

	/**
	 * Deletes one item from the graph.
	 * 
	 * @param item the item to be deleted.
	 */
	public final void deleteItem( GraphItem item ) {
		startActionEdition();
		item.remove();
		removeItem(item,true);
		endActionEdition();
	}

	/**
	 * Adds the specified item at the specified coordinates of the graph.
	 * 
	 * @param item the item to be added.
	 * @param xycoord the coordinate.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */	
	public final void addItem( GraphItem item, Point xycoord, boolean historical ) {
		_graphItems.add(item);
		_graphItemCount++;
		// Keep the 0 at the end of the add method !!!!
		add(item,new XYConstraint(xycoord.x,xycoord.y,0,0));
		setComponentZOrder(item, 0); // Put it in front
		if (historical) {
			validateAction(new AddItemEdit(this,item));
		}
	}

	/**
	 * Removes the related graphical component
	 * of the item from the graph representation.
	 * 
	 * @param item the item to be removed.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void removeItem( GraphItem item, boolean historical ) {
		super.remove(item);
		_graphItems.remove(item);
		item.remove();
		if (historical) {
			validateAction(new RemoveItemEdit(this,item));
		}
	}

	/**
	 * Moves the specified item to the new coordinates.
	 * 
	 * @param item the item to be moved.
	 * @param loc the new location.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void moveItem( GraphItem item, Point loc, boolean  historical ) {
		if (historical) {
			validateAction(new MoveItemEdit(this,item,item.getLocation(),loc));
		}
		item.setLocation(loc);
		revalidate();
	}

	/**
	 * Moves a list of items with the specified translation value.
	 * 
	 * @param litems the list of items.
	 * @param translation the translation value.
	 */
	public final void moveItems( LinkedList<GraphItem> litems, Point translation ) {
		for (GraphItem item : litems) {
			Point xy = item.getLocation();
			xy.translate(-translation.x,-translation.y);
			moveItem(item,xy,true);
		}
		repaint();
	}

	/**
	 * Moves a list of items with the specified translation value.
	 * 
	 * @param litems the list of items.
	 * @param from the origin point.
	 * @param to the destination point.
	 */
	public final void moveItems( LinkedList<GraphItem> litems, Point from, Point to ) {
		Point translation = new Point(from.x-to.x,from.y-to.y);
		for (GraphItem item : litems) {
			Point origin = new Point(item.getLocation().x + translation.x, item.getLocation().y + translation.y);
			validateAction(new MoveItemEdit(this,item,origin,item.getLocation()));
		}
		repaint();
	}

	/**
	 * Moves the given point from the start coordinates
	 * to the destination coordinates.
	 *
	 * @param link the link to be modified.
	 * @param from the start coordinates.
	 * @param to the destination coordinates.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	private final void moveLink( Link link, Point from, Point to, boolean historical ) {
		if (historical) {
			validateAction(new MoveLinkEdit(this,_selectedLink,from,to));
		}
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * @return true if the graph has been modified, false if not.
	 */
	public boolean getStatus(){
		return _isModified;
	}

	/**
	 * Resets all items of the graph.
	 */
	public void resetAll( ) {
		for (GraphItem gi : _graphItems)
			gi.resetAll();
	}

	/**
	 * Resets all items of the graph.
	 */
	public void resetSelectedItems( ) {
		if (_selectedItems.size() > 0) {
			for (GraphItem gi : _selectedItems) {
				gi.resetAll();
			}
		}
	}

	/**
	 * Stops the current execution.
	 */
	public final void stopExecution( ) {
		_executionState=false;
		Ariane.workspace.setExecution(false);
	}

	/**
	 * Determines if the current execution session is stopped.
	 * An execution can be stopped either by user or by fatal error.
	 * 
	 * @return true if execution is stopped.
	 */
	public static final boolean isExecutionStopped( ) {
		return !_executionState;
	}

	/**
	 * Executes the current selected item
	 * in a separate Thread.
	 */
	public final void execute( ) {
		if (_selectedItems.size() == 0) {
			Ariane.workspace.printWarning(Ariane.lang.getString("Graph.execution.warning1"));
		} else {
			_executionState=true;
			// Creates a thread and executes the graph with it.
			ThreadGraphExecution thread = new ThreadGraphExecution(_selectedItems);
			thread.start();
		}
	}

	/**
	 * Executes the current selected item in a separate Thread.
	 * @param loader the loader used as the directory source.
	 */
	public final void executeFolder( LoaderItem loader ) {
		File loaderFile = new File(loader.getBinary());
		if (loaderFile==null || !loaderFile.exists()) {
			Ariane.workspace.printErrorMessage(Ariane.lang.getString("Graph.execution.error1")+loader.getName()+"("+loader.getId()+")");
			return;
		}

		// Get the list of items to be  executed starting from the loader.
		LinkedList<GraphItem> graphItems = new LinkedList<GraphItem>();
		LinkedList<GraphItem> visitedItems = new LinkedList<GraphItem>();
		LinkedList<GraphItem> stack = new LinkedList<GraphItem>();
		stack.add(loader);
		while (stack.size()>0) {
			GraphItem item = stack.removeLast();
			boolean lastItem=true;
			if (visitedItems.contains(item)) {
				continue;
			}
			visitedItems.add(item);
			for ( InSlotParameter slot : item.getInSlotParameters()) {
				if (slot.isConnected()) {
					GraphItem neighbour = slot.getConnectedSlot().getGraphItem();
					stack.add(neighbour);
					lastItem=false;
				}
			}
			for ( InSlotImage slot : item.getInSlotImages()) {
				if (slot.isConnected()) {
					GraphItem neighbour = slot.getConnectedSlot().getGraphItem();
					stack.add(neighbour);
				}
			}
			InSlotMask maskSlot = item.getInSlotMask();
			if (maskSlot != null && maskSlot.isConnected()) {
				GraphItem neighbour = maskSlot.getConnectedSlot().getGraphItem();
				stack.add(neighbour);
			}
			for ( OutSlotImage slot : item.getOutSlotImages()) {
				if (slot.isConnected()) {
					lastItem=false;
					for ( InSlot neighbourSlot : slot.getConnectedSlots()) {
						if (neighbourSlot.isConnected()) {
							stack.add(neighbourSlot.getGraphItem());
						}
					}
				}
			}
			if (lastItem) {
				graphItems.add(item);
			}
		}

		// Get the savers to store data in the appropriate files.
		HashMap<SaverItem,File> saverDirectories = new HashMap<SaverItem,File>();
		for ( GraphItem item : graphItems ) {
			if (item instanceof SaverItem ) {
				File saverFile = new File(item.getBinary());
				if (saverFile == null || saverFile.getParentFile() == null || !saverFile.getParentFile().exists()) {
					Ariane.workspace.printErrorMessage(Ariane.lang.getString("Graph.execution.error1")+item.getName()+"("+item.getId()+")");
					return;
				}
				saverDirectories.put((SaverItem)item, saverFile);
			}
		}

		int suffixPosition=loaderFile.getName().lastIndexOf('.');
		final String loaderSuffix= (suffixPosition>0)? loaderFile.getName().substring(suffixPosition) : "";
		File[] files=loaderFile.getParentFile().listFiles( new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				File file = new File(dir,name);
				if (file.isDirectory() || file.isHidden()) {
					return false;
				}
				String suffix="";
				int suffixPosition=name.lastIndexOf('.');
				if (suffixPosition>0) {
					suffix = name.substring(suffixPosition);
				}
				return suffix.equals(loaderSuffix);
			}
		});

		if (files == null) {
			Ariane.workspace.printErrorMessage(Ariane.lang.getString("Graph.execution.error2")+loaderFile.getParent());
			return;
		}
		Arrays.sort(files);

		_executionState = true;
		ThreadDirectoryExecution thread = new ThreadDirectoryExecution(graphItems, loader, loaderFile, files, saverDirectories);
		thread.start();
	}

	///// 
	///// P R E F E R E N C E S
	/////

	/** The default background color. */
	private static final Color _defaultBackgroundColor = new Color(200,200,200);
	/** The current background color. */
	private static Color _backgroundColor = _defaultBackgroundColor;

	/** The default image link color. */
	private static final Color _defaultImageLinkColor = new Color(0,51,255);
	/** The current image link color. */
	private static Color _imageLinkColor = _defaultImageLinkColor;

	/** The default value link color. */
	private static final Color _defaultValueLinkColor = new Color(255,0,0);
	/** The current value link color. */
	private static Color _valueLinkColor = _defaultValueLinkColor;

	/** The default mask link color. */
	private static final Color _defaultMaskLinkColor = new Color(0,0,0);
	/** The current mask link color. */
	private static Color _maskLinkColor = _defaultMaskLinkColor;

	/**
	 * @return the default background color.
	 */
	public static Color getDefaultBackgroundColor( ) { return _defaultBackgroundColor; }

	/**
	 * @return the current background color.
	 */
	public static Color getBackgroundColor( ) { return _backgroundColor; }

	/**
	 * Sets the background color.
	 * 
	 * @param color the new color.
	 */
	public static void setBackgroundColor( Color color ) {
		_backgroundColor=color;
	}

	/**
	 * Returns the image link color.
	 * 
	 * @return the color.
	 */
	public static final Color getDefaultImageLinkColor( ) { return _defaultImageLinkColor; }

	/**
	 * @return the image link color.
	 */
	public static final Color getImageLinkColor( ) { return _imageLinkColor; }

	/**
	 * Sets the image link color.
	 * 
	 * @param color the new color.
	 */
	public static final void setImageLinkColor( Color color ) { _imageLinkColor=color; }

	/** @return the default mask link color. */
	public static final Color getDefaultMaskLinkColor( ) { return _defaultMaskLinkColor; }

	/** @return the current mask link color. */
	public static final Color getMaskLinkColor( ) { return _maskLinkColor; }
	/**
	 * Sets the mask link color.
	 * 
	 * @param color the new color.
	 */
	public static final void setMaskLinkColor( Color color ) { _maskLinkColor=color; }

	/**
	 * @return the default value link color.
	 */
	public static final Color getDefaultValueLinkColor( ) { return _defaultValueLinkColor; }

	/**
	 * @return the current value link color.
	 */
	public static final Color getValueLinkColor( ) { return _valueLinkColor; }

	/**
	 * Sets the value link color.
	 * 
	 * @param color the new color.
	 */
	public static final void setValueLinkColor( Color color ) { _valueLinkColor=color; }

	///// 
	///// P A I N T I N G  /  P R I N T I N G
	/////

	/**
	 * Prints the current graph in the given page.
	 * 
	 * @param g the graphics context.
	 * @param pageFormat the properties of the page to be printed.
	 * @param pageIndex the index of the page to be printed.
	 * @return NO_SUCH_PAGE in case of error, or PAGE_EXISTS.
	 */
	@Override
	public final int print( Graphics g, PageFormat pageFormat, int pageIndex ) {
		if (pageIndex > 0)
			return(NO_SUCH_PAGE);
		Graphics2D g2d = (Graphics2D)g;

		// Print the title: the graph name.
		//		g2d.drawString(Ariane.lang.getString("Graph.print.title")+" "+getName(),
		//				(int)pageFormat.getImageableX(),
		//				(int)pageFormat.getImageableY()+16);

		double scaleW=pageFormat.getImageableWidth()/getWidth();
		double scaleH=(pageFormat.getImageableHeight()-24)/getHeight(); // -24 for title.

		if (scaleH < scaleW) scaleW=scaleH;
		if (scaleW > 1.0) scaleW=1.0;
		g2d.setPaint(Color.black);

		g2d.translate(pageFormat.getImageableX(), pageFormat.getImageableY()+24);
		g2d.scale(scaleW,scaleW);
		printAll(g2d);
		return(PAGE_EXISTS);
	}

	/**
	 * Overrides the <i>paintComponent</i>
	 * to add the links drawing.
	 * 
	 * @param gc the graphics context.
	 */
	@Override
	public final void paintComponent( Graphics gc ) {
		super.paintComponent(gc);
		drawBindings(gc);
		if (_rubberBandBox!=null) {
			_rubberBandBox.erase(gc);
		}
	}

	/**
	 * Overrides the <i>printComponent</i>
	 * to add the links drawing.
	 * 
	 * @param gc the graphics context.
	 */
	@Override
	public final void printComponent( Graphics gc ) {
		drawBindings(gc);
	}

	/**
	 * Forces repainting of all items in the graph.
	 */
	public final void repaintAll( ) {
		setBackground(_backgroundColor);
		for (int i = 0; i < _graphItems.size(); i++) {
			_graphItems.get(i).draw();
		}
	}

	/**
	 * Draws all links of the graph.
	 * 
	 * @param gc the graphics context.
	 */
	private void drawBindings( Graphics gc ) {
		for (int i = 0; i < _graphItems.size(); i++) {
			GraphItem op = _graphItems.get(i);
			// Input images.
			for (int j = 0; j < op.getInSlotImages().size(); j++)
				((op.getInSlotImages().get(j))).drawLink(gc);
			// Mask.
			if (op.getInSlotMask() != null)
				(op.getInSlotMask()).drawLink(gc);
			// Parameters.
			for (int j = 0; j < op.getInSlotParameters().size(); j++)
				((op.getInSlotParameters().get(j))).drawLink(gc);
		}
	}

	///// 
	///// I T E M S   S E L E C T I O N
	/////

	/** The list of all selected items. */
	private final LinkedList<GraphItem> _selectedItems = new LinkedList<GraphItem>();

	/** The current link selected, if any. */
	private Link _selectedLink;

	/** Stores dragging information. */
	private DraggingSelection _draggingSelection;

	/** Currently dragging a link? */
	private boolean _draggingLink = false;

	/** the origin of a movement. */
	private Point _movementOrigin;

	/** The current selection rectangle. */
	private RubberBandBox _rubberBandBox;

	/**
	 * Manages the event "button pressed". If the button is:
	 * <ul>
	 * <li>the right button: displays the popup menu (UNIX).
	 * <li>the left button: selects a link or a point in the link
	 *                  or initialize a new selection rectangle.
	 * </ul>
	 * 
	 * @param e the event properties.
	 */
	private final void onMousePressed( MouseEvent e ) {
		if (e.isPopupTrigger()){	// Used by Unix/Linux: display popup menu
			displayPopupMenu(e.getX(), e.getY());
		} else {
			Link link = detectLink(e.getX(), e.getY());
			unSelectAll();
			Ariane.workspace.displayProperties(null);
			if ( link != null) {	// if a link is selected then reshape the link.
				_selectedLink = link;
				link.setSelected(true, e.getPoint());
				setCursor(link.getCursor());
				_draggingLink = true;
				_movementOrigin = e.getPoint();
			} else {				// begin a selection rectangle.
				_rubberBandBox=new RubberBandBox(e.getX(), e.getY());
			}
		}
	}

	/**
	 * Manages the event "drag" on the graph editor:
	 * displays the selection rectangle from the mouse pointer.
	 * 
	 * @param e the event properties.
	 */
	private final void onMouseDragged( MouseEvent e ) {
		if (_rubberBandBox!=null) { // case of rubber band box
			_rubberBandBox.update(e.getPoint());
		}
		if (_draggingLink) { // drag a point.
			_selectedLink.move(e.getPoint());
			repaint();
		}
	}

	/**
	 * Manages the event "button released". If the button is :
	 * <ul>
	 * <li>right button: displays the popup menu (Windows)
	 * <li>left button: updates items and points in the selection rectangle.
	 * </ul>
	 * 
	 * @param e the event properties.
	 */
	private final void onMouseReleased( MouseEvent e ) {
		if (e.isPopupTrigger()) { // Used by Windows.
			displayPopupMenu(e.getX(), e.getY());
		} else {
			if (_rubberBandBox!=null) {
				for (GraphItem item : _graphItems) {
					// add all items in the selection rectangle in the selection.
					if (_rubberBandBox.contains(new Rectangle(item.getLocation(), item.getSize()))) {
						addItemInSelection(item,true);
					}
				}
				_rubberBandBox.erase();
				_rubberBandBox=null;
			} else if (_draggingLink) {
				startActionEdition();
				moveLink(_selectedLink,_movementOrigin,e.getPoint(),true);
				endActionEdition();
				// restore the cursor.
				setCursor(null);
				_selectedLink.setSelected(false, null);
				_selectedLink = null;
				_draggingLink = false;
			}
			repaint();
		}
	}

	/**
	 * Handles the key event.
	 * 
	 * @param e a key event.
	 */
	public void onKeyPressed( KeyEvent e ) {
		if (_selectedItems.size() == 0) return;
		switch (e.getKeyCode()) {
		case KeyEvent.VK_RIGHT:
		case KeyEvent.VK_KP_RIGHT:
			moveSelection(new Point(-1,0));
			break;
		case KeyEvent.VK_LEFT:
		case KeyEvent.VK_KP_LEFT:
			// check if one of the selected items will disappear!!
			for (GraphItem x : _selectedItems) {
				Point p = x.getLocation();
				if (p.x==0) return;
			}
			moveSelection(new Point(1,0));
			break;
		case KeyEvent.VK_UP:
		case KeyEvent.VK_KP_UP:
			// check if one of the selected items will disappear!!
			for (GraphItem x : _selectedItems) {
				Point p = x.getLocation();
				if (p.y==0) return;
			}
			moveSelection(new Point(0,1));
			break;
		case KeyEvent.VK_DOWN:
		case KeyEvent.VK_KP_DOWN:
			moveSelection(new Point(0,-1));
			break;
		}
	}

	public void displayOutputs( int output ) {
		if (output>=0 && output<=9) {
			for (GraphItem x : _selectedItems) {
				if (x.getOutSlotImages().size()>output) {
					x.getOutSlotImages().get(output).showImage();
				}
			}
		}
	}

	/**
	 * Select items in the graph (except virtual operator).
	 * Builds the related selection rectangle.
	 */
	public final void selectAll( ) {
		unSelectAll();
		for (int i=0; i< _graphItems.size(); i++) {
			GraphItem item=_graphItems.get(i);
			if (!(item instanceof VirtualItem))
				addItemInSelection(item,true);
		}
	}

	/**
	 * Unselects all items inside the current selection rectangle.
	 */
	public final void unSelectAll( ) {
		for (GraphItem gi : _selectedItems) {
			gi.setSelected(false);
		}
		_selectedLink = null;
		_selectedItems.clear();
		_rubberBandBox= null;
		repaint();
	}

	public final void selectItem( GraphItem item ) {
		unSelectAll();
		_selectedItems.add(item);
		item.setSelected(true);
	}

	public final GraphItem getFirstSelectedItem() {
		if (_selectedItems.size()>0) {
			return _selectedItems.get(0);
		}
		return null;
	}

	/**
	 * Adds the given item in the current selection according to the
	 * specified added value:
	 * <ul>
	 * <li>Shift pressed: adds the item to the current selection.
	 * <li>Shift pressed and item already in the selection: removes it from the selection.
	 * <li>Shift not pressed : builds a new selection with this item.
	 * </ul>
	 * 
	 * @param item the item to be added.
	 * @param shift key shift pressed?
	 */
	public final void addItemInSelection( GraphItem item, boolean shift ) {
		if (shift) {
			if (item.isSelected()){
				_selectedItems.remove(item);
				item.setSelected(false);
			} else {
				item.setSelected(true);
				_selectedItems.add(item);
			}
		} else {
			selectItem(item);
		}
	}

	/**
	 * Returns the coordinate of the minimum rectangle
	 * that includes all items in the graph.
	 * 
	 * @return the rectangle.
	 */
	private final Rectangle getSelectedRectangle( ) {
		int xs=100000, ys=100000, xe=0,ye=0;
		for (int i=0; i< _selectedItems.size();i++){
			GraphItem item = _selectedItems.get(i);
			Point itemcoord = item.getLocation();
			if (itemcoord.x<xs)
				xs=itemcoord.x;
			if (itemcoord.y<ys)
				ys=itemcoord.y;
			if (item.getWidth()+itemcoord.x>xe)
				xe=item.getWidth()+itemcoord.x;
			if (item.getHeight()+itemcoord.y>ye)
				ye=item.getHeight()+itemcoord.y;
		}
		return new Rectangle(xs,ys,xe-xs,ye-ys);
	}

	/**
	 * Drags the current selection of items and points
	 * on the panel.
	 * 
	 * @param e a mouse event with coordinate and keyboard state.
	 */
	public final void dragSelection( MouseEvent e ) {
		Point xycoord = e.getPoint();
		if (_draggingSelection == null) {	// start dragging.
			_draggingSelection = new DraggingSelection(getSelectedRectangle(), xycoord);
		} else {							// dragging the current selection...
			Point oldOrigin = _draggingSelection.getLocation();
			Point shift = _draggingSelection.getOffset();
			Point newOrigin;
			if (e.isShiftDown()) {
				if (Math.abs(oldOrigin.x-xycoord.x-shift.x) > Math.abs(oldOrigin.y-xycoord.y-shift.y)) {
					newOrigin = new Point(xycoord.x+shift.x, oldOrigin.y);
				} else {
					newOrigin = new Point(oldOrigin.x, xycoord.y+shift.y);
				}
			} else {
				newOrigin = new Point(xycoord.x+shift.x, xycoord.y+shift.y);
			}
			if (newOrigin.x<0) newOrigin.x = 0;
			if (newOrigin.y<0) newOrigin.y = 0;
			Point translation = _draggingSelection.getTranslation(newOrigin);
			scrollRectToVisible(new Rectangle(e.getX(), e.getY(), 1, 1));
			dragSelection(translation);
		}
	}

	/**
	 * Drops the current selection.
	 * Moves each item in the selection from the relative coordinate
	 * of the selection rectangle.
	 * 
	 * @return <i>true</i> if selection can be moved.
	 */
	public final boolean dropSelection( ) {
		if (_draggingSelection!=null) {
			moveSelection(_draggingSelection.getOriginLocation(), _draggingSelection.getLocation());
			_draggingSelection=null;
			return true;
		}
		return false;
	}

	/**
	 * Moves all items in the current selection from
	 * the specified coordinates to the specified coordinates.
	 * 
	 * @param from the origin coordinates.
	 * @param to the destination coordinates.
	 */
	private void moveSelection( Point from, Point to ) {
		startActionEdition();
		moveItems(_selectedItems,from, to);
		endActionEdition();
	}

	/**
	 * Moves all items in the current selection.
	 * 
	 * @param translation the translation vector.
	 */
	private void moveSelection( Point translation ) {
		startActionEdition();
		moveItems(_selectedItems,translation);
		endActionEdition();
	}

	/**
	 * Moves all items in the current selection.
	 * 
	 * @param translation the translation vector.
	 */
	private void dragSelection( Point translation ) {
		for (GraphItem item : _selectedItems) {
			Point xy = item.getLocation();
			xy.translate(-translation.x,-translation.y);
			moveItem(item,xy,false);
		}
		repaint();
	}

	/**
	 * Deletes items in the current selection.
	 */
	public final void deleteSelectedItems( ) {
		deleteItems(_selectedItems);
		unSelectAll();
	}

	/**
	 * Copies each current selected items.
	 * This is done by building the related XML tree
	 * and setting it to the clipboard.
	 * 
	 * @return the root of tree of items.
	 */
	public final XMLElement copySelectedItems( ) {
		try {
			XMLElement e=new XMLElement("ClipBoard");
			XMLElement items = new XMLElement("ItemList");
			for (int i = 0; i < _selectedItems.size(); i++)
				items.addChildren(_selectedItems.get(i).save(null));

			XMLElement connections = new XMLElement("ConnectionList");
			for (int i = 0; i < _selectedItems.size(); i++)
				connections.addChildren(_selectedItems.get(i).saveSlots());
			e.addChildren(items);
			e.addChildren(connections);
			return e;
		} catch (IOException ex) {
			return null;
		}
	}

	/**
	 * Pastes the current clipboard content.
	 * This is done by building graph items from the
	 * XML tree contained in the clipboard.
	 * Only connections between items that are not in
	 * the same selection are kept.
	 * 
	 * @param clipBoard the current clipboard to be pasted.
	 */
	public final void pasteSelectedItems( XMLElement clipBoard ) {
		try {
			Iterator<XMLElement> it = clipBoard.getChildren().iterator();
			startActionEdition();
			ArrayList<GraphItem> items = new ArrayList<GraphItem>();
			while (it.hasNext()) {
				XMLElement e = it.next();
				if (e.getName().equalsIgnoreCase("ItemList")) {
					items = GraphItem.load(this,e.getChildren(),true);
					Point min=items.get(0).getLocation();
					for ( GraphItem item : items) {
						Point point = item.getLocation();
						if (point.y < min.y) min.y=point.y;
						if (point.x < min.x) min.x=point.x;
					}
					Point viewPosition = ((JViewport)getParent()).getViewPosition();
					Point translation = new Point(viewPosition.x+5-min.x, viewPosition.y+5-min.y);

					// translate all item to be visible into the current viewport.
					for (GraphItem item : items) {
						Point coordinates = item.getLocation();
						coordinates.translate(translation.x,translation.y);
						add(item,new XYConstraint(coordinates.x,coordinates.y,0,0));
						setComponentZOrder(item, 0); // Put it in front
					}
				} else if (e.getName().equalsIgnoreCase("ConnectionList")) {
					Slot.load(this,e.getChildren());
				}
			}
			// start the new selection
			unSelectAll();
			for (int i=0; i< _graphItems.size(); i++) {
				GraphItem item = _graphItems.get(i);
				// Discard virtual operators.
				if (item.getCurrentReadPasteSession()==Workspace.getCurrentReadPasteSession() && !(item instanceof VirtualItem)) {
					addItemInSelection(item,true);
				}
			}
			endActionEdition();
			scrollRectToVisible(getSelectedRectangle());
		} catch(IOException ex){
		}
	}

	/**
	 * Finds the link under the mouse coordinates if any.
	 * 
	 * @param x the x coordinate of the mouse pointer.
	 * @param y the y coordinate of the mouse pointer.
	 * @return the link selected or null if any.
	 */
	private final Link detectLink( double x, double y ) {
		Link closest = null;
		double dist = MOUSE_PRECISION;
		for (GraphItem currentGI : getGraphItems()) {
			LinkedList<InSlot> allSlots = new LinkedList<InSlot>();
			allSlots.addAll(currentGI.getInSlotImages());
			allSlots.addAll(currentGI.getInSlotParameters());
			if (currentGI.getInSlotMask()!=null) allSlots.add(currentGI.getInSlotMask());
			for (InSlot slot : allSlots) {
				if (slot.getConnectedSlot() != null){
					Link currentL = slot.getLink();
					double dist1 = currentL.distance(x,y);
					if (dist1<=dist) {
						closest = currentL;
						dist=dist1;
					}
				}
			}
		}
		return closest;
	}

	///// 
	///// I N P U T   /   O U T P U T
	/////

	/**
	 * Determines if the file has never been saved
	 * i.e., it has no name.
	 * 
	 * @return true if the file wasn't previously saved.
	 */
	private final boolean firstSave( ) {
		return _filename == null;
	}

	/**
	 * Sets the new filename (and the new name) of the graph.
	 * 
	 * @param filename the complete path of the new name.
	 */
	private final void setFileName( File filename ) {
		if (!filename.equals(_filename)) {
			_filename = filename;
			_name=getAbsoluteName(filename.getName());
			renameFrame();
		}
		setModified(false);
	}

	/**
	 * Returns the current name of the file.
	 * 
	 * @return a filename or null.
	 */
	public final File getFileName( ) {
		return _filename;
	}

	/**
	 * Saves the graph in the current file name. If the graph
	 * is not already associated with a file, use the saveAs()
	 * function to choose the file name.
	 */
	public void save( ) {
		if (firstSave()) {
			saveAs();
		} else {
			saveGraph(getFileName());
		}
	}

	/**
	 * Saves the graph in a new file. The directory is given as
	 * the preferred directory where to store graphs, but the user
	 * can select an other directory.
	 */
	public void saveAs( ) {
		String currentDir = Workspace.getCurrentDir(); 
		File currentFileName = new File(currentDir);

		JFileChooser fileChooser = new JFileChooser(currentFileName);
		fileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
		fileChooser.setDialogTitle(Ariane.lang.getString("FileChooser.titleSaveGraph"));
		fileChooser.setFileFilter( new javax.swing.filechooser.FileFilter() {
			@Override
			public boolean accept( File F ) {
				String Fname = F.getName();
				return(Fname.endsWith(Ariane.DEFAULT_FILE_EXT) || F.isDirectory() );
			}
			@Override
			public String getDescription( ) {
				return "Ariane (*"+Ariane.DEFAULT_FILE_EXT+")";
			}
		});

		if (fileChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION){
			File theFile = fileChooser.getSelectedFile();
			if (theFile != null){
				String fullFileName = fileChooser.getSelectedFile().getPath();
				if (!fullFileName.endsWith(Ariane.DEFAULT_FILE_EXT)){
					fullFileName += Ariane.DEFAULT_FILE_EXT;
				}
				theFile = new File(fullFileName);
				// Overwrite the chosen file?
				if (theFile.exists()) {
					if (JOptionPane.showConfirmDialog(Ariane.workspace,
							Ariane.lang.getString("Workspace.dialog.message.part1")
							+theFile.getName()
							+Ariane.lang.getString("Workspace.dialog.message.part2"),
							Ariane.lang.getString("Workspace.dialog.title"),
							JOptionPane.YES_NO_OPTION)==JOptionPane.NO_OPTION) {
						return;
					}
				}
				saveGraph(theFile);
				currentDir = fileChooser.getSelectedFile().getParent();
				Ariane.workspace.addRecentProjects(theFile.getAbsolutePath());
				Workspace.setCurrentDir(currentDir);
			}
		}
	}

	/**
	 * Saves the current graph in the specified file.
	 * 
	 * @param f a regular XML file.
	 */
	private void saveGraph( File f ) {
		try {
			XMLFileOutputStream out = new XMLFileOutputStream(f,null);
			out.write(getXML(f));
			out.close();
			// Everything is OK.
			setFileName(f);
			Ariane.workspace.printSystemMessage(getFileName( ) + Ariane.lang.getString("Workspace.saving.success"));
		} catch (Exception ex) {
			//			ex.printStackTrace();
			Ariane.workspace.printErrorMessage(Ariane.lang.getString("Workspace.saving.error1")+getName()+" "+ex.getLocalizedMessage());
		}
	}

	/**
	 * Saves the current graph in a related XML tree.
	 * 
	 * @param file the current filename of the project.
	 * @return the resulted XML tree.
	 * @throws IOException in case of IO errors.
	 */
	public final XMLElement getXML( File file ) throws IOException {
		XMLElement e=null;
		if (_isRoot){
			e = new XMLElement("Graph");
		} else {
			e = new XMLElement("SubGraph");
		}
		// Graph contents (or sub-graph contents).
		XMLElement items = new XMLElement("ItemList");
		for (int i = 0; i < _graphItems.size(); i++){
			items.addChildren(_graphItems.get(i).save(file));
		}
		XMLElement connections = new XMLElement("ConnectionList");
		for (int i = 0; i < _graphItems.size(); i++){
			connections.addChildren(_graphItems.get(i).saveSlots());
		}
		e.addChildren(items);
		e.addChildren(connections);
		return e;
	}

	/**
	 * Loads a graph from an XML tree.
	 * 
	 * @param xmlTree the XML tree of all graph items.
	 * @throws IOException in case of IO errors.
	 */
	public final void setXML( LinkedList<XMLElement> xmlTree ) throws IOException {
		boolean found=false;
		if (_isRoot) {
			xmlTree = xmlTree.getFirst().getChildren();
			if (xmlTree.getFirst().getName().equalsIgnoreCase("Graph") ||
					xmlTree.getFirst().getName().equalsIgnoreCase("Graphe")) {
				xmlTree = xmlTree.getFirst().getChildren();
				found = true;
			}
		} else {
			for (int i=0; i< xmlTree.size(); i++) {
				if (xmlTree.get(i).getName().equalsIgnoreCase("SubGraph") ||
						xmlTree.get(i).getName().equalsIgnoreCase("SousGraphe")) {
					xmlTree = xmlTree.get(i).getChildren();
					found = true;
					break;
				}
			}
		}

		// Error incorrect file?
		if (!found)
			return;

		Iterator<XMLElement> it = xmlTree.iterator();
		while (it.hasNext()) {
			XMLElement e = it.next();
			if (e.getName().equalsIgnoreCase("ItemList") || e.getName().equalsIgnoreCase("ListeElements")) {
				GraphItem.load(this,e.getChildren(),false);
			} else if (e.getName().equalsIgnoreCase("ConnectionList") || e.getName().equalsIgnoreCase("ListeConnexions")) {
				Slot.load(this,e.getChildren());
			}
		}
		_undoRedoManager.cancelCurrentActionEdition(); // reset the undo manager.
	}

	///// 
	///// A U T O S C R O L L I N G
	/////

	/** The width of the border where the autoscrolling is sensitive. */
	private static int _autoscrollMargin = 45;

	/**
	 * Returns the Insets describing the autoscrolling region
	 * or border relative to the geometry of the implementing Component.
	 * 
	 * @return the inset region.
	 */
	@Override
	public final Insets getAutoscrollInsets( ) {
		Dimension size = getSize();
		Rectangle rect = getVisibleRect();

		// Notice the right and bottom suffers from the
		// fact that autoscroll use the top left item coordinates
		// to decide to scroll. So the top left insets and bottom right insets are uneven.
		return new Insets(rect.y + _autoscrollMargin, rect.x + _autoscrollMargin,
				size.height - rect.y - rect.height + _autoscrollMargin,
				size.width - rect.x - rect.width + _autoscrollMargin);
	}

	/**
	 * Notifies the panel to autoscroll according to the mouse coordinates.
	 * 
	 * @param location the mouse coordinates.
	 */
	@Override
	public final void autoscroll( Point location ) {
		int top = 0, left = 0, bottom = 0, right = 0;
		Dimension size = getSize();
		Rectangle rect = getVisibleRect();
		int bottomEdge = rect.y + rect.height;
		int rightEdge = rect.x + rect.width;

		if (location.y-rect.y <= _autoscrollMargin && rect.y > 0)
			top = _autoscrollMargin;
		if (location.x-rect.x <= _autoscrollMargin && rect.x > 0)
			left = _autoscrollMargin;
		if (bottomEdge-location.y <= _autoscrollMargin && bottomEdge < size.height)
			bottom = _autoscrollMargin;
		if (rightEdge-location.x <= _autoscrollMargin && rightEdge < size.width)
			right = _autoscrollMargin;
		rect.x += right - left;
		rect.y += bottom - top;
		location.translate(right - left, bottom - top);
		scrollRectToVisible(rect);
	}

	///// 
	///// U N D O / R E D O (H I S T O R Y)
	/////
	///// Redirects each event to the UndoRedoManager.

	/** The edit manager. */
	private final UndoRedoManager _undoRedoManager = new UndoRedoManager();

	/**
	 * Validates the given modification, it means adds
	 * the given modification in the current undo/redo history.
	 * 
	 * @param action the change to be added.
	 */
	public final void validateAction( UndoableEdit action ) {
		_undoRedoManager.validateAction(action);
	}

	/**
	 * Starts a new list of actions.
	 * An action is composed of several modifications.
	 */
	public final void startActionEdition( ) {
		_undoRedoManager.startActionEdition();
	}

	/**
	 * Ends the current list of actions.
	 */
	public final void endActionEdition( ) {
		_undoRedoManager.endActionEdition();
	}

	/**
	 * @return true if action undoing is possible.
	 */
	public boolean canUndo() {
		return _undoRedoManager.canUndo();
	}

	/**
	 * @return true if action redoing is possible.
	 */
	public boolean canRedo() {
		return _undoRedoManager.canRedo();
	}

	/**
	 * Undoes the last do action.
	 * Removes the last undo item from the undo list
	 * and adds it in the redo list.
	 */
	public final void undo( ) {
		_undoRedoManager.undo();
	}

	/**
	 * Redoes the last undo action.
	 * Removes the last redo item from the redo list
	 * and adds it in the undo list.
	 */
	public final void redo( ) {
		_undoRedoManager.redo();
	}

	/**
	 * UndoRedoManager manages a list of UndoableEdits, providing a way
	 * to undo or redo the appropriate actions.
	 */
	private class UndoRedoManager extends UndoManager {
		/** The current compound edit to store several edit in a same one. */
		private CompoundEdit _edit;

		/**
		 * Creates a new UndoRedoManager.
		 */
		public UndoRedoManager() {
			setLimit(PreferenceChooser.getInt("UndoHistorySize"));
		}

		/**
		 * Validates the given modification, it means add
		 * the given modification in the current undo/redo history.
		 * 
		 * @param action the change to be added.
		 */
		final void validateAction( UndoableEdit action ) {
			if (_edit == null) {
				_edit = new CompoundEdit();
			}
			_edit.addEdit(action);
		}

		/**
		 * Starts a new list of actions.
		 * An action is composed of several modifications.
		 */
		public final void startActionEdition( ) {
			_edit = new CompoundEdit();
		}

		/**
		 * Ends the current list of actions.
		 */
		public final void endActionEdition( ) {
			if (_edit != null && _edit.isSignificant()) {
				_edit.end();
				addEdit(_edit);
			}
			setModified(true);
			Ariane.workspace.update();
		}

		/**
		 * Cancels the current edited action.
		 */
		void cancelCurrentActionEdition() {
			_edit=null;
		}

		/**
		 * Undoes the last do action.
		 * Removes the last undo item from the undo list
		 * and adds it in the redo list. 
		 */
		@Override
		public final void undo( ) {
			try {
				super.undo();
				Ariane.workspace.update();
			} catch (javax.swing.undo.CannotUndoException e) {
				Ariane.workspace.printErrorMessage(Ariane.lang.getString("Graph.history.error1"));
			}
		}

		/**
		 * Redoes the last undo action.
		 * Removes the last redo item from the redo list
		 * and adds it in the undo list.
		 */
		@Override
		public final void redo( ) {
			try {
				super.redo();
				Ariane.workspace.update();
			} catch (javax.swing.undo.CannotRedoException e) {
				Ariane.workspace.printErrorMessage(Ariane.lang.getString("Graph.history.error1"));
			}
		}
	}

	/**
	 * Stores dragging information.
	 */
	class DraggingSelection {
		/** The initial selection rectangle. */
		final private Rectangle _selectionRectangle;
		/** The shift between the mouse pointer and the origin of the selection rectangle. */
		private Point _offset;
		/** the current location of the selection rectangle. */
		private Point _location;

		/**
		 * Creates a new dragging event.
		 * 
		 * @param rectangle the selection rectangle.
		 * @param xycoord the mouse pointer.
		 */
		public DraggingSelection (Rectangle rectangle, Point xycoord ) {
			_selectionRectangle=rectangle;
			_offset = new Point(rectangle.x-xycoord.x,rectangle.y-xycoord.y);
			_location = rectangle.getLocation();
		}

		/**
		 * @return the selection rectangle location;
		 */
		public final Point getOriginLocation() {
			return _selectionRectangle.getLocation();
		}

		/**
		 * @return the offset between the origin of the selection rectangle
		 *          and the mouse pointer.
		 */
		public final Point getOffset() {
			return _offset;
		}

		/**
		 * @return the current location of the selection rectangle.
		 */
		public final Point getLocation() {
			return _location;
		}

		/**
		 * Sets the new location.
		 * 
		 * @param location the new selection rectangle location.
		 */
		public final void setLocation( Point location ) {
			_location=location;
		}

		/**
		 * Returns the translation between the old location 
		 * and the new location of the selection rectangle. 
		 * 
		 * @param newOrigin the new selection rectangle origin.
		 * @return the translation vector.
		 */
		public final Point getTranslation( Point newOrigin )  {
			Point translation = new Point();
			translation.x=_location.x;
			translation.y=_location.y;
			_location.x = newOrigin.x;
			_location.y = newOrigin.y;
			translation.x -= newOrigin.x;
			translation.y -= newOrigin.y;
			return translation;
		}
	}

	/**
	 * The rubber band box is used to select components in the graph.
	 */
	class RubberBandBox {
		private int _startX, _startY;
		private Rectangle _currentBox= new Rectangle();
		private final Color _shadowColor =new Color(100, 149, 237);

		public RubberBandBox( int x, int y ) {
			_startX = x;
			_startY = y;
		}

		/**
		 * Draws the current selection rectangle.
		 */
		private final void erase( ) {
			repaint();
		}

		/**
		 * Draws the current selection rectangle.
		 * 
		 * @param ghostgraphics the graphic context.
		 */
		private final void erase( Graphics ghostGraphics ) {
			Graphics2D g2d = (Graphics2D)ghostGraphics;
			Composite originalComposite = g2d.getComposite();
			g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f));
			g2d.setPaint(_shadowColor);
			ghostGraphics.fillRect(_currentBox.x,_currentBox.y,_currentBox.width,_currentBox.height);
			g2d.setComposite(originalComposite);
		}

		public final boolean contains ( Rectangle r ) {
			return _currentBox.contains(r);
		}

		public void update( Point xycoord ) {
			int w = Math.abs(_startX-xycoord.x);
			int h = Math.abs(_startY-xycoord.y);
			int x = Math.min(_startX, xycoord.x);
			int y = Math.min(_startY, xycoord.y);

			if (_currentBox.width != w || _currentBox.height != h) {
				repaint();
				_currentBox.x=x;
				_currentBox.y=y;
				_currentBox.width=w;
				_currentBox.height=h;
				// If needed : autoscroll.
				Rectangle r = new Rectangle(xycoord.x-_autoscrollMargin, xycoord.y-_autoscrollMargin,1,1);
				scrollRectToVisible(r);
				repaint();
			}
		}
	}
}