/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph;

/**
 * A <code>XYConstraint</code> defines a constraint that can be
 * specified for locating a component in a XYLayout.
 * A constraints can be either:
 * <ul>
 * <li>a bounds: x,y,height,width
 * <li>an alignment: GridBagConstraints.EAST WEST, NORHT SOUTH.
 * </ul>
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 24, 2004
 */
public final class XYConstraint implements Cloneable {
	/** The alignment constraints. */
	private int _alignement;
	/** The rank in the same alignment. */
	private int _rank;
	/** The x constraint. */
	private int _x;
	/** The y constraint. */
	private int _y;
	/** The width constraint. */
	private int _width;
	/** The height constraint. */
	private int _height;

	/**
	 * Creates an empty constraint.
	 */
	public XYConstraint( ) {
		this(0, 0, 0, 0);
	}

	/**
	 * Creates a bounds constraint with the
	 * specified values.
	 * 
	 * @param x the x constraint.
	 * @param y the y constraint.
	 * @param width the width constraint.
	 * @param height the height constraint.
	 */
	public XYConstraint( int x, int y, int width, int height ) {
		_rank = -1;
		_x = x;
		_y = y;
		_width = width;
		_height = height;
	}

	/**
	 * Creates an alignment region constraint with the
	 * specified values.
	 * 
	 * @param alignment type of alignment between
	 *        GridBagConstraints.{EAST WEST, NORHT SOUTH}.
	 * @param rank the rank in the same alignment (>=1).
	 */
	public XYConstraint( int alignment, int rank ) {
		_alignement = alignment;
		_rank = rank;
	}

	/**
	 * Returns the rank part of the constraint.
	 * 
	 * @return the rank of the slot.
	 */
	public final int getRank( ) { return _rank; }

	/**
	 * Returns the alignment part of the constraint.
	 * 
	 * @return the alignment.
	 */
	public final int getAlignement( ) { return _alignement; }

	/**
	 * Returns the x part of the constraint.
	 * 
	 * @return the x coordinate.
	 */
	public final int getX( ) { return _x; }

	/**
	 * Returns the y part of the constraint.
	 * 
	 * @return the y coordinate.
	 */
	public final int getY( ) { return _y; }

	/**
	 * Returns the width part of the constraint.
	 * 
	 * @return the width.
	 */
	public final int  getWidth( ) { return _width; }

	/**
	 * Returns the height part of the constraint.
	 * 
	 * @return the height.
	 */
	public final int getHeight( ) { return _height; }

	/**
	 * Defines equality between two constraints.
	 * 
	 * @param that a constraints.
	 */
	@Override
	public final boolean equals( Object that ) {
		if (that instanceof XYConstraint) {
			XYConstraint other = (XYConstraint)that;
			return other._x == _x && other._y == _y && other._width == _width && other._height == _height;
		}
		return false;
	}

	/**
	 * Clones a constraints.
	 * 
	 * @return the new XYConstraints.
	 */
	@Override
	public final Object clone( ) {
		return new XYConstraint(_x, _y, _width, _height);
	}
}