/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotMask;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotResult;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>ControlStructureForItem</code> implements 
 * the iterative control structure by a graph item.
 * The sequence of Items to has be iterated is inclosed
 * inside a subgraph.
 * The loop is made from the value of the parameter $0 until
 * the value of the parameter $1 by step of the value of the parameter $2.
 *
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD -Jul 13, 2010 (display loop variable during execution)
 */
public final class ControlStructureForItem extends CompositeItem {
	/** The tag name of the item. */
	public static final String TAGNAME="For";

	/**
	 * Creates a new <code>ControlStructureFor</code> on the specified
	 * graph with the given attributes values.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the name of the item.
	 * @param identifier the identification number (or -1)
	 */
	public ControlStructureForItem( TreeItem treeItem, Graph graph, String name, int identifier ) {
		super(treeItem, graph, identifier, name );
		askProperties(null,1,1,1);
		// used to display the current loop value
		setSignature(_treeItem.getSignature());
		setComment(PreferenceChooser.EMPTY_STRING);
	}

	/**
	 * Creates a <code>ControlStructureForItem</code> on the specified
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ControlStructureForItem( Graph graph, LinkedList<XMLElement> xmlTree ) throws IOException {
		super(graph,-1);
		load(xmlTree);
		// used to display the current loop value
		setComment(PreferenceChooser.EMPTY_STRING);
		if (_treeItem==null) {
			_treeItem = restoreDefaultTreeItem();
		}
		setSignature(_treeItem.getSignature());
		draw();
	}

	/**
	 * Creates a <code>ControlStructureForItem</code>
	 * as a subroutine (undeditable item) on the specified
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ControlStructureForItem( Graph graph, LinkedList<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
		super(graph,-1,treeItem);
		load(xmlTree);
		// used to display the current loop value
		setComment(PreferenceChooser.EMPTY_STRING);
		setSignature(_treeItem.getSignature());
		setAbstractName(treeItem.getName());
		setNickName(treeItem.getName());
		draw();
	}

	@Override
	protected TreeItem restoreDefaultTreeItem() {
		setNickName(getNickName());
		setAbstractName("for");
		return Ariane.workspace.getTree().getTreeItem("for",null);
	}

	/**
	 * Builds the prototype of the item from the given values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected final void setPrototype( String name, int iIm, int iP, int oIm ) {
		setPrototype(name, iIm, 3+iP, oIm, iIm, 1);
	}

	/**
	 * Creates the subgraph with the convenient prototype.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 * @return the subgraph.
	 */
	@Override
	protected SubGraph createSubGraph( String name, int iIm, int iP, int oIm ) {
		return createSubGraph(getNickName(), iIm, iP-2, oIm, iIm, 0);
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String tagName( ) { return TAGNAME; }

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Resets the item and erases the comment inside the item glyph.
	 */
	@Override
	protected final void reset() {
		super.reset();
		setComment(PreferenceChooser.EMPTY_STRING);
	}

	/**
	 * Executes the iteration.
	 * The loop is made from the parameter value $0
	 * to the parameter value $1
	 * by step of the parameter value $2.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the for has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			VirtualItem virtualOperator = getVirtualOperator();

			// Get and check the 3 parameter values used in the conditional expression
			int loop[] = new int[3];
			for (int i=0;i<3;i++) {
				InSlotParameter params = getInSlotParameters().get(i);
				if (params.hasNumericalValue(imperative)) {
					loop[i]=(int)params.getNumericalValue();
				} else {
					throw new IllegalArgumentException(Ariane.lang.getString("Operator.error4")+params.getRank());
				}
			}

			// get the rest of the parameters values
			for (int i=3; i<inSlotParameters.size(); i++) {
				InSlotParameter params = getInSlotParameters().get(i);
				OutSlotResult paramd = virtualOperator.getVirtualInSlotParameters().get(i-2);
				if (params.hasValue(imperative)) {
					paramd.setValue(params.getValue());
				} else {
					throw new IllegalArgumentException(Ariane.lang.getString("Operator.error1")+params.getRank());
				}
			}

			// Get the first value of the inner parameter
			OutSlotResult paramd = virtualOperator.getVirtualInSlotParameters().getFirst();

			// Get and check mask image (-> execute previous operators if needed).
			// Note: Mask is optional!
			InSlotMask inMasks = getInSlotMask();
			OutSlotImage inMaskd = (OutSlotImage)virtualOperator.getVirtualInSlotMask();
			if (inMasks.isConnected() && inMasks.hasValue(false)) {
				inMaskd.setValue(inMasks.getValue());
			}

			// Gets and checks input images.
			// (Sets the input image of the virtual operator).
			Iterator<InSlotImage> its1 = getInSlotImages().iterator();
			Iterator<OutSlotImage> itd1 = virtualOperator.getVirtualInSlotImages().iterator();
			while (its1.hasNext()) {
				InSlotImage inImgs = its1.next();
				OutSlotImage inImgd = itd1.next();
				if (inImgs.hasValue(imperative)) {
					inImgd.setValue(inImgs.getValue());
				} else {
					throw new IllegalArgumentException(Ariane.lang.getString("Operator.error3")+inImgs.getRank());
				}
			}

			changeStatus(Status.RUNNING);
			// Iteration
			Ariane.workspace.printSystemMessage("> For loop from "+loop[0]+" to "+ loop[1] + " step "+ loop[2]+" do");
			for (int loopv=loop[0]; loopv<=loop[1] && !Graph.isExecutionStopped(); loopv+=loop[2]) {
				reset();
				changeStatus(Status.RUNNING);
				Ariane.workspace.printSystemMessage("> loop indice "+loopv);

				// Set the value of the inner parameter with the current loop value.
				paramd.setValue(""+loopv);
				setComment(""+loopv);

				// The execution of the while body is split in 2 steps
				// to avoid problems when two operators use the same entry.
				
				// 1-step: Execute the graph by trying to get the intermediate images
				for (int k=0; k<getInSlotImages().size(); k++) {
					InSlotImage outImgi = virtualOperator.getVirtualInterSlotImage(k);
					outImgi.hasValue(imperative);
				}

				// 2- set the result to each the intermediate images for a new loop
				for (int k=0; k<getInSlotImages().size(); k++) {
					InSlotImage outImgi = virtualOperator.getVirtualInterSlotImage(k);
					OutSlotImage outImgd = virtualOperator.getVirtualInSlotImage(k);
					if (outImgi.hasValue(imperative)) {
						DataFile.copy(outImgi.getValue(),outImgd.getTempName());
						outImgd.setValue();
					}
				}
			}

			Ariane.workspace.printSystemMessage("> end loop for");

			// Sets the new output images.
			Iterator<OutSlotImage> its2 = getOutSlotImages().iterator();
			Iterator<InSlotImage>itd2 = virtualOperator.getVirtualOutSlotImages().iterator();
			while (its2.hasNext()) {
				OutSlotImage outImgs = its2.next();
				InSlotImage outImgd = itd2.next();
				if (outImgd.hasValue(imperative)) {
					outImgs.setValue(outImgd.getValue());
				}
			}

			OutSlotResult results = getOutSlotResult();
			InSlotParameter resultd = (InSlotParameter)virtualOperator.getVirtualOutSlotResult();
			if (!resultd.hasValue(imperative)) {
				throw new IllegalArgumentException(Ariane.lang.getString("MacroOperator.error2"));
			}
			results.setValue(resultd.getValue());
			changeStatus(Status.EXECUTED);
			return true;
		} catch(IllegalArgumentException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				changeStatus(Status.ERROR);
				getGraph().stopExecution();
			}
		} catch(Exception e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				changeStatus(Status.ERROR);
				getGraph().stopExecution();
			}
		}
		return false;
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Asks interactively the prototype of the structure.
	 */
	@Override
	protected final void askProperties( ){
		askProperties(null,getInSlotImages().size(),0,getOutSlotImages().size());
	}
}
