/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import java.io.IOException;
import java.util.LinkedList;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>ControlStrutureMergeItem</code> merges two inputs
 * slots into a unique output slot.
 * The result is chosen as the first validated input.
 *
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 */
public class ControlStructureMergeItem extends ControlStructure {
	/** The tag name of the item. */
	public static final String TAGNAME="Merge";

	/**
	 * Creates <code>ControlStructureMergeItem</code>
	 * with the specified attributes.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the item name.
	 * @param identifier the identification number (or -1).
	 */
	public ControlStructureMergeItem( TreeItem treeItem, Graph graph, String name, int identifier ) {
		super(treeItem, graph, identifier,name);
		// 2 inputs.
		for (int i=0; i<2; i++) inSlotImages.add(new InSlotImage(this,i));
		// 1 output
		outSlotImages.add(new OutSlotImage(this, 0));
		draw();
	}

	/**
	 * Creates a <code>ControlStructureMergeItem</code>
	 * from the given XML data.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ControlStructureMergeItem( Graph graph, LinkedList<XMLElement> xmlTree ) throws IOException {
		super(graph, -1);
		load(xmlTree);
		draw();
	}

	/**   
	 * Builds or changes the prototype of the control structure
	 * from the specified values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected final void setPrototype( String name, int iIm, int iP, int oIm ) {}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public String tagName() { return TAGNAME; }

	///// 
	/////  E X E C U T I O N 
	/////

	/**
	 * Executes a merge operation, it means looks at each
	 * of the two inputs and returns the first validated result.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			for (int i=0; i<2; i++) {
				InSlotImage isi = inSlotImages.get(i);
				if (isi.hasValue(false)) {
					(outSlotImages.getFirst()).setValue(isi.getValue());
					changeStatus(Status.RUNNING);
					getOutSlotResult().setValue(Float.toString(i));
					changeStatus(Status.EXECUTED);
					return true;
				}
			}
			// No input validated -> try to find the reason for that...
			// so reexecute the two paths with imperative tests.
			for (int i=0; i<2; i++) {
				InSlotImage isi = inSlotImages.get(i);
				if (isi.hasValue(true)) {
					(outSlotImages.getFirst()).setValue(isi.getValue());
					changeStatus(Status.RUNNING);
					getOutSlotResult().setValue(Float.toString(i));
					changeStatus(Status.EXECUTED);
					return false;
				}
			}
			throw new IOException(Ariane.lang.getString("controlstructuremerge.error1"));
		} catch(IOException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				changeStatus(Status.ERROR);
				getGraph().stopExecution();
			}
		}
		return false;
	}
}