/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;
import java.io.IOException;
import java.util.LinkedList;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>ControlStructureSelectItem</code> implements the a
 * selector of multiple inputs. The selected input
 * is given by the parameter value.
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 */
public class ControlStructureSelectItem extends ControlStructure {
	/** The tag name of the item. */
	public static String TAGNAME="Select";

	/**
	 * Creates a <code>ControlStructureSelectItem</code>
	 * with the given attributes.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the item name.
	 * @param identifier the identification number (or -1).
	 */
	public ControlStructureSelectItem( TreeItem treeItem, Graph graph, String name, int identifier ) {
		super(treeItem, graph,identifier,name);
		askProperties(null,1,0,0);
		setSignature(treeItem.getSignature());
	}

	/**
	 * Creates a <code>ControlStructureSelectItem</code> on the given
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public ControlStructureSelectItem( Graph graph, LinkedList<XMLElement> xmlTree ) throws IOException {
		super(graph, -1);
		load(xmlTree);
		draw();
	}

	/**   
	 * Builds or changes the prototype of the control structure
	 * from the given values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected final void setPrototype( String name, int iIm, int iP, int oIm ) {
		setAbstractName(name);
		for (int i=0; i<iIm; i++) inSlotImages.add(new InSlotImage(this,i));
		outSlotImages.add(new OutSlotImage(this, 0));
		inSlotParameters.add(new InSlotParameter(this,0));
		draw();
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String tagName( ) {
		return TAGNAME;
	}

	///// 
	/////  E X E C U T I O N 
	/////

	/**
	 * Executes a <i>Select</i>, it means select
	 * the input rank indicated hy
	 * the parameter value.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			// Gets and checks the parameter (-> executes forward operators if needed).
			int inputRank;
			InSlotParameter param = inSlotParameters.getFirst();
			if (param.hasNumericalValue(imperative)) {
				inputRank = (int)param.getNumericalValue();
			} else {
				throw new IllegalArgumentException(Ariane.lang.getString("Operator.error4")+param.getRank());
			}

			// Checks consistency of the parameter.
			if (inputRank<0 || inputRank >= inSlotImages.size()) {
				throw new ArithmeticException(Ariane.lang.getString("controlstructure.error1")); 
			}

			// Gets and checks the input image (-> executes forward operators if needed).
			InSlotImage isi = (inSlotImages.get(inputRank));
			if (isi != null && (isi.hasValue(imperative))) {
				// Validate the output.
				(outSlotImages.getFirst()).setValue(isi.getValue());
				getOutSlotResult().setValue(Float.toString(inputRank));
			}else {
				throw new IllegalArgumentException(Ariane.lang.getString("Operator.error3")+isi.getRank());
			}
			changeStatus(Status.EXECUTED);
			return true;
		} catch (IllegalArgumentException e){
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				changeStatus(Status.ERROR);
				getGraph().stopExecution();
			}
		} catch (Exception e){
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				changeStatus(Status.ERROR);
				getGraph().stopExecution();
			}
		}
		return false;
	}

	///// 
	/////  P R O P E R T I E S
	/////

	/**   
	 * Asks interactively the prototype of the structure.
	 */
	protected final void askProperties(){
		askProperties(null,1,0,0);
	}
}