/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.item.etc.LogicalExpression;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotMask;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotResult;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * A <code>ControlStructureWhileItem</code> implements the
 * repetitive structure control by a graph item.
 * The sequence of Items to has be repeated is inclosed
 * inside a subgraph.
 * 
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001 
 * @version Régis CLOUARD - Jun 11, 2010 (Add several optional parameters)
 * @version Régis CLOUARD - Mar 24, 2011 (fix bug on execution)
 */
public final class ControlStructureWhileItem extends CompositeItem {
	/** The tag name of the item. */
	public static final String TAGNAME="While";

	/** The boolean expression of the conditional test. */
	private LogicalExpression _condition=null;

	/**
	 * Creates a new <code>ControlStructureWhileItem</code> on the specified
	 * graph with the given attributes.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param name the default name of the component.
	 * @param identifier the identification number (or -1).
	 */
	public ControlStructureWhileItem( TreeItem treeItem, Graph graph, String name, int identifier) {
		super(treeItem, graph, identifier, name );
		askProperties(null,1,1,1);
		// used to display the current loop value
		setSignature(_treeItem.getSignature());
		setComment(PreferenceChooser.EMPTY_STRING);
	}

	/**
	 * Creates a <code>ControlStructureWhileItem</code> on the specified
	 * graph from the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
 */
	public ControlStructureWhileItem( Graph graph, LinkedList<XMLElement> xmlTree ) throws IOException {
		super(graph,-1);
		load(xmlTree);
		if (getFormula() != null) {
			_condition= new LogicalExpression(getFormula());
		} else {
			if (getBinary() != null){ // Keep for sake of compatibility
				_condition= new LogicalExpression(getBinary());
				setFormula(getBinary());
				setBinary(null);
			}
			setFormula(""); // IMPERATIVE to signify that it uses formula.
		}
		// used to display the current loop value
		setComment(PreferenceChooser.EMPTY_STRING);
		if (_treeItem==null) {
			_treeItem = restoreDefaultTreeItem();
		}
		setSignature(_treeItem.getSignature());
		draw();
	}

	/**
	 * Creates a <code>ControlStructureWhileItem</code> 
	 * as a subroutine (undeditable item) on the specified
	 * graph from the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
 */
	public ControlStructureWhileItem( Graph graph, LinkedList<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
		super(graph,-1, treeItem);
		load(xmlTree);
		if (getFormula() != null) {
			_condition= new LogicalExpression(getFormula());
		} else {
			if (getBinary() != null){ // Keep for sake of compatibility
				_condition= new LogicalExpression(getBinary());
				setFormula(getBinary());
				setBinary(null);
			}
			setFormula(""); // IMPERATIVE to signify that it uses formula.
		}
		// used to display the current loop value
		setComment(PreferenceChooser.EMPTY_STRING);
		setSignature(_treeItem.getSignature());
		setAbstractName(treeItem.getName());
		setNickName(treeItem.getName());
		draw();
	}

	@Override
	protected TreeItem restoreDefaultTreeItem() {
		setNickName(getNickName());
		setAbstractName("while");
		return Ariane.workspace.getTree().getTreeItem("while",null);
	}

	/**
	 * Builds the prototype of the item from the given values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected final void setPrototype( String name, int iIm, int iP, int oIm ) {
		super.setPrototype(name, iIm, 2+iP, oIm, iIm, 2);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		draw();
	}

	/**
	 * Creates the subgraph with the convenient prototype.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 * @return the subgraph.
	 */
	@Override
	protected final SubGraph createSubGraph( String name, int iIm, int iP, int oIm ) {
		return createSubGraph(getNickName(), iIm, iP, oIm, iIm, 2);
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String tagName( ){
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes the repetition while the related condition is true.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			if (getFormula().length()==0) {
				throw new ArithmeticException(Ariane.lang.getString("arithmetic.error0"));
			}

			VirtualItem virtualOperator = getVirtualOperator();
			// Get the 2 first parameters used in the conditional expression
			double[] val = new double[2];
			for (int i=0;i<2;i++) {
				InSlotParameter params = getInSlotParameters().get(i);
				if (params.hasNumericalValue(imperative)) {
					val[i]=(int)params.getNumericalValue();
				} else {
					throw new IllegalArgumentException(Ariane.lang.getString("Operator.error4")+ params.getRank());
				}
			}

			// Get the rest of the parameters
			for (int i=2;i<inSlotParameters.size();i++) {
				InSlotParameter params = getInSlotParameters().get(i);
				OutSlotResult paramd = virtualOperator.getVirtualInSlotParameters().get(i);
				if (params.hasValue(imperative)) {
					paramd.setValue(params.getValue());
				} else {
					throw new IllegalArgumentException(Ariane.lang.getString("Operator.error1")+params.getRank());
				}
			}

			// Get and check the mask image (-> execute previous operators if needed)
			// Note: Mask is optional
			InSlotMask inMasks = getInSlotMask();
			OutSlotImage inMaskd = (OutSlotImage)virtualOperator.getVirtualInSlotMask();
			if (inMasks.isConnected() && inMasks.hasValue(false)) {
				inMaskd.setValue(inMasks.getValue());
			}

			// Gets and check the input images
			Iterator<InSlotImage> its=getInSlotImages().iterator();
			Iterator<OutSlotImage> itd=virtualOperator.getVirtualInSlotImages().iterator();
			while (its.hasNext()) {
				InSlotImage inImgs = its.next();
				OutSlotImage inImgd = itd.next();
				if (inImgs.hasValue(imperative)) {
					inImgd.setValue(inImgs.getValue());
				} else {
					throw new IllegalArgumentException(Ariane.lang.getString("Operator.error3")+inImgs.getRank());
				}
			}

			changeStatus(Status.RUNNING);

			// Repetition
			int iteration=0; // Tip: use to avoid infinite loop
			Ariane.workspace.printSystemMessage("> While loop ("+_condition.toHtml()+") do");
			while (_condition.evaluate(val[0],val[1]) && !Graph.isExecutionStopped()) {
				iteration++;
				Ariane.workspace.printSystemMessage(">cycle: "+iteration);
				setComment(""+iteration);

				// Set the value of the 2 inner parameters.
				for (int i=0; i<2;i++)
					virtualOperator.getVirtualInSlotParameter(i).setValue(Float.toString((float)val[i]));
				reset();
				changeStatus(Status.RUNNING);

				// The execution of the while body is split in 2 steps
				// to avoid problems when two operators use the same entry.
				
				// 1-step: Execute the graph by trying to get the intermediate images
				for (int k=0; k<getInSlotImages().size(); k++){
					InSlotImage outImgi = virtualOperator.getVirtualInterSlotImage(k);

					outImgi.hasValue(imperative);
				}

				// 2- set the result to each the intermediate images for a new loop
				for (int k=0; k<getInSlotImages().size(); k++) {
					InSlotImage outImgi = virtualOperator.getVirtualInterSlotImage(k);
					OutSlotImage outImgd = virtualOperator.getVirtualInSlotImage(k);

					if (outImgi.hasValue(imperative)) {
						DataFile.copy(outImgi.getValue(),outImgd.getTempName());
						outImgd.setValue();
					}
				}

				
				// Get the new parameters value.
				for (int i=0;i<2;i++) {
					InSlotParameter params = virtualOperator.getVirtualInterSlotResult(i);
					if (params.isConnected() && params.hasNumericalValue(imperative)) {
						val[i]=(int)params.getNumericalValue();
					}
				}
			}
			Ariane.workspace.printSystemMessage("end while loop");

			// Set the new output images.
			Iterator<OutSlotImage> its1 = getOutSlotImages().iterator();
			Iterator<InSlotImage> itd1 = virtualOperator.getVirtualOutSlotImages().iterator();
			while (its1.hasNext()) {
				OutSlotImage outImgs = its1.next();
				InSlotImage outImgd = itd1.next();
				if (outImgd.hasValue(imperative)) {
					outImgs.setValue(outImgd.getValue());
				}
			}

			OutSlotResult results = getOutSlotResult();
			InSlotParameter resultd = (InSlotParameter)virtualOperator.getVirtualOutSlotResult();
			if (!resultd.hasValue(imperative)) {
				throw new IllegalArgumentException(Ariane.lang.getString("MacroOperator.error2"));
			}
			results.setValue(resultd.getValue());
			changeStatus(Status.EXECUTED);
			return true;
		} catch(IllegalArgumentException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
		} catch(Exception e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
		}
		return false;
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Changes the current boolean expression with the given one.
	 * 
	 * @param condition the conditional expression.
	 * @return true if the formula is valid, false otherwise.
	 */
	@Override
	public final boolean changeFormula( String condition ) {
		if (getFormula() != null && getFormula().equals(condition))
			return true;
		try {
			_condition= new LogicalExpression(condition);
			setFormula(condition);
			return true;
		} catch(Exception ex) {
			Ariane.workspace.printErrorMessage(ex.getLocalizedMessage());
		}
		return false;
	}

	/**
	 * Asks interactively the prototype of the item.
	 */
	@Override
	protected final void askProperties( ) {
		askProperties(null,1,0,0);
	}
}