/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.etc.StringExpression;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.util.LinkedList;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;

/**
 * A <code>FormatItem</code> implements a calculator item.
 * 
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 19, 2004
 * @version Frédéric LAIR - Jun 28, 2007 (add more than two parameters)
 * @version Régis CLOUARD- Mar 30, 2011 (refine)
 */
public final class FormatItem extends ControlStructure {
	/** The tag name of the item. */
	public static String TAGNAME="Format";

	/** The arithmetic expression. */
	private StringExpression _expression=null;

	/**
	 * Creates a new <code>FormatItem</code> on the specified
	 * graph with the given attribute values.
	 * 
	 * @param treeItem the source tree item.
	 * @param graph the graph that holds the item.
	 * @param name the default name of the item.
	 * @param identifier the identification number (or -1).
	 */
	public FormatItem( TreeItem treeItem, Graph graph, String name, int identifier ) {
		super(treeItem, graph, identifier,name);
		askProperties(null,0,2,0);
		draw();
	}

	/**
	 * Creates a <code>FormatItem</code> on the specified
	 * graph with the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param elements the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public FormatItem( Graph graph, LinkedList<XMLElement> elements ) throws IOException {
		super(graph, -1);
		load(elements);
		// Compile formula if it exists.
		if (getFormula() != null) {
			_expression = new StringExpression(getFormula(), inSlotParameters.size()); 
		} else {
			if (getBinary() != null){ // Keep this for sake of compatibility
				_expression = new StringExpression(getBinary(), inSlotParameters.size());
				setFormula(getBinary());
				setBinary(null);
			} else {
				setFormula(""); // IMPERATIVE to signify that it uses formula.
			}
		}
		draw();
	}

	/**
	 * Builds the prototype of the control structure from the given values.
	 * 
	 * @param name the name of the item.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected final void setPrototype( String name, int iIm, int iP, int oIm ) {
		setAbstractName(name);
		setFormula(""); // IMPERATIVE to signify that it uses formula.
		for (int i=0; i < iP; i++) {
			inSlotParameters.add(new InSlotParameter(this,i));
		}
	}

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * Returns the nickname of the item if any
	 * otherwise returns the name.
	 * 
	 * @return the nickname.
	 */
	@Override
	public final String getNickName( ) {
		if (getFormula()==null || getFormula().length()==0) {
			return super.getNickName();
		} else {
			return ""+_expression;
		}
	}

	/**
	 * @return the tag name.
	 */
	@Override
	public final String tagName( ) {
		return TAGNAME;
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes the <code>FormatItem</code>, means
	 * compute a result from the arithmetic formula.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			if (getFormula().length()==0) {
				throw new MissingFormulaException(Ariane.lang.getString("arithmetic.error0"));
			}
			int nbParam = inSlotParameters.size();
			String[] val = new String[nbParam];

			for (int i=0; i<nbParam; i++) {
				InSlotParameter param = inSlotParameters.get(i);
				if (param.hasValue(imperative)){
					val[i] = param.getValue();
				} else {
					throw new IllegalArgumentException(Ariane.lang.getString("Operator.error1")+param.getRank());
				}
			}
			changeStatus(Status.RUNNING);

			String result = _expression.evaluate(val);
			getOutSlotResult().setValue(result);
			Ariane.workspace.printSystemMessage("format("+getFormula()+")="+result);
			changeStatus(Status.EXECUTED);
			return true;
		} catch (MissingFormulaException e){
//			e.printStackTrace();
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.MISSINGFORMULA, getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
		} catch (ArithmeticException e){
//			e.printStackTrace();
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.ARITHMETICS, getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
		}catch(IllegalArgumentException e){
//			e.printStackTrace();
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.MISSINGPARAMETER, getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
		}
		return false;
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected final JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem item;

		// Action: executer
		item = new JMenuItem(Ariane.lang.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				getGraph().execute();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				resetAll();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		// Edition: delete
		item = new JMenuItem(Ariane.lang.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e){
				delete();
			}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		return popupMenu;
	}

	/**
	 * Changes the arithmetic formula.
	 * 
	 * @param textFormula a formula.
	 */
	@Override
	public final boolean changeFormula( String textFormula ) {
		if ((getFormula()==null || !getFormula().equals(textFormula))) {
			try {
				_expression= new StringExpression(textFormula, inSlotParameters.size());
				setFormula(textFormula);
			} catch(Exception ex) {
				Ariane.workspace.printErrorMessage(ex.getLocalizedMessage());
				return false;
			}
		}
		return true;
	}
}