/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.command.ChangeFormulaEdit;
import fr.greyc.pantheon.ariane.graph.command.RenameItemEdit;
import fr.greyc.pantheon.ariane.graph.item.etc.ItemSignatureParser;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.graph.slot.InSlot;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotMask;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlot;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotResult;
import fr.greyc.pantheon.ariane.graph.slot.Slot;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.RoundedBorder;
import fr.greyc.pantheon.ariane.utils.RoundedPanel;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Point;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.MalformedParameterizedTypeException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPopupMenu;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.border.Border;
import javax.swing.border.EtchedBorder;

/**
 * A <code>GraphItem</code> defines an item that can be displayed in a graph.
 * It is graphically represented by a box with input and output slots.
 * @see OperatorItem
 * @see MacroOperatorItem
 * @see ControlStructure
 * 
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Dec 06, 2004
 * @version Frédéric LAIR - Jun 11, 2007
 * @version Régis CLOUARD - Jun 10, 2010 (add nickname) 
 */
public abstract class GraphItem extends RoundedPanel {
	/** The tree item that contains the properties of this kind of graph item. */
	protected TreeItem _treeItem;

	/**
	 * The various status of a graph item
	 */
	public enum Status {
		/** Unexecuted state. */
		UNEXECUTED,
		/** Running state. */
		RUNNING,
		/** Executed state. */
		EXECUTED,
		/** Error state.*/
		ERROR
	}

	/** The related graph that supports this item. */
	private Graph _graph = null;
	/** The identifier number; supposed to be unique. */
	private int _identifier;
	/** The identifier number read during a reading or a pasting. */
	private int _readIdentifier=-1;
	/** The current read or paste session number. */
	private int _currentReadPasteSession=0;

	/** The icon for unexecuted status. */
	private static ImageIcon _unexecutedStatusIcon;
	/** The icon for running status. */
	private static ImageIcon _runningStatusIcon;
	/** The icon for executed status. */
	private static ImageIcon _executedStatusIcon;
	/** The icon for error status. */
	private static ImageIcon _errorStatusIcon;

	private final static Color selectedShadowColor = new Color(255, 150, 150);
	/** The color for selected shadow. */
	private final static Color selectedHighlightColor = new Color(255, 0, 00);

	protected final static Border unselectedBorder =  new RoundedBorder(EtchedBorder.RAISED);
	private final static Border selectedBorder = new RoundedBorder(EtchedBorder.LOWERED, selectedHighlightColor, selectedShadowColor);

	/** The label that visualizes the status of the item. */
	private JLabel _labelStatus;

	/** The popup menu.*/
	private static JPopupMenu _popupMenu=null;

	/** The name of the item. */
	private String _name=null;

	/** The nickname of the item. */
	private String _nickName=null;

	/** The label that displays the name in the property viewer. */
	private JTextField _textFieldName;

	/** The label that displays the name on the item representation. */
	private JLabel _labelName;

	/** An optional comment that can be displayed inside the graph item. */
	private final JComponent _comment;

	/** The name of the executable command. Ex: "pbmp2pan". */
	private String _binary=null;

	/** The name of executable category. */
	private String _category=null;

	/** The arithmetic or logical expression: ex: $0<=$1. */
	private String _formula=null;

	/** The list of input image slots. */
	protected LinkedList<InSlotImage> inSlotImages = new LinkedList<InSlotImage>();

	/** The list of output slots. */
	protected LinkedList<OutSlotImage> outSlotImages = new LinkedList<OutSlotImage>();

	/** The list of input parameter slots. */
	protected LinkedList<InSlotParameter> inSlotParameters = new LinkedList<InSlotParameter>();

	/** The list of output result slots. */
	protected LinkedList<OutSlotResult> outSlotResults = new LinkedList<OutSlotResult>();

	/** The input mask slot if exists */
	protected InSlotMask inSlotMask = null;

	/** The command for getting execution result. Ex: "pstatus". */
	private static String commandStatus="";

	/** The command for displaying output images. Ex: "pvisu". */
	private static String commandVisu="";

	/** The current status of the item. */
	private Status _executionStatus=Status.UNEXECUTED;
	/** An empty constant JPanel for item which hasn't properties (eg. notes). */
	protected static final PropertyViewer BLANK_PROPERTY_PANEL = new PropertyViewer();
	/** The window of properties */
	private PropertyViewer _properties=null;

	/**
	 * Creates an new graph item on the given graph with the given
	 * identifier number. If the identifier = -1 then creates
	 * a new unique identifier number.
	 * 
	 * @param graph the graph that will support this item.
	 * @param identifier the identification number (or -1).
	 */
	public GraphItem( Graph graph, int identifier ) {
		_graph = graph;
		setId(identifier);
		_comment = getCommentEditor();
		initialize();
	}

	/**
	 * Creates an new graph item on the given graph with the given
	 * identifier number. If the identifier = -1 then creates
	 * a new unique identifier number.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the graph that will support this item.
	 * @param identifier the identification number.
	 * @param name the name of the item.
	 * @param category the category of the item
	 */
	public GraphItem( TreeItem treeItem, Graph graph, int identifier, String name, String category ) {
		_graph = graph;
		_treeItem = treeItem;
		_name = name;
		_category = category;
		setId(identifier);
		_comment = getCommentEditor();
		initialize();
	}

	/**
	 * Initializes the graph item.
	 */
	private final void initialize( ) {
		_labelName = new JLabel();
		_labelName.setFont(PreferenceChooser.defaultFont);
		_labelName.setForeground(new Color(0,0,0));
		_textFieldName = new JTextField(20);
		_textFieldName .addFocusListener(new FocusListener() {
			@Override
			public final void focusLost(FocusEvent e) {
				setNickName(((JTextField)(e.getSource())).getText(), true);
			}
			@Override
			public final void focusGained(FocusEvent e){ }
		});

		_labelStatus = new JLabel();
		setLayout(new GridBagLayout());
		addListeners();
		// reads icons.
		if (_unexecutedStatusIcon == null) {
			ClassLoader classLoader = getClass().getClassLoader();
			_unexecutedStatusIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"StatusIcon0.png"));
			_runningStatusIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"StatusIcon1.png"));
			_executedStatusIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"StatusIcon2.png"));
			_errorStatusIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"StatusIcon3.png"));
		}
	}

	/**
	 * Initializes the mouse listeners. Redirects mouse click and double click
	 * to onMouseClicked(MouseEvent e) and onMouseDoubleClicked(MouseEvent e)
	 * respectively. Same for onMousePressed(MouseEvent e) and
	 * onMouseReleased(MouseEvent e).
	 */
	private final void addListeners( ) {
		addMouseListener(new MouseAdapter() {
			@Override
			public final void mouseClicked( MouseEvent e ) {
				if (e.getClickCount()>1) { onMouseDoubleClicked(e); }
			}
			@Override
			public final void mousePressed( MouseEvent e ) {
				if (!e.isShiftDown() && !e.isControlDown()) {
					requestFocusInWindow();
				} else {
					Ariane.workspace.displayProperties(null);
				}
				onMousePressed(e);
			}
			@Override
			public final void mouseReleased( MouseEvent e ) {
				displayPropertyViewer();
				onMouseReleased(e);
			}
		});
		addMouseMotionListener(new MouseMotionAdapter() {
			@Override
			public final void mouseDragged( MouseEvent e ) {
				onMouseDragged(e);
			}
		});

		addFocusListener(new FocusListener() {
			@Override
			public final void focusLost( FocusEvent e ) {
			}
			@Override
			public final void focusGained( FocusEvent e ) {
				displayPropertyViewer();
				if (!isSelected()) {
					GraphItem g = GraphItem.this;
					g.getGraph().addItemInSelection(g,false);
				}
			}
		});
		addKeyListener( new KeyAdapter() {
			@Override
			public final void keyPressed( KeyEvent e ) {
				// redirect events to the related graph.
				getGraph().onKeyPressed(e);
			}
		});
	}

	/**
	 * Removes the item from the graph, but it is still in the history.
	 * Disconnects it from all other items.
	 */
	public void remove( ) {
		disconnectAllSlots();
		if (_properties!=null)
			_properties.setVisible(false);
	}

	/**
	 * Disposes definitively the item from the graph.
	 * Tries to avoid memory leak.
	 */
	public void dispose( ) {
		disconnectAllSlots();
		if (_properties!=null)
			_properties.setVisible(false);

		for (InSlotImage inSlotImage : inSlotImages)
			inSlotImage .dispose();
		inSlotImages.clear();
		for (InSlotParameter inSlotParameter: inSlotParameters) 
			inSlotParameter.dispose();
		inSlotParameters.clear();

		if (inSlotMask!=null)
			inSlotMask.dispose();
		inSlotMask=null;

		for (OutSlotImage outSlotImage : outSlotImages) 
			outSlotImage.dispose();
		outSlotImages.clear();

		for (OutSlotResult outSlotResult : outSlotResults) 
			outSlotResult.dispose();
		outSlotResults.clear();
		_graph=null;
	}

	/**
	 * Deletes the items from the graph.
	 * Pushes it into the undo list.
	 */
	protected final void delete( ) {
		_graph.deleteItem(this);
	}

	/**
	 * Sets the default value the parameters.
	 * @param signature the formatted string with the values (e.g. $0:toto=0; $1:titi=10; ..) 
	 */
	protected final void setSignature( String signature ) {
		if (signature!=null && signature.trim().length()!=0) {
			// Set the default value for the parameters
			try {
				String[][] values = ItemSignatureParser.parse(signature, inSlotParameters.size());
				for ( int i=0; i<inSlotParameters.size(); i++ ) {
					inSlotParameters.get(i).setName(values[i][0]);
					inSlotParameters.get(i).setValue(values[i][1]);
				}
			} catch ( MalformedParameterizedTypeException e) {
				Ariane.workspace.printErrorMessage(getName()+Ariane.lang.getString("Operator.error6"));
			}
		}
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes the item.
	 * 
	 * @return true if the item has been successfully executed.
	 */
	public final boolean execute( ) {
		boolean isMacroOperator = _graph.getSuperGraphItem()!=null;
		if (isMacroOperator) {
			return _graph.getSuperGraphItem().execute();
		} else if (this instanceof ViewerItem ) {
			return execute(true);
		} else {
			boolean result = execute(true);
			// Display all the outputs...
			for (OutSlotImage slot : getOutSlotImages() ) {
				slot.showImage();
			}
			return result;
		}
	}

	/**
	 * Executes the item.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the item has been executed successfully.
	 */
	public final boolean execute( boolean imperative ) {
		if (Graph.isExecutionStopped()) {
			return false;
		}
		if (getExecutionStatus() == Status.EXECUTED && !(this instanceof ViewerItem)) {
			return true;
			}
		if (primitiveExecute(imperative)) {
			updatePropertyViewer();
			return true;
		}
		return false;
	}

	/**
	 * This is the internal execution function.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the item has been executed successfully.
	 */
	abstract protected boolean primitiveExecute( boolean imperative );

	/**
	 * Returns the related subgraph of the item.
	 * 
	 * @return a graph or null.
	 */
	public SubGraph getSubGraph( ){
		return null;
	}

	public final Status getExecutionStatus( ) {
		return _executionStatus;
	}

	/**
	 * Changes the current status of the item from
	 * the specified value. Changes also its parent's status.
	 * 
	 * @param value the new status (UNEXECUTED, EXECUTED, ERROR).
	 */
	protected final void changeStatus( Status value ) {
		if (_executionStatus != value) {
			_executionStatus = value;
			if (_executionStatus == Status.ERROR) {
				if (_graph.getSuperGraphItem()!=null)
					_graph.getSuperGraphItem().changeStatus(value);
			}
			displayStatus();
		}
	}

	/**
	 * A hook for customizing the comment editor.
	 * @return a JLabel.
	 */
	protected JComponent getCommentEditor() {
		JLabel label = new JLabel();
		label.setFont(PreferenceChooser.fontDialogPLAIN8);
		label.setBorder(BorderFactory.createEmptyBorder(2,2,2,2) );
		return label;
	}

	protected final void setComment( String comment ) {
		if (_comment instanceof JLabel) {
			((JLabel)_comment).setText(comment);
		} else if (_comment instanceof JTextArea) {
			((JTextArea)_comment).setText(comment);
		}
	}

	/**
	 * Displays the current status of the item
	 * in the related item widget.
	 */
	private final void displayStatus( ) {
		switch(_executionStatus) {
		case UNEXECUTED:
			_labelStatus.setIcon(_unexecutedStatusIcon);
			break;
		case RUNNING:
			_labelStatus.setIcon(_runningStatusIcon);
			break;
		case EXECUTED:
			_labelStatus.setIcon(_executedStatusIcon);
			break;
		case ERROR:
			_labelStatus.setIcon(_errorStatusIcon);
			break;
		}
	}

	/**
	 * Resets the item and all its forwarded connected items
	 * and its parent.
	 */
	public void resetAll( ) {
		if (_executionStatus != Status.UNEXECUTED) {
			reset();
			if (outSlotImages != null) {
				for (OutSlotImage slot: outSlotImages) {
					slot.reset();
					// Reset all connected items.
					if (slot.isConnected())
						for (InSlot inSlot: slot.getConnectedSlots())
							inSlot.getGraphItem().resetAll();
				}
			}

			if (outSlotResults.size() > 0) {
				getOutSlotResult().reset();
				if (getOutSlotResult().isConnected())
					for (InSlot inSlot : getOutSlotResult().getConnectedSlots())
						inSlot.getGraphItem().resetAll();
			}
			// Its parent
			if (_graph.getSuperGraphItem()!=null) {
				_graph.getSuperGraphItem().resetAll();
			}
			updatePropertyViewer();
		}
	}

	/**
	 * Resets the item without all its forwarded connected items
	 * and without its parent.
	 */
	protected void reset() {
		changeStatus(Status.UNEXECUTED);
	}

	/**
	 * Disconnects all input and output slots.
	 */
	public final void disconnectAllSlots( ) {
		for (OutSlot slot : outSlotImages)
			slot.disconnectAll();
		if (inSlotParameters != null) {
			for (InSlot slot : inSlotParameters)
				slot.disconnect();
		}
		for (InSlot slot : inSlotImages) {
			slot.disconnect();
		}
		if (inSlotMask != null) inSlotMask.disconnect();
		if (outSlotResults.size() > 0) getOutSlotResult().disconnectAll();
	}

	///// 
	///// A C C E S S O R S
	/////

	/**
	 * Returns the identifier number of the item.
	 * 
	 * @return an integer.
	 */
	public int getId( ){
		return _identifier;
	}

	/**
	 * Sets the identifier of the item. This identifier is
	 * guaranty to be unique in a same graph.
	 * If the parameter identifier (i.e., the previous identifier
	 * in a previous session) is > -1, then the identifier is kept
	 * in _readIdentifier.
	 * 
	 * @param identifier the previous identifier number.
	 */
	protected final void setId( int identifier ) {
		// Set a unique identifier number!
		_identifier = _graph.getGraphItemCount()+1;
		// Keeps the current read session.
		_currentReadPasteSession = Workspace.getCurrentReadPasteSession();
		if (identifier != -1)
			_readIdentifier = identifier;
	}

	/**
	 * Returns the current identifier of the read session
	 * (use for loading and pasting items).
	 * 
	 * @return the number of the current loading session.
	 */
	public final int getCurrentReadPasteSession( ){
		return _currentReadPasteSession;
	}

	/**
	 * Returns the identifier number of the item
	 * in a previous session (saved or pasted).
	 * 
	 * @return the identifier number.
	 */
	public final int getReadIdentifier( ){
		return _readIdentifier;
	}

	/**
	 * Returns the category of the item.
	 * 
	 * @return the name of the category.
	 */
	public final String getCategory( ) {
		if (_treeItem==null) return _category;
		return _treeItem.getCategory();
	}

	/**
	 * Returns the address of the related documentation HTML page.
	 * 
	 * @return the complete or relative address.
	 */
	public String getDocURL( ) {
		if (_treeItem==null) return "";
		return _treeItem.getDocURL();
	}

	/**
	 * Returns the absolute path of the executable if possible.
	 * Possible cases are:
	 * <ul>
	 * <li>If the binary already contains the absolute path,
	 * <li>if the binary is in the associated library.
	 * </ul>
	 * In all other cases, the binary is supposed to be accessible by
	 * the environment variable PATH, and the function returns the binary itself.
	 * 
	 * @param binary the abstract name of the executable.
	 * @return the absolute path to a known executable or the binary itself.
	 */
	public static final String getAbsoluteBinary( String binary ) {
		// !!!
		// !!! Note: even if I modify the environment variable PATH, the processBuilder cannot find
		// !!! the required Shell command. It seems that it needs the absolute path !
		// !!!
		if (binary==null) return null;
		// User operator
		File unixFile;
		File windowsFile;
		String binaryDirectory= PreferenceChooser.get("BinaryDirectory");

		if (binaryDirectory!=null) {
			unixFile= new File(binaryDirectory+File.separator+File.separator+binary);
			windowsFile= new File(binaryDirectory+File.separator+File.separator+binary+".exe");
			if (windowsFile.exists() && windowsFile.isFile()) {
				// An operator of the library
				return windowsFile.getAbsolutePath();
			}
			if (unixFile.exists() && unixFile.isFile()) {
				// An operator of the library
				return unixFile.getAbsolutePath();
			}
		}

		// Already an absolute path
		if (binary.indexOf(File.separator)>=0) {
			return binary;
		}

		// Unix case
		unixFile = new File(Ariane.getLibraryPath()+File.separator+"bin"+File.separator+binary);
		windowsFile = new File(Ariane.getLibraryPath()+File.separator+"bin"+File.separator+binary+".exe");
		if (windowsFile.exists()) {
			// An operator of the library
			return windowsFile.getAbsolutePath();
		}
		if (unixFile.exists()) {
			// An operator of the library
			return unixFile.getAbsolutePath();
		}
		// Default case: return the binary itself
		return binary;
	}

	/**
	 * Returns the name of the executable.
	 * 
	 * @return the name.
	 */
	public String getBinary( ) {
		return _binary;
	}

	/**
	 * Returns the name of the executable using a relative
	 * path name if filepath is specified.
	 * 
	 * @param filePath the file path or null.
	 * @return the name.
	 */
	public String getBinary( String filePath ) {
		return _binary;
	}

	/**
	 * Sets the path to the executable.
	 * final 
	 * @param binary the complete path.
	 */
	protected void setBinary( String binary ) {
		_binary=binary;
	}

	/**
	 * Returns the name of the formula using a relative
	 * path name if it is a file and the filepath is specified.
	 * 
)	 * @param filePath the file path or null.
	 * @return the name.
	 */
	public String getFormula( String filepath ) {
		return _formula;
	}

	public boolean displayFormula() {
		return true;
	}

	/**
	 * Returns the formula of the logical or arithmetic expression.
	 * 
	 * @return the formula.
	 */
	public final String getFormula( ) {
		return _formula;
	}

	/**
	 * Returns the formula of the logical or arithmetic expression.
	 * 
	 * @param textFormula the formula.
	 * @return true if the formula is valid, false otherwise.
	 */
	public boolean changeFormula( String textFormula ) { return true; }

	/**
	 * Sets the new formula.
	 * 
	 * @param formula the formula.
	 */
	protected void setFormula( String formula ) {
		_formula=formula;
		resetAll();
		getGraph().setModified(true);
		_labelName.setText(getNickName());
		getGraph().repaint();
	}

	/**
	 * Returns the name of the item.
	 * 
	 * @return the name.
	 */
	@Override
	public final String getName( ) {
		return _name;
	}

	/**
	 * Returns the nickname of the item if any
	 * otherwise returns the name.
	 * 
	 * @return the nickname.
	 */
	public String getNickName( ) {
		return (_nickName==null)?getName():_nickName;
	}

	/**
	 * Returns the absolute name of the item, including
	 * the graph and subgraphs names.
	 * 
	 * @return the absolute name.
	 */
	protected final String getAbsoluteName( ) {
		return getGraph().getName()+"/"+getNickName();
	}

	/**
	 * Sets the item name.
	 * 
	 * @param name the new name.
	 */
	public final void setAbstractName( String name ) {
		if (name.length() > 0)
			_name = name;
	}

	public final void setTemporaryNickName( String name ) {
		_nickName = name;
		updateName();
	}

	/**
	 * Sets the item nickname.
	 * 
	 * @param name the new nickname.
	 */
	public void setNickName( String name ) {
		if (name != null && name.length() > 0) {
			_nickName = name;
			_textFieldName.setText(name);
		} else {
			_nickName = null;
			_textFieldName.setText("");
		}
		updatePropertyViewer();
		getGraph().setModified(true);
	}

	/**
	 * Updates the related property viewer if any.
	 */
	public final void updatePropertyViewer() {
		if (_properties!=null) _properties.update(this);
	}

	/**
	 * Sets the new nickname to the item.
	 * 
	 * @param newName the new nickname.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void setNickName( String newName, boolean historical ) {
		String oldName = _nickName;
		setNickName(newName);
		if (historical) {
			getGraph().startActionEdition();
			getGraph().validateAction(new RenameItemEdit(this, newName, oldName));
			getGraph().endActionEdition();
		}
		updateName();
		revalidate();
		getGraph().repaint();
	}

	/**
	 * Returns the graph that holds this item.
	 * 
	 * @return the Graph.
	 */
	public final Graph getGraph( ) {
		return _graph;
	}

	/**
	 * @return the related tag name.
	 */
	public abstract String tagName( );

	/**
	 * Returns the list of input images.
	 * 
	 * @return the LinkedList of InSlotImage
	 */
	public final LinkedList<InSlotImage> getInSlotImages( ) {
		return inSlotImages;
	}

	/**
	 * Returns the list of output images.
	 * 
	 * @return the LinkedList of OutSlotImage
	 */
	public final LinkedList<OutSlotImage> getOutSlotImages( ) {
		return outSlotImages;
	}

	/**
	 * Returns the input mask.
	 * 
	 * @return InSlotMask
	 */
	public final InSlotMask getInSlotMask( ) {
		return inSlotMask;
	}

	/**
	 * Returns the list of input parameters.
	 * 
	 * @return the LinkedList of InSlotParameter
	 */
	public final LinkedList<InSlotParameter> getInSlotParameters( ) {
		return inSlotParameters;
	}

	/**
	 * Returns the list of output results.
	 * 
	 * @return the LinkedList of OutSlotResult
	 */
	public final LinkedList<OutSlotResult> getOutSlotResults( ) {
		return outSlotResults;
	}

	/**
	 * Returns the first output result.
	 * 
	 * @return the first OutSlotResult
	 */
	public final OutSlotResult getOutSlotResult( ) {
		if (outSlotResults.size()>0)
			return outSlotResults.getFirst();
		return null;
	}

	/**
	 * Returns the slot related to the given rank <i>rank</i>
	 * of the given type.
	 * 
	 * @param typeSlot the type of the slot.
	 * @param rank the rank of the slot in the item.
	 * @return a slot or null.
	 */
	public final Slot getSlot( String typeSlot, int rank ) {
		if (typeSlot.equalsIgnoreCase("InSlotImage"))
			return getInSlotImages().get(rank);
		else if (typeSlot.equalsIgnoreCase("InSlotMask"))
			return getInSlotMask();
		else if (typeSlot.equalsIgnoreCase("InSlotParameter"))
			return getInSlotParameters().get(rank);
		else if (typeSlot.equalsIgnoreCase("OutSlotImage")) {
			if (rank>=getOutSlotImages().size()) return null;
			return getOutSlotImages().get(rank);
		}
		else if (typeSlot.equalsIgnoreCase("OutSlotResult"))
			return getOutSlotResults().get(rank);
		return null;
	}

	///// 
	///// P R E F E R E N C E S
	///// 

	/**
	 * @return the name of status binary command.
	 */
	public static final String getCommandStatus( ) {
		return commandStatus;
	}

	/**
	 * The name of status binary command.
	 * 
	 * @param cmd the command name.
	 */
	public static final void setCommandStatus( String cmd ) {
		commandStatus = cmd;
	}

	/**
	 * @return the name of the visualization binary software.
	 */
	public static final String getCommandVisu( ) {
		return commandVisu;
	}

	/**
	 * Sets the name of the visualization binary software.
	 * 
	 * @param cmd the command name.
	 */
	public static final void setCommandVisu( String cmd ) {
		commandVisu = cmd;
	}

	///// 
	///// D I S P L A Y I N G  &  P R I N T I N G
	///// 

	/**
	 * Displays the item content.
	 */
	public void draw( ) {
		removeAll();
		Color gc= getBackgroundColor();
		setBackground(gc);
		setBorder(unselectedBorder);
		int max = (inSlotImages.size()>outSlotImages.size()) ? inSlotImages.size():outSlotImages.size();
		int nbparam = inSlotParameters.size();
		if (nbparam == 0) { nbparam=1; }

		_labelName.setText(getNickName());
		setToolTipText(toString());

		// layout
		GridBagConstraints c = new GridBagConstraints();
		c.gridheight = 1;

		// parameters slots
		c.anchor = GridBagConstraints.NORTHWEST;
		c.gridy = 0;
		c.insets = new Insets(1, 5, 1, 0);

		for (int i = 0; i < inSlotParameters.size(); i++) {
			Slot slot = (inSlotParameters.get(i));
			slot.updateBackgroundColor(gc);
			c.gridx = i;
			add(slot, c);
			c.insets = new Insets(1, 0, 1, 0);
		}

		// The mask slot
		c.insets = new Insets(0, 0, 0, 5);
		c.anchor = GridBagConstraints.NORTHEAST;
		c.gridx = nbparam;
		if (inSlotMask != null) {
			inSlotMask.updateBackgroundColor(gc);
			add(inSlotMask, c);
		}
		c.insets = new Insets(0, 0, 0, 0);
		// The name
		c.gridwidth = nbparam + 1;
		c.gridy++; c.gridx = 0;
		c.anchor = GridBagConstraints.CENTER;
		add(_labelName, c);

		// The lights ramp that shows the current status.
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.gridy++; c.gridx = 0;
		displayStatus();
		add(_labelStatus, c);

		// ???
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.gridy++; c.gridx = 0;
		add(_comment, c);

		// The input slots
		c.insets = new Insets(0, 0, 1, 1);
		c.gridy++; c.gridx = 0;
		c.gridwidth = nbparam;
		c.anchor = GridBagConstraints.WEST;
		for (int i = 0; i < inSlotImages.size(); i++){
			Slot slot = (inSlotImages.get(i));
			slot.updateBackgroundColor(gc);
			c.gridy += i;
			add(slot, c);
		}

		// The output slots
		c.insets = new Insets(0, 0, 1, 1);
		c.gridx = nbparam; c.gridwidth = 1; c.anchor = GridBagConstraints.EAST;
		for (int i = 0; i < outSlotImages.size(); i++) {
			Slot slot = (outSlotImages.get(i));
			slot.updateBackgroundColor(gc);
			c.gridy += i;
			add(slot, c);
		}

		// The result
		c.gridy += max;
		c.insets = new Insets(0, 0, 2, 10);
		if (outSlotResults.size() > 0){
			getOutSlotResult().updateBackgroundColor(gc);
			c.gridx = nbparam;
			c.gridwidth = 1;
			c.gridheight = 1;
			c.anchor = GridBagConstraints.SOUTHEAST;
			add(getOutSlotResult(), c);
		}
	}

	/**
	 * Centers the specified text in the component.
	 * If the length of the text is larger than the component
	 * width, then use several lines.
	 * 
	 * @param text the text to be centered.
	 * @param row the line where to display the text.
	 * @param multiline can print the text with several lines?
	 * @param gc the graphics context.
	 */
	protected final void centerText( String text, Graphics gc, int row, boolean multiline ) {
		int column;
		Dimension size = getSize();
		FontMetrics fm;
		fm = gc.getFontMetrics();

		if (text==null) return;
		if (fm.stringWidth(text)>size.width) {
			if (multiline) {
				// Use several lines and do not center text.
				String tmp;
				while(text.length()>0) {
					int i=text.length();
					tmp=text;
					while (fm.stringWidth(tmp)>size.width-4){
						tmp=text.substring(0,--i);
					}
					gc.drawString(tmp, 2, row);
					row+=fm.getHeight();
					text=text.substring(i);
				}
			}
			gc.drawString(text, 2, row);
		} else { // Centers the text
			column = (size.width-fm.stringWidth(text))/2;
			gc.drawString(text, column, row);
		}
	}

	/**
	 * Cancels printing of item contents.
	 * 
	 * @param gc the graphics context.
	 */
	@Override
	public final void printChildren( Graphics gc ) { }

	/**
	 * Prints the border of the item;
	 * 
	 * @param gc the graphics context.
	 */
	@Override
	public final void printBorder( Graphics gc ) {
		Dimension size;
		size = new Dimension(getSize());
		Graphics2D g2d=(Graphics2D)gc;
		g2d.setColor(Color.black);
		BasicStroke lineStroke=new BasicStroke(2.0F, BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND);
		g2d.setStroke(lineStroke);
		g2d.drawRect(0,0,size.width,size.height);
	}

	/**
	 * Prints the items. Just draw a line to split the
	 * name of the item and the content.
	 * 
	 * @param gc the graphics context.
	 */
	@Override
	public void printComponent( Graphics gc ) {
		Dimension size;
		int row;
		size = new Dimension(getSize());
		gc.setColor(Color.white);
		gc.fillRect(0,0,size.width,size.height);

		gc.setColor(Color.black);
		gc.setFont(PreferenceChooser.defaultFont);
		FontMetrics fm;
		fm = gc.getFontMetrics(); 
		row = fm.getHeight();
		// For aestheticism purpose, split title and parameters.
		gc.drawLine(0,row+4,size.width,row+4);
	}

	/** 
	 * @return the current background color.
	 */
	public abstract Color getBackgroundColor( );

	/**
	 * Shows the related documentation HTML page on a 
	 * separated window.
	 */
	public final void showMan( ) {
		if (getDocURL() != null) {
			Ariane.workspace.displayMan(Ariane.lang.getString("HTMLViewer.manual")+" "+_name, getDocURL(), null);
		}
	}

	///// 
	///// S E L E C T I O N
	///// 

	/** The status of the item between selected or not. */
	private boolean _selected = false;

	public final boolean isSelected( ) {
		return _selected;
	}

	public final void setSelected( boolean selected ) {
		if (_selected != selected) {
			for ( InSlot slot : inSlotImages ) {
				slot.setSelected(selected);
			}
			for ( InSlot slot : inSlotParameters ) {
				slot.setSelected(selected);
			}
			for ( OutSlot outSlot : outSlotImages ) {
				for ( InSlot inSlot : outSlot.getConnectedSlots() ) {
					inSlot.setSelected(selected);
				}
			}
			_selected = selected;
			setBorder((selected)?selectedBorder:unselectedBorder);
		}
	}

	protected final void onMouseDragged( MouseEvent e ) {
		if ((_popupMenu!=null) && (_popupMenu.isVisible()))
			return;
		if (!_selected) {
			getGraph().addItemInSelection(this,e.isShiftDown()||e.isControlDown());
		}
		e.translatePoint(getLocation().x,getLocation().y);
		getGraph().dragSelection(e);
	}

	protected final void onMousePressed( MouseEvent e ) {
		getParent().setComponentZOrder(this, 0); // Put it in front
		if (e.isPopupTrigger()) { // Unix version -> open the popup menu
			getGraph().addItemInSelection(this, false);
			drawPopup(e);
		}
	}

	protected final void onMouseReleased( MouseEvent e ) {
		if (e.isPopupTrigger()) { // Window version -> open the popup menu
			getGraph().addItemInSelection(this,false);
			drawPopup(e);
		} else {
			if ((_popupMenu==null || !_popupMenu.isVisible()) && !getGraph().dropSelection()) {
				if (e.isShiftDown()||e.isControlDown()) {
					getGraph().addItemInSelection(this,true);
				}
			}
		}
	}

	protected void onMouseDoubleClicked( MouseEvent e ) { }

	///// 
	///// M E N U
	///// 

	/**
	 * Displays the property viewer in the workspace.
	 */
	public final void displayPropertyViewer() {
		Ariane.workspace.displayProperties(getPropertyViewer());
	}

	/**
	 * Creates a related property viewer.
	 * 
	 * @return a dialog box.
	 */
	protected PropertyViewer createPropertyViewer() {
		return new PropertyViewer(Ariane.lang, this);
	}

	/**
	 * @return the current property window.
	 */
	private final PropertyViewer getPropertyViewer() {
		if (_properties==null || _properties.getLanguage()!=Ariane.lang ) {
			_properties = createPropertyViewer();
		} 
		_properties.setVisible(true);
		return _properties;
	}

	/**
	 * @return the list of specific properties (other than common properties).
	 */
	public ArrayList<PropertyViewer.PropertyCouple> getSpecificProperties( ) { return null; }

	/**
	 * Changes the formula expression and store
	 * it in undo -redo manager.
	 * @param component the component that displays the formula.
	 * @param newValue the new formula expression.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void changeFormula( final JComponent component, String newValue, boolean historical ) {
		String oldFormula = getFormula();
		changeFormula(newValue);

		// Very ugly but YAGNI...
		if (component instanceof JTextField) {
			((JTextField)component).setText(getFormulaText());
		} else if (component instanceof JButton) {
			((JButton)component).setText(getFormulaText());
		}
		updatePropertyViewer();
		if (historical) {
			getGraph().startActionEdition();
			getGraph().validateAction(new ChangeFormulaEdit(this, component, getFormula(), oldFormula));
			getGraph().endActionEdition();
		}
	}

	/**
	 * Changes the binary expression and store
	 * it in undo -redo manager.
	 * @param component the component that displays the formula.
	 * @param newValue the new formula expression.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public final void changeBinary( final JComponent component, String newValue, boolean historical ) {
		String oldBinary = getBinary();
		setBinary(newValue);
		// Very ugly but YAGNI...
		if (component instanceof JTextField) {
			((JTextField)component).setText(getBinary());
		} else if (component instanceof JButton) {
			((JButton)component).setText(getBinary());
		}
		updatePropertyViewer();
		if (historical) {
			getGraph().startActionEdition();
			getGraph().validateAction(new ChangeFormulaEdit(this, component, getBinary(), oldBinary));
			getGraph().endActionEdition();
		}
	}

	/**
	 * @return the user's representation of the formula.
	 */
	public String getFormulaText() {
		return getFormula();
	}

	/**
	 * @return the text associated to the formula editor.
	 */
	public String getFormulaLabel( ) {
		return Ariane.lang.getString("PropertyViewer.formula");
	}

	/**
	 * Builds a component appropriate to display and change the formula.
	 * 
	 * @return a new component.
	 */
	public JComponent getFormulaEditor( ) {
		JTextField formulaText = new JTextField(getFormulaText(), 20);
		formulaText.addFocusListener(new FocusListener() {
			@Override
			public final void focusLost(FocusEvent e) {
				changeFormula((JTextField)e.getSource(), ((JTextField)e.getSource()).getText(), true);
			}
			@Override
			public final void focusGained(FocusEvent e){
			}
		});
		return formulaText;
	}

	/**
	 * @return a component appropriate to edit and possibly change the name of the item.
	 */
	public JComponent getNameEditor() {
		return _textFieldName;
	}

	/**
	 * @return a component appropriate to edit and possibly change the name of the item.
	 */
	public JComponent getBinaryEditor() {
		if (getBinary() != null) {
			JTextField c = new JTextField(getBinary());
			c.setEnabled(false);
			c.setEditable(false);
			return c;
		}
		return null;
	}

	/**
	 * Updates the item content after modification
	 * through the properties dialog box.
	 */
	public final void updateName() {
		_labelName.setText(getNickName());
		setToolTipText(toString());
	}

	/**
	 * Creates and shows the related popup menu.
	 * 
	 * @param e the MouseEvent.
	 */
	protected final void drawPopup( MouseEvent e ) {
		_popupMenu = initPopup();
		_popupMenu.show(this, e.getX(), e.getY());
	}

	/**
	 * Creates the related popup menu.
	 * 
	 * @return a convenient popup menu.
	 */
	protected abstract JPopupMenu initPopup( );

	///// 
	/////  I N P U T S   &   O U T P U T S
	///// 

	/**
	 * Saves the item as a XML tree.
	 * 
	 * @param file the filename of the project.
	 * @return the xml Tree.
	 * @throws IOException in case of output error.
	 */
	public XMLElement save( File file) throws IOException {
		XMLElement e = new XMLElement(tagName());
		e.addChildren(saveProperties(file));
		return e;
	}

	/**
	 * Saves properties of the graph item in an XML tree.
	 * Saves only filled properties.
	 * 
	 * @param file the filename of the project.
	 * @return the XML tree.
	 * @throws IOException in case of output error.
	 */
	public final LinkedList<XMLElement> saveProperties( File file ) throws IOException {
		LinkedList<XMLElement> xmlTree = new LinkedList<XMLElement>();
		xmlTree.add(new XMLElement("NickName", _nickName));
		xmlTree.add(new XMLElement("Name", getName()));
		xmlTree.add(new XMLElement("Identifier", Integer.toString(getId())));
		xmlTree.add(new XMLElement("X-coordinate", Integer.toString(getLocation().x)));
		xmlTree.add(new XMLElement("Y-coordinate", Integer.toString(getLocation().y)));

		String current_path=null;
		if (file!= null) {
			current_path = file.getParentFile().getAbsolutePath();
		}
		String tmp = getBinary(current_path);
		if (tmp != null) {
			xmlTree.add(new XMLElement("Binary", tmp));
		}
		if (getCategory() != null) {
			xmlTree.add(new XMLElement("Category", getCategory()));
		}
		if (getFormula() != null) {
			xmlTree.add(new XMLElement("Formula", getFormula(current_path)));
		}
		if (inSlotImages.size()>0) {
			xmlTree.add(new XMLElement("InputImages", Integer.toString(inSlotImages.size())));
		}
		if (inSlotParameters.size()>0) {
			xmlTree.add(new XMLElement("Parameters", Integer.toString(inSlotParameters.size())));
		}
		if (outSlotImages.size()>0)
			xmlTree.add(new XMLElement("OutputImages",Integer.toString(outSlotImages.size())));

		if (this instanceof IScriptArgument) {
			IScriptArgument item = (IScriptArgument)this;
			xmlTree.add(new XMLElement("Is-Argument",""+item.isArgument()));
		}
		return xmlTree;
	}

	/**
	 * Saves each of its slots and its related connection
	 * in an XML tree.
	 * 
	 * @return the xml Tree.
	 * @throws IOException in case of output error.
	 */
	public final LinkedList<XMLElement> saveSlots( ) throws IOException {
		LinkedList<XMLElement> l= new LinkedList<XMLElement>();
		for (int i = 0; i < inSlotImages.size(); i++) {
			XMLElement e = inSlotImages.get(i).save();
			if (e != null)
				l.add(e);
		}

		for (int i = 0; i < inSlotParameters.size(); i++) {
			XMLElement e = inSlotParameters.get(i).save();
			if (e != null)
				l.add(e);
		}
		if (inSlotMask != null) {
			XMLElement e = inSlotMask.save();
			if (e!= null)
				l.add(e);
		}
		return l;
	}

	/**
	 * Loads properties of the item from the given XML tree.
	 * 
	 * @param xmlTree the tree of XML elements.
	 * @throws IOException in case of input error.
	 */
	public void load( LinkedList<XMLElement> xmlTree ) throws IOException {
		loadProperties(xmlTree);
	}

	/**
	 * @return the related tree item in the explorer if any.
	 */
	protected TreeItem getRelatedTreeItem( ) {
		return Ariane.workspace.getTree().getTreeItem(getName(), _category);
	}

	/**
	 * Loads graph item properties from the given XML tree.
	 * 
	 * @param xmlTree the tree of properties.
	 * @throws IOException in case of input error.
	 */
	protected final void loadProperties( LinkedList<XMLElement> xmlTree ) throws IOException {
		XMLElement e;
		int inParams=0, inImages=0, outImages=0;
		Point coordinates=new Point(0,0);

		Iterator<XMLElement> it = xmlTree.iterator();
		while(it.hasNext()){
			e = it.next();
			// Inside <Operateur> or <DataFile> or ...
			// French tags are kept for sake of compatibility.
			if (e.getName().equalsIgnoreCase("NickName")) {
				setNickName(e.getValue());
			} else if (e.getName().equalsIgnoreCase("Name") || e.getName().equalsIgnoreCase("Nom")) {
				setAbstractName(e.getValue());
			} else if (e.getName().equalsIgnoreCase("Identifier") || e.getName().equalsIgnoreCase("Identifieur"))
				setId(Integer.parseInt(e.getValue()));
			else if (e.getName().equalsIgnoreCase("X-coordinate") || e.getName().equalsIgnoreCase("Abscisse"))
				coordinates.x = Integer.parseInt(e.getValue());
			else if (e.getName().equalsIgnoreCase("Y-coordinate") || e.getName().equalsIgnoreCase("Ordonnee"))
				coordinates.y = Integer.parseInt(e.getValue());
			else if (e.getName().equalsIgnoreCase("Binary") || e.getName().equalsIgnoreCase("Binaire")) {
				setBinary(e.getValue());
			} else if (e.getName().equalsIgnoreCase("Category") || e.getName().equalsIgnoreCase("Categorie")) {
				_category = e.getValue();
			} else if (e.getName().equalsIgnoreCase("Formula")) {
				setFormula(e.getValue());
			} else if (e.getName().equalsIgnoreCase("Parameters") || e.getName().equalsIgnoreCase("ParametresEntree")) {
				inParams = Integer.parseInt(e.getValue());
			} else if (e.getName().equalsIgnoreCase("InputImages") || e.getName().equalsIgnoreCase("ImagesEntree")) {
				inImages = Integer.parseInt(e.getValue());
			} else if (e.getName().equalsIgnoreCase("OutputImages") || e.getName().equalsIgnoreCase("ImagesSortie")) {
				outImages = Integer.parseInt(e.getValue());
			} else if (e.getName().equalsIgnoreCase("Is-Argument")) {
				if (this instanceof IScriptArgument) {
					IScriptArgument item = (IScriptArgument)this;
					item.setIsArgument(Boolean.parseBoolean(e.getValue()));
				}
			}
		}
		for (int i = 0; i < inImages; i++) { inSlotImages.add(new InSlotImage(this,i)); }
		for (int i = 0; i < inParams; i++) { inSlotParameters.add(new InSlotParameter(this,i)); }
		for (int i = 0; i < outImages; i++) { outSlotImages.add(new OutSlotImage(this, i)); }
		if (_treeItem==null) {
			_treeItem = getRelatedTreeItem();
		}
		setLocation(coordinates);
	}

	/**
	 * Loads a graph item from the given XML tree.
	 * 
	 * @param graph the graph.
	 * @param xmlTree the XML tree.
	 * @param translation translate or not the item location (+5 pixels).
	 * @throws IOException in case of input error.
	 */
	public static final ArrayList<GraphItem> load( Graph graph, LinkedList<XMLElement> xmlTree, boolean translation ) throws IOException {
		ArrayList<GraphItem> items = new ArrayList<GraphItem>();
		if (xmlTree == null) // Empty tree.
			return items;

		Iterator<XMLElement> ite = xmlTree.iterator();
		while(ite.hasNext()) {
			XMLElement el = ite.next();
			GraphItem item=null;
			if (el.getName().equalsIgnoreCase(OperatorItem.TAGNAME) || el.getName().equalsIgnoreCase("Operateur")) {
				item = new OperatorItem(graph, el.getChildren());
			} else if (el.getName().equalsIgnoreCase(NoteItem.TAGNAME)) {
				item = new NoteItem(graph,el.getChildren());
			} else if (el.getName().equalsIgnoreCase("Fichier")) {
				// Kept for the sake of compatibility
				item = DataFile.getInstance(graph,el.getChildren());
			} else if (el.getName().equalsIgnoreCase(LoaderItem.TAGNAME)) {
				item = new LoaderItem(graph,el.getChildren());
			} else if (el.getName().equalsIgnoreCase(SaverItem.TAGNAME)) {
				item = new SaverItem(graph,el.getChildren());
			} else if (el.getName().equalsIgnoreCase(ViewerItem.TAGNAME)) {
				item = new ViewerItem(graph,el.getChildren());
			} else if (el.getName().equalsIgnoreCase(FileNameItem.TAGNAME)
					|| el.getName().equalsIgnoreCase("Datafile")) {
				item = new FileNameItem(graph,el.getChildren());
			} else if (el.getName().equalsIgnoreCase(ConstantItem.TAGNAME)) {
				item = new ConstantItem(graph,el.getChildren());
			} else if (el.getName().equalsIgnoreCase(SystemItem.TAGNAME)) {
				item = new SystemItem(graph,el.getChildren());
			} else if (el.getName().equalsIgnoreCase(ScriptItem.TAGNAME)) {
				item = new ScriptItem(graph,el.getChildren());
			} else if (el.getName().equalsIgnoreCase(MacroOperatorItem.TAGNAME)
					|| el.getName().equalsIgnoreCase("MacroOperateur")) {
				item = new MacroOperatorItem(graph,el.getChildren());
			} else if (el.getName().equalsIgnoreCase(ControlStructureMergeItem.TAGNAME)) {
				item = new ControlStructureMergeItem(graph, el.getChildren());
			} else if (el.getName().equalsIgnoreCase(EvalItem.TAGNAME)) {
				item = new EvalItem(graph, el.getChildren());
			} else if (el.getName().equalsIgnoreCase(FormatItem.TAGNAME)) {
				item = new FormatItem(graph, el.getChildren());
			} else if (el.getName().equalsIgnoreCase(ControlStructureIfItem.TAGNAME)) {
				item = new ControlStructureIfItem(graph, el.getChildren());
			} else if (el.getName().equalsIgnoreCase(ControlStructureSelectItem.TAGNAME)) {
				item = new ControlStructureSelectItem(graph, el.getChildren());
			} else if (el.getName().equalsIgnoreCase(ControlStructureWhileItem.TAGNAME)) {
				item = new ControlStructureWhileItem(graph, el.getChildren());
			} else if (el.getName().equalsIgnoreCase("Until")) { // Keep for compatibility
				item = new ControlStructureWhileItem(graph,el.getChildren());
			} else if (el.getName().equalsIgnoreCase(ControlStructureForItem.TAGNAME)) {
				item = new ControlStructureForItem(graph,el.getChildren());
			}
			if (item != null) {
				Point coordinates = item.getLocation();
//				if (translation) {
//					coordinates.translate(+5,+5);
//				}
				graph.addItem(item, coordinates, true);
				items.add(item);
			}
		}
		return items;
	}
}