/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.SubGraph;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.InSlotMask;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotResult;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>MacroOperatorItem</code> is a graph item that is composed by
 * a subgraph of items.
 *
 * @author Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Alban DOUBLET, Guillaume CHATELET - Jun 10, 2002
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Jan 16, 2007
 */
public final class MacroOperatorItem extends CompositeItem {
	/** The tag name of the item. */
	public static final String TAGNAME="MacroOperator";

	/**
	 * Creates a <code>MacroOperatorItem</code> on the given graph with 
	 * properties interactively asked to the user.
	 * Initializes the mouse listeners, the popup menu,
	 * and displays it in the graph.
	 * 
	 * @param treeItem the reference tree item.
	 * @param graph the parent graph.
	 * @param identifier the identification number (or -1).
	 */
	public MacroOperatorItem( TreeItem treeItem, Graph graph, int identifier ) {
		super(treeItem, graph,identifier,"" );
		setAbstractName("macro-operator");
		askProperties("macro-operator"+getId(),1,1,1);
	}

	/**
	 * Creates a <code>MacroOperatorItem</code> on the given graph with
	 * the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @throws IOException XML errors.
	 */
	public MacroOperatorItem( Graph graph, LinkedList<XMLElement> xmlTree ) throws IOException {
		super(graph,-1);
		load(xmlTree);
		// Keep the 4 following lines for sake of compatibility
		if (_treeItem==null) {
			_treeItem = restoreDefaultTreeItem();
		}
		setSignature(_treeItem.getSignature());
		inSlotMask = new InSlotMask(this);
		draw();
	}

	/**
	 * Creates a <code>MacroOperatorItem</code> 
	 * as a subroutine (undeditable item) on the given graph with
	 * the given XML data.
	 * 
	 * @param graph the parent graph.
	 * @param xmlTree the XML tree of data.
	 * @param treetem the related TreeItem.
	 * @throws IOException XML errors.
	 */
	public MacroOperatorItem( Graph graph, LinkedList<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
		super(graph,-1, treeItem);
		load(xmlTree);
		setSignature(_treeItem.getSignature());
		inSlotMask = new InSlotMask(this);
		setAbstractName(treeItem.getName());
		setNickName(treeItem.getName());
		draw();
	}

	@Override
	protected TreeItem restoreDefaultTreeItem() {
		setNickName(getNickName());
		setAbstractName("macro-operator");
		return Ariane.workspace.getTree().getTreeItem("macro-operator",null);
	}

	/**
	 * Builds or changes the prototype of the macro-operator
	 * from the given values.
	 * Creates and changes also the related subgraph.
	 * 
	 * @param name the name of the macro-operator.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 */
	@Override
	protected final void setPrototype( String name, int iIm, int iP, int oIm ) {
		setPrototype(name, iIm, iP, oIm,0,0);
	}

	/**
	 * Creates the subgraph with the convenient prototype.
	 * 
	 * @param name the name of the macro-operator.
	 * @param iIm the number of input images.
	 * @param iP the number of parameters.
	 * @param oIm the number of output images.
	 * @return the subgraph.
	 */
	@Override
	protected final SubGraph createSubGraph( String name, int iIm, int iP, int oIm ) {
		return createSubGraph(getNickName(),iIm,iP,oIm,0,0);
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public final String tagName ( ) { return TAGNAME; }

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes the content of the macro-operator.
	 * Execution consists in:
	 * <ul>
	 * <li>Gets input data.
	 * <li>Executes the subgraph.
	 * <li>Gets the results.
	 * </ul>
	 * Considers that a fatal error is in case of :
	 * <ul>
	 * <li>imperative: any errors.
	 * <li>non imperative: only systems or slot errors (unset).
	 * </ul>
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			// Makes links between the Macro inSlots and the VirtualOperator inSlots.
			VirtualItem virtualOperator = getVirtualOperator();

			// Gets and checks parameters (-> executes forward operators if needed).
			Iterator<InSlotParameter> its = getInSlotParameters().iterator();
			Iterator<OutSlotResult> itd   = virtualOperator.getVirtualInSlotParameters().iterator();
			while (its.hasNext()) {
				InSlotParameter params = its.next();
				OutSlotResult paramd = itd.next();
				if (params.hasValue(imperative)) {
					paramd.setValue(params.getValue());
				} else {
					throw new IllegalArgumentException(Ariane.lang.getString("Operator.error1")+params.getRank());
				}
			}

			// Gets and checks mask image (-> executes forward operators if needed).
			// Note: Mask is optional.
			InSlotMask inMasks = getInSlotMask();
			OutSlotImage inMaskd = (OutSlotImage)virtualOperator.getVirtualInSlotMask();
			if (inMasks.isConnected() && inMasks.hasValue(false)) {
				inMaskd.setValue(inMasks.getValue());
			}

			// Gets and checks input images (-> executes forward operators if needed).
			Iterator<InSlotImage> its1  = getInSlotImages().iterator();
			Iterator<OutSlotImage> itd1 = virtualOperator.getVirtualInSlotImages().iterator();
			while (its1.hasNext()) {
				InSlotImage inImgs = its1.next();
				OutSlotImage inImgd = itd1.next();
				if (inImgs.hasValue(imperative)) {
					inImgd.setValue(inImgs.getValue());
				} else {
					throw new IllegalArgumentException(Ariane.lang.getString("Operator.error3")+inImgs.getRank());
				}
			}

			changeStatus(Status.RUNNING);

			// Sets the output images (executes the subgraph if needed).
			Iterator<OutSlotImage> its2 = getOutSlotImages().iterator();
			Iterator<InSlotImage> itd2 = virtualOperator.getVirtualOutSlotImages().iterator();
			while (its2.hasNext()){
				OutSlotImage outImgs = its2.next();
				InSlotImage outImgd = itd2.next();
				if (outImgd.hasValue(imperative)) {
					outImgs.setValue(outImgd.getValue());
				}
			}

			// Sets the result.
			OutSlotResult results = getOutSlotResult();
			InSlotParameter resultd = (InSlotParameter)virtualOperator.getVirtualOutSlotResult();
			if (!resultd.hasValue(imperative)) {
				throw new IllegalArgumentException(
						Ariane.lang.getString("MacroOperator.error2")
						);
			}
			results.setValue(resultd.getValue());
			changeStatus(Status.EXECUTED);
			return true;
		} catch (IllegalArgumentException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
		} catch (Exception e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(getAbsoluteName() +"("+getId()+") : "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
		}
		return false;
	}

	///// 
	///// P R O P E R T I E S
	/////

	/**
	 * Asks interactively the prototype of the Macro.
	 */
	@Override
	protected final void askProperties() {
		askProperties(getName(), getInSlotImages().size(), getInSlotParameters().size(), getOutSlotImages().size());
	}
}