/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.item;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.etc.PropertyViewer;
import fr.greyc.pantheon.ariane.graph.slot.InSlotImage;
import fr.greyc.pantheon.ariane.graph.slot.OutSlotImage;
import fr.greyc.pantheon.ariane.ui.Workspace;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Component;
import java.awt.HeadlessException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

/**
 * A <code>saverItem</code> manages saving
 * of a data file in a graph.
 * 
 * @version Régis CLOUARD - Dec 09, 2007 (created from ImageFile)
 */
public final class SaverItem extends DataFile {
	/** The tag name of the item. */
	public static final String TAGNAME = "Saver";
	/** The button for loading a new file.  */
	private JButton _jbuttonSelect;
	/** The component that holds the current name. */
	private JTextField _nameEditor = new JTextField(getNickName());
	/** The component that holds the current name. */
	private JTextField _binaryEditor = new JTextField(getBinary());

	private static StaticSaveFileChooser _fileChooser = new StaticSaveFileChooser();

	/**
	 * Creates a <code>SaverItem</code> on the specified graph
	 * with the specified properties.
	 * Initializes the mouse listeners, the popup menu, and
	 * display it in the graph.
	 * 
	 * @param treeItem the source tree item.
	 * @param graph the graph that hold the item.
	 * @param name the name of the item.
	 * @param identifier the identification number of the item (or -1).
	 * @param category the category name of the operator.
	 */
	public SaverItem( TreeItem treeItem, Graph graph, String name, int identifier, String category ) {
		super(treeItem, graph, name, identifier, null, category,"saver");
		inSlotImages.add(new InSlotImage(this,0));
		outSlotImages.add(new OutSlotImage(this,0));
		draw();
	}

	/**
	 * Creates a <code>SaverItem</code> on the specified graph
	 * with the given XML data.
	 * 
	 * @param graph the graph that holds the item.
	 * @param xmlTree an XML tree of data.
	 * @throws IOException XML errors.
	 */
	public SaverItem( Graph graph,LinkedList<XMLElement> xmlTree ) throws IOException {
		super(graph, xmlTree,"saver");
		// For compatibility reasons
		if (outSlotImages.size() == 0)
			outSlotImages.add(new OutSlotImage(this, 0));
		draw();
	}

	/**
	 * Returns the address of the related documentation HTML page.
	 * 
	 * @return the complete or relative address.
	 * Kept for sake of compatibility.
	 */
	@Override
	public String getDocURL( ) {
		if (_treeItem!=null)
			return _treeItem.getDocURL();
		return "saver.html";
	}

	/**
	 * Returns the tag name.
	 * 
	 * @return a string.
	 */
	@Override
	public String tagName( ) {
		return TAGNAME;
	}

	/**
	 * Sets the name of the related physical file.
	 * If the name of the file is specified as a relative path 
	 * than the absolute name is built with the directory of the project.
	 * 
	 * @param binary a file name (absolute or relative)
	 */
	@Override
	public void setBinary( String binary ) {
		if (binary==null || binary.length()==0) {
			super.setBinary("");
		} else {
			String current_directory = null;
			if (getGraph().getRoot().getFileName()!=null) {
				current_directory = getGraph().getRoot().getFileName().getParent();
			}
			super.setBinary(DataFile.setPathToFilename(binary, current_directory));
		}
	}

	/**
	 * Returns the name of the file either with relative path if the file is
	 * in the filePath directory or with an absolute path otherwise.
	 * 
	 * @param directory a home directory.
	 * @return the file name.
	 */
	@Override
	public String getBinary( String directory ) {
		return getRelativePath(super.getBinary(), directory);
	}

	///// 
	///// E X E C U T I O N
	/////

	/**
	 * Executes a Saver. It means: sets the input file into
	 * the related physical file.
	 * 
	 * @param imperative is FAILURE a fatal error?
	 * @return true if the operator has been executed without errors.
	 */
	@Override
	protected final boolean primitiveExecute( boolean imperative ) {
		try {
			_currentFilename = null;
			InSlotImage isi = (inSlotImages.getFirst());
			if (isi.hasValue(imperative)) {
				if (getBinary() == null || getBinary().length()==0) {
					// No output filename specified
					throw new MissingInputException(Ariane.ErrorCode.MISSINGFILE, Ariane.lang.getString("ImageFile.error7"));
				} else {
					// An output filename is specified
					copy(isi.getValue(),getBinary());
					String exec = Ariane.lang.getString("ImageFile.save")+ ": " + getBinary();
					Ariane.workspace.printSystemMessage(exec);
					_currentFilename = getBinary();
				}
				(outSlotImages.getFirst()).setValue(_currentFilename);
			} else
				throw new MissingInputException(Ariane.ErrorCode.MISSINGINPUT, Ariane.lang.getString("ImageFile.error5"));
		} catch (MissingInputException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(e.errorCode, getAbsoluteName() +"("+getId()+"): "+ e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		} catch (IOException e) {
			if (imperative && !Graph.isExecutionStopped()) {
				Ariane.workspace.printErrorMessage(Ariane.ErrorCode.INPUTOUTPUT, getAbsoluteName() +"("+getId()+"): "+Ariane.lang.getString("ImageFile.error6")+e.getLocalizedMessage());
				getGraph().stopExecution();
				changeStatus(Status.ERROR);
			}
			return false;
		}

		// Validate the result.
		changeStatus(Status.EXECUTED);
		return true;
	}

	///// 
	///// M E N U 
	/////

	/**
	 * Builds the contextual menu as a Popup Menu.
	 * 
	 * @return a convenient popup menu.
	 */
	@Override
	protected JPopupMenu initPopup( ) {
		JPopupMenu popupMenu = new JPopupMenu();
		JMenuItem item;

		item= new JMenuItem(Ariane.lang.getString("PopupMenu.execute"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { getGraph().execute();}
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.reset"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { resetAll(); }
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_MASK));
		popupMenu.add(item);

		item = new JMenuItem(Ariane.lang.getString("PopupMenu.remove"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { delete(); }
		});
		item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		popupMenu.add(item);

		popupMenu.addSeparator();
		item = new JMenuItem(Ariane.lang.getString("PopupMenu.selectFile"));
		item.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) { selectFile();}
		});
		popupMenu.add(item);

		return popupMenu;
	}

	/**
	 * Opens a windows from which user can type a file name.
	 * Keeps the directory as the current directory.
	 */
	private void selectFile( ) {
		String path;

		if (_currentWorkingDir == null)
			_currentWorkingDir = Workspace.getCurrentDir();
		_fileChooser.setCurrentDirectory(new File(_currentWorkingDir));
		final String currentFile = getBinary();
		if ( currentFile != null && currentFile.trim().length()>0) {
			_fileChooser .setSelectedFile(new File(currentFile));
		}

		if (_fileChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
			try {
				path = ""+_fileChooser.getSelectedFile();
				_currentFilename=null;
				_currentWorkingDir = _fileChooser.getSelectedFile().getParent();
				changeBinaryName(_nameEditor, _binaryEditor, _jbuttonSelect, path, true);
			} catch (Exception e) {
				Ariane.workspace.printErrorMessage(Ariane.lang.getString("ImageFile.error1") + e);
			}
		}
	}

	/**
	 * @return the list of specific properties (other than common properties).
	 */
	@Override
	public ArrayList<PropertyViewer.PropertyCouple> getSpecificProperties( ) {
		ArrayList<PropertyViewer.PropertyCouple> properties = new ArrayList<PropertyViewer.PropertyCouple>();

		String name = "";
		if (getBinary()!=null) {
			File file = new File(getBinary());
			name = file.getName();
		}

		_jbuttonSelect = new JButton();
		if (name.length()==0)
			_jbuttonSelect.setText(Ariane.lang.getString("PropertyViewer.saveas")+"...");
		else 
			_jbuttonSelect.setText(name);
		_jbuttonSelect.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				selectFile();
				_jbuttonSelect.setText(getNickName());
			}
		});

		properties.add(new PropertyViewer.PropertyCouple(Ariane.lang.getString("PropertyViewer.path"), _jbuttonSelect));
		return properties;
	}

	/**
	 * Invoked when the mouse button has been
	 * double clicked (pressed and released) on a graph item.
	 * Shows the dialog box for setting the item's properties.
	 * 
	 * @param e the event attributes.
	 */
	@Override
	protected void onMouseDoubleClicked( MouseEvent e ) {
		_jbuttonSelect.doClick();
	}

	/**
	 * @return a component appropriate to edit
	 *               and possibly change the name of the item.
	 */
	@Override
	public JComponent getNameEditor() {
		_nameEditor = new JTextField(getNickName());
		_nameEditor.setEnabled(false);
		_nameEditor.setEditable(false);
		return _nameEditor;
	}

	/**
	 * @return a component appropriate to edit
	 *               and possibly change the name of the item.
	 */
	@Override
	public JComponent getBinaryEditor() {
		return null;
	}

	/**
	 * Defined to keep the size and location of the 
	 * previous opening of the JFileChooser;
	 */
	private static class StaticSaveFileChooser extends JFileChooser {
		private JDialog _dialog;
		public StaticSaveFileChooser() {
			setDialogTitle(Ariane.lang.getString("FileChooser.titleSaveFile"));
			setDialogType(JFileChooser.SAVE_DIALOG);
		}
		@Override
		protected JDialog createDialog( Component parent ) throws HeadlessException {
			JDialog dialog = super.createDialog(parent);
			if (_dialog!=null) {
				dialog.setLocation(_dialog.getLocation());
				dialog.setSize(_dialog.getSize());
			}
			_dialog=dialog;
			return dialog;
		}
	}
}