/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.command.ConnectItemEdit;
import fr.greyc.pantheon.ariane.graph.command.DisconnectItemEdit;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.VirtualItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragSource;
import java.awt.dnd.DropTargetDropEvent;
import java.io.IOException;

/**
 * The class <code>InSlot</code> defines input connection slot.
 * Only one <code>OutSlot</code> can be connected to one
 * <code>InSlot</code>. It is a drop site for an <code>OutSlot</code>.
 * Three types of input slots are available:
 * <ul>
 * <li>@see InSlotImage
 * <li>@see InSlotMask
 * <li>@see InSlotParameter
 * </ul>
 * 
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Frédéric LAIR - May 23, 2007
 * @version Régis CLOUARD - 2007-11-24 (restructuration)
 */
public abstract class InSlot extends Slot {
	/** The connected output slot. */
	private transient OutSlot _connectedSlot = null;
	/** The DataFlavor as a Transferable object. */
	public static final DataFlavor[] flavors ={ new DataFlavor("application/x-java-serialized-object-in-slot", "InSlot Flavor")};
	/** The link between this and the connected slot. */
	private transient Link _link = null;

	/**
	 * Creates an input slot on the given graph item.
	 * 
	 * @param graph the graph item that will hold the slot.
	 * @param rank the rank in the list of input slot of the graph item.
	 */
	public InSlot( GraphItem graph, int rank ) {
		super(graph, rank);
		initDragSource();
	}

	/**
	 * @return the link of the slot.
	 */
	public final Link getLink( ) {
		return _link;
	}

	/**
	 * Sets the link for representing the connection.
	 * 
	 * @param l a link.
	 */
	protected final void setLink( Link l ) {
		_link = l;
	}

	///// 
	///// D R A G   a n d   D R O P
	/////

	/**
	 * Initializes the slot as a draggable component.
	 */
	private final void initDragSource( ) {
		dragSource = DragSource.getDefaultDragSource();
		dragSource.createDefaultDragGestureRecognizer(this, DnDConstants.ACTION_MOVE, this);
	}

	/**
	 * Gets the transfered object.
	 * 
	 * @return the list of recognized flavors.
	 */
	@Override
	public final DataFlavor[] getTransferDataFlavors( ) {
		return flavors;
	}

	/** 
	 * Is the object the transferable object?
	 * 
	 * @param flavor the flavor to be tested.
	 * @return true if the flavor is transferable.
	 */
	@Override
	public final boolean isDataFlavorSupported( DataFlavor flavor ) { 
		return flavor.equals(flavors[0]);
	}

	/**
	 * Gets the transfered data.
	 * 
	 * @param flavor the flavor of the transfered object.
	 * @return the data.
	 * @throws UnsupportedFlavorException unrecognized flavors.
	 * @throws IOException in case of IO errors.
	 */
	@Override
	public final Object getTransferData( DataFlavor flavor ) throws UnsupportedFlavorException, IOException {
		if (isDataFlavorSupported(flavor))
			return this;
		else
			throw new UnsupportedFlavorException(flavor);
	}

	/**
	 * The DropTarget intercepts drop() notifications before the
	 * registered DropTargetListener gets them.
	 * Connects the two slots if the connection is possible:
	 * <ul>
	 * <li>drag and drop slots are different ;
	 * <li>slots are compatibles.
	 * </ul>
	 * 
	 * @param e the DropTargetDropEvent.
	 */
	@Override
	public final void drop( DropTargetDropEvent e ) {
		setNormalBorder();
		try {
			if (e.isDataFlavorSupported(OutSlot.flavors[0])) {// Come from an OutSlot?
				OutSlot outSlot = (OutSlot)(e.getTransferable().getTransferData(OutSlot.flavors[0]));
				outSlot.setNormalBorder();
				boolean connect = false;
				// check if the slot are in the same graph!
				if (getGraph() != (outSlot).getGraph())
					throw new Exception(Ariane.lang.getString("Slot.error3"));
				// check slots consistency.
				if (((this instanceof InSlotParameter) && (outSlot instanceof OutSlotResult))
						||
						((this instanceof InSlotImage) && (outSlot instanceof OutSlotImage))
						||
						((this instanceof InSlotMask) && (outSlot instanceof OutSlotImage))) {
					connect = true;
				}
				if (connect) {
					getGraph().startActionEdition();
					outSlot.connectTo(this,true);
					getGraph().endActionEdition();
					getGraph().unSelectAll();
					getGraph().requestFocusInWindow();
					e.acceptDrop(DnDConstants.ACTION_MOVE);
					e.dropComplete(true);
				} else
					throw new Exception(Ariane.lang.getString("Slot.error2"));
			} else
				throw new Exception(Ariane.lang.getString("Slot.error2"));
		} catch (UnsupportedFlavorException ex) {
			e.rejectDrop();
			e.dropComplete(false);
//			ex.printStackTrace();
			Ariane.workspace.printErrorMessage(Ariane.lang.getString("Slot.error1")+ ex.getLocalizedMessage());
		} catch (Exception ex) {
			e.rejectDrop();
//			ex.printStackTrace();
			e.dropComplete(false);
			Ariane.workspace.printErrorMessage(Ariane.lang.getString("Slot.error1")+ ex.getLocalizedMessage());
		}
	}

	///// 
	///// C O N N E C T I O N   /   D I S C O N N E C T I O N 
	/////

	/**
	 * Returns the connected slot.
	 * 
	 * @return an OutSlot or null.
	 */
	public final OutSlot getConnectedSlot( ) {
		return _connectedSlot;
	}

	/**
	 * Tests if the slot is connected.
	 * 
	 * @return true if the slot is connected.
	 */
	@Override
	public final boolean isConnected( ) {
		return (_connectedSlot != null);
	}

	/**
	 * @return a new link.
	 */
	abstract protected Link createLink();

	/**
	 * Connects the slot with the given slot,
	 * or disconnects the slot if the given slot is null.
	 * Resets the related graph item.
	 * 
	 * @param slot the slot to be connected or null.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	protected final void setConnectedSlot( OutSlot slot , boolean historical ) {
		getGraphItem().resetAll();
		getGraphItem().getGraph().setModified(true);
		if (slot == null) { // case of disconnection.
			setIcon(getDisconnectedIcon());
			if (historical) {
				getGraph().validateAction(new DisconnectItemEdit(getGraph(),this,_connectedSlot));
			}
			_link.dispose();
		} else { // case of connection
			setIcon(getConnectedIcon());
			_link = createLink();
			if (historical) {
				getGraph().validateAction(new ConnectItemEdit(getGraph(),this,slot));
			}
		}
		_connectedSlot = slot;
	}

	/**
	 * Disconnects the slot.
	 * 'passive' method : calls the disconnection method
	 * of the connected slot.
	 */
	public final void disconnect( ) {
		if (_connectedSlot != null) {
			_connectedSlot.disconnect(this,true);
			disposeLink();
		}
	}

	/**
	 * Connects the slot with the given slot.
	 *<br>
	 * 'Passive' method. First of all, disconnects the slot if needed.
	 * Then calls the connect method of the given slot.
	 * 
	 * @param outSlot the slot to be connected.
	 */
	public final void connectFrom( OutSlot outSlot ) {
		if (outSlot != _connectedSlot) {
			disconnect();
			outSlot.connectTo(this,true);
		}
	}

	/**
	 * Disposes the link.
	 */
	public void disposeLink( ) { }

	/**
	 * Saves the slot and its values as an XML tree.
	 * 
	 * @return the root of the XML tree.
	 * @throws IOException in case of output errors.
	 */
	public abstract XMLElement save( ) throws IOException;

	/**
	 * Returns the current color of that link type.
	 * 
	 * @return the user defined color.
	 */
	protected abstract Color getLinkColor( );

	/**
	 * Draws the link between the slot and its connected slot.
	 * (Try a sophisticated algorithm).
	 * 
	 * @param gc the graphic context.
	 */
	public final void drawLink( Graphics gc ) {
		if (isConnected()) {
			_link.draw(gc,getLinkColor(), isSelected());
		}
	}

	/**
	 * Selects the slot and the connected slot.
	 * @param isSelected true if it is selected, false otherwise.
	 */
	@Override
	public final void setSelected( boolean isSelected ) {
		if (getGraphItem() instanceof VirtualItem) return;
		if (_connectedSlot!=null) {
			_connectedSlot.setSelected(isSelected);
			super.setSelected(isSelected);
		}
	}
}