/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.graph.slot;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.command.ChangeParameterValueEdit;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.ui.PreferenceChooser;
import fr.greyc.pantheon.ariane.xml.XMLElement;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.io.IOException;

import javax.swing.ImageIcon;
import javax.swing.JTextField;

/**
 * The class <code>InSlotParameter</code> defines an input parameter connection slot.
 * An input parameter slot can be connected to a OutSlotResult
 * or can hold its own value (either float or String value).
 *
 * @author Régis CLOUARD - Oct 01, 1998
 * @version Sébastien LEFÈVRE, Richard LOBRÉAU - Feb 21, 1999
 * @version Adrien SCHRÈQUE, Hicham AMADI - Nov 05, 2000
 * @version Mathieu JOETS - Jul 09, 2001
 * @version Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 25, 2004
 * @version Frédéric LAIR - May 23, 2007
 * @version Régis Clouard - Dec 23, 2012 (add parameter name)
 */
public final class InSlotParameter extends InSlot {
	/** The value of the parameter (float or string values are available). */
	private String _value;
	private String _name;

	/** The visual representation of the value. */
	private transient JTextField _textfield;

	/** The connected rendering of the slot. */
	private transient static ImageIcon _connectedIcon;

	/** The disconnected rendering of the slot. */
	private transient static ImageIcon _disconnectedIcon;

	/**
	 * Creates an input parameter slot on the given graph item.
	 * 
	 * @param g the graph item that will hold the slot.
	 * @param rank the rank in the list of input slot of the graph item.
	 */
	public InSlotParameter( GraphItem g, int rank ) {
		super(g, rank);
		_textfield= new JTextField(getValue(), 10);
		_textfield.setEnabled(true);
		_textfield.setEditable(true);
		_textfield.addFocusListener(new FocusListener() {
			@Override
			public void focusLost(FocusEvent e){ changeValue(_textfield.getText(), true); }
			@Override
			public void focusGained(FocusEvent e){ }
		});

		// Load connected and disconnected icons.
		if (_connectedIcon == null) {
			ClassLoader classLoader = getClass().getClassLoader();
			_disconnectedIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"SlotValueOff.png"));
			_connectedIcon = new ImageIcon(classLoader.getResource(Ariane.iconDir+"SlotValueOn.png"));
		}
		setIcon(_disconnectedIcon);
		dropTarget = new DropTarget(this, DnDConstants.ACTION_LINK, this, true);
	}

	/**
	 * Creates an new link for a new connection.
	 * @return a new link.
	 */
	@Override
	protected Link createLink() {
		_textfield.setEnabled(false);
		_textfield.setEditable(false);
		return new ValueLink(this);
	}

	/**
	 * Disposes the link after a disconnection.
	 */
	@Override
	public final void disposeLink( ) {
		_value = null;
		_textfield.setEnabled(true);
		_textfield.setEditable(true);
		_textfield.setText(_value);
	}

	/**
	 * Returns a textual description of the slot.
	 * for debugging purpose only.
	 * 
	 * @return a string.
	 */
	@Override
	public String toString( ) {
		return "Parameter $"+getRank();
	}

	/**
	 * Returns the icon for connected rendering.
	 * 
	 * @return an icon.
	 */
	@Override
	protected ImageIcon getConnectedIcon( ) {
		return _connectedIcon;
	}

	/**
	 * Returns the icon for disconnected rendering.
	 * 
	 * @return	an icon.
	 */
	@Override
	protected ImageIcon getDisconnectedIcon( ) {
		return _disconnectedIcon;
	}

	///// 
	/////  L I N K
	/////

	/**
	 * Overrides printing to display a string in the related
	 * graph item.
	 * 
	 * @param gc the graphic context.
	 */
	@Override
	public final void printChildren( Graphics gc ) {
		if (isConnected()) {
			super.printChildren(gc);
		} else {
			gc.setFont(PreferenceChooser.fontDialogPLAIN6);
			Rectangle r=getBounds();
			gc.setColor(new Color(255,255,255));
			gc.fillRect((int)r.getX(),(int)r.getY(),(int)r.getWidth(), (int)r.getHeight()) ;
			gc.setColor(new Color(0,0,0));
			gc.drawString(toString(), (int)r.getX(),(int)(r.getY()+(2+r.getHeight()/2.0)));
		}
	}

	/**
	 * Returns the anchor coordinates of the slot.
	 * 
	 * @return anchor point.
	 */
	@Override
	public final Point getAnchor( ) {
		return( new Point((int)(getLocation().getX() + getWidth()/2+ getGraphItem().getLocation().getX() ), 
				(int)(getLocation().getY() +getGraphItem().getLocation().getY() - 3)));
	}

	/**
	 * Returns the current color of that link type.
	 * 
	 * @return the user defined color.
	 */
	@Override
	protected Color getLinkColor( ) {
		return Graph.getValueLinkColor();
	} 

	///// 
	/////  V A L U E 
	/////

	/**
	 * Tests if the parameter has a validate value.
	 * If it is connected, tests if the connected slot
	 * has a validate value.
	 * 
	 * @param imperative is this value really necessary?
	 * @return true if the parameter has a validate value.
	 */
	public final boolean hasValue( boolean imperative ) {
		// The parameter is connected.
		if (isConnected()) {
			// Gets the value of the connected slot.
			OutSlotResult res = (OutSlotResult)getConnectedSlot();
			if (res.hasValue(imperative)) {
				setValue(res.getValue());
			} else {
				return false;
			}
		}
		// The parameter is not connected, tests the value.
		return (_value != null);
	}

	/**
	 * Tests if the parameter has a float value and is validated.
	 * If it is connected tests if the connected slot
	 * has a float value and is validated.
	 * 
	 * @param imperative is this value really necessary?
	 * @return true if the parameter has a validate value and is a float.
	 */
	public final boolean hasNumericalValue( boolean imperative ) {
		if (hasValue(imperative)) {
			try {
				new Float(getValue()).floatValue();
				return true;
			} catch (Exception ex) {
				return false;
			}
		} else
			return false;
	}

	@Override
	public void setName( String name ) {
		_name = name;
	}

	@Override
	public String getName( ) {
		if (_name != null) {
			return _name;
		} else {
			return Ariane.lang.getString("PropertyViewer.parameter") + " $" +getRank();
		}
	}

	public boolean hasName() {
		return _name!=null;
	}

	/**
	 * Sets the value of the parameter.
	 * 
	 * @param value a value.
	 */
	public final void setValue( String value ) {
		if ((value!= null) && (value.length() > 0)) {
			_value = value.trim();
		} else {
			_value = null;
		}
		_textfield.setText(_value);
	}

	/**
	 * Returns the value as a String value.
	 * 
	 * @return the value.
	 */
	public final String getValue( ) {
		return _value;
	}

	/**
	 * Returns the value as a float value.
	 * 
	 * @return the value.
	 */
	public final float getNumericalValue( ) { 
		try {
			float x = new Float(getValue()).floatValue();
			return x;
		} catch (Exception ex) {
			Ariane.workspace.printErrorMessage(Ariane.lang.getString("Slot.error4"));
			return 0.0f;
		}
	}

	/**
	 * Changes the current value of the slot.
	 * 
	 * @param value the new value.
	 * @param historical true if the action must be stored in the action
	 *        history (for use with undo/redo).
	 */
	public void changeValue( String value, boolean historical ) {
		if (value==null && getValue()==null) return;
		if ((value==null || !value.equals(getValue()))) {
			String oldValue = getValue();
			setValue(value);
			getGraphItem().resetAll();
			getGraphItem().updatePropertyViewer();
			getGraph().setModified(true);
			if (historical) {
				getGraph().startActionEdition();
				getGraph().validateAction(new ChangeParameterValueEdit(this, value, oldValue));
				getGraph().endActionEdition();
			}
		}
	}

	/**
	 * Creates the textField that manages the parameter value.
	 * 
	 * @return the JTextField.
	 */
	public final JTextField getParameterEditor( ) {
		return _textfield;
	}

	/**
	 * Saves the slot with its values as an XML tree.
	 * 
	 * @return the root of the XML tree.
	 * @throws IOException XML problem.
	 */
	@Override
	public final XMLElement save( ) throws IOException {
		XMLElement e = new XMLElement("Connection");
		e.addChildren(new XMLElement("EndItemId",Integer.toString(getGraphItem().getId())));
		e.addChildren(new XMLElement("EndSlotType","InSlotParameter"));
		e.addChildren(new XMLElement("EndSlotIndex",Integer.toString(getRank())));
		if (isConnected()) {
			getConnectedSlot().save(e);
			if (getLink().getPath().size() != 0) {
				e.addChildren(new XMLElement("UserDefined", ""+getLink().isUserDefined()));
				XMLElement pa = new XMLElement("LinkPath");
				for (int i = 0; i<getLink().getPath().size(); i++) {
					XMLElement po = new XMLElement("LinkPoint");
					po.addChildren(new XMLElement("IdPoint",Integer.toString(i)));
					po.addChildren(new XMLElement("X-coordinate", Integer.toString((int)getLink().getPath().get(i).getX())));
					po.addChildren(new XMLElement("Y-coordinate", Integer.toString((int)getLink().getPath().get(i).getY())));
					pa.addChildren(po);
				}
				e.addChildren(pa);
			}
		} else {
			e.addChildren(new XMLElement("ParameterValue", getValue()));
		}
		return e;
	}
}