/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.utils.Tree;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JToolBar;
import javax.swing.KeyStroke;
import javax.swing.border.EtchedBorder;

/**
 * An explorer manages an operators tree,
 * from where user can select items to compose
 * a graph.
 * It also provides a toolbar for custom facilities (search, documentation).
 * 
 * @version Régis CLOUARD - Dec 02, 2007
 * @version Régis CLOUARD - Jan 01, 2011
 */
public class Explorer extends JPanel implements ILocalizable {
	/** The tree of items. */
	private Tree _tree;

	/**
	 * Creates the explorer with the specified tree.
	 * 
	 * @param tree a tree.
	 */
	public Explorer( Tree tree ) {
		super(new BorderLayout());
		_tree = tree;
		add(createToolbar(),BorderLayout.NORTH);
		add(new JScrollPane(tree),BorderLayout.CENTER);
	}

	/** The search action. */
	private Action _searchA;
	/** The open documentation action. */
	private Action _indexA;

	/** The search button */
	private JButton _searchB;
	/** The index button. */
	private JButton _indexB;

	/** The search menu items */
	private JMenuItem _searchI;
	/** The index menu item. */
	private JMenuItem _indexI;

	/** The explorer menu. */
	private JMenu _explorerMenu;

	/**
	 * Creates the related toolbar.
	 * 
	 * @return the toolbar.
	 */
	private JToolBar createToolbar() {
		JToolBar toolbar = new JToolBar();
		toolbar.setFloatable(false);
		toolbar.setBorder(new EtchedBorder(EtchedBorder.LOWERED));
		toolbar.setBorderPainted(true);
		toolbar.setOpaque(false);

		ClassLoader classLoader= getClass().getClassLoader();

		_searchA = new AbstractAction(null, new ImageIcon(classLoader.getResource(Ariane.iconDir+"search.png")))
		{ @Override
			public void actionPerformed(ActionEvent e) { searchAction(); } };
		_indexA = new AbstractAction(null, new ImageIcon(classLoader.getResource(Ariane.iconDir+"index.png")))
		{ @Override
			public void actionPerformed(ActionEvent e) { indexAction(); } };

		_searchB=new JButton(_searchA);
		_searchB.setBorderPainted(false);
		toolbar.add(_searchB);

		_indexB=new JButton(_indexA);
		_indexB.setBorderPainted(false);
		toolbar.add(_indexB);

		return toolbar;
	}

	/**
	 * Creates the related menu.
	 * 
	 * @return the customized menu.
	 */
	public JMenu getMenu( ) {
		_explorerMenu = new JMenu();
		_explorerMenu.setMnemonic('R');

		_searchI = new JMenuItem(_searchA);
		_searchI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F, KeyEvent.CTRL_MASK));
		_explorerMenu.add(_searchI);

		_indexI = new JMenuItem(_indexA);
		_indexI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_I, KeyEvent.CTRL_MASK));
		_explorerMenu.add(_indexI);

		return _explorerMenu;
	}

	/**
	 * Sets the new text string for all components of the explorer
	 * according to the current language.
	 */
	@Override
	public void setInternationalText( ) {
		_searchI.setText(Ariane.lang.getString("Menu.explorer.search"));
		_searchB.setToolTipText(Ariane.lang.getString("Menu.explorer.search"));
		_indexI.setText(Ariane.lang.getString("Menu.explorer.index"));
		_indexB.setToolTipText(Ariane.lang.getString("Menu.explorer.index"));
		_explorerMenu.setText(Ariane.lang.getString("Menu.explorer.title"));
	}

	/**
	 * Searches operators in the explorer.
	 * Highlights all operators found in the tree.
	 */
	private void searchAction( ) {
		new SearchDialog(_tree).setVisible(true);
	}

	/**
	 * Displays the index page of the library manual.
	 */
	private void indexAction( ) {
		Ariane.workspace.displayMan("Index",_tree.getIndexDoc(), null);
	}
}