/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Rectangle;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetAdapter;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.Hashtable;

import javax.swing.AbstractAction;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.plaf.basic.BasicTabbedPaneUI;

import fr.greyc.pantheon.ariane.graph.Graph;

/**
 * The <code>GraphEditor</code> is the editor of graph.
 * Each graph is displayed in a <code>JTabbedPane</code>.
 * Beware of the fact that each graph is first for all inserted
 * in a JScrollPane.
 * 
 * @author Régis CLOUARD - Dec 13, 2004
 * @version Régis CLOUARD - Dec 01, 2007 (change JInternalFrame to JTabbedPane) 
 */
public final class GraphEditor extends JTabbedPane implements ChangeListener {
	/** The workspace that support the graph. */
	private final Workspace _workspace;
	/** The index table of current graphs. */
	private final Hashtable<Component, Graph> _indexTable = new Hashtable<Component, Graph>();

	/**
	 * Creates the tabbed pane manager.
	 * 
	 * @param workspace the workspace that supports the graph editor.
	 */
	public GraphEditor( Workspace workspace) {
		_workspace = workspace ;
		// when all the tabs will not fit within a single run -> scrolling.
		setUI(new CloseableTabbedPaneUI());
		setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
		addChangeListener(this);
		Font font = getFont();
		setFont(font.deriveFont(Font.PLAIN,11f));

		new DropTarget(this, new DropTargetAdapter() {
			@Override
			public final void dragEnter( DropTargetDragEvent e ) {
				e.acceptDrag(DnDConstants.ACTION_MOVE);
			}

			@Override
			public final void drop( DropTargetDropEvent e ) {
				if (getTabCount()==0) {
					_workspace.newGraph().getDropTarget().drop(e);
				} else {
					e.rejectDrop();
				}
			}
		});
	}

	/**
	 * Returns the JScrollPane component that holds the
	 * specified graph or null if any.
	 * 
	 * @param graph the graph.
	 * @return the JScrollPane that holds the graph.
	 */
	private JScrollPane getScrollPane( Graph graph ) {
		for (int i = 0; i < getTabCount(); i++) {
			JScrollPane scrollpane = (JScrollPane)getComponentAt(i);
			if (_indexTable.get(scrollpane) == graph) {
				return scrollpane;
			}
		}
		return null;
	}

	/**
	 * Returns the graph supported by the given scrollpane.
	 * 
	 * @param scrollpane a scrollpane.
	 * @return the graph or null if any.
	 */
	private Graph getGraph( Component scrollpane ) {
		if (scrollpane == null) { return null; }
		return _indexTable.get(scrollpane);
	}

	/**
	 * Returns the graph supported by the nth scrollpane
	 * in the JTabbedPane.
	 * 
	 * @param index the index of the scrollpane.
	 * @return the graph or null if any.
	 */
	private Graph getGraphAt( int index ) {
		if (index<0) { return null; }
		return getGraph(getComponentAt(index));
	}

	/**
	 * Checks whether the specified graph already exists.
	 * If the graph already exists, puts it in front.
	 * 
	 * @param filename a regular file name.
	 * @return true if the graph already exists.
	 */
	public boolean exists( String filename ) {
		for (int i = 0; i < getTabCount(); i++) {
			JScrollPane scrollpane = (JScrollPane)getComponentAt(i);
			Graph graph = _indexTable.get(scrollpane);
			if (graph.getFileName()!=null && graph.getFileName().equals(new File(filename))) {
				super.setSelectedComponent(scrollpane);
				_workspace.setCurrentGraph(graph);
				return true;
			}
		}
		return false;
	}

	private String buildRootName( String name ) {
		return "<html><b><font color=#EFC700>"+name+"</font></b></html>";
	}

	/**
	 * Adds a new frame in the editor.
	 * 
	 * @param graph the graph to be added.
	 */
	public void add( Graph graph ) {
		JScrollPane scrollpane = getScrollPane(graph);
		if ( scrollpane == null ) {
			scrollpane = new JScrollPane(graph);
			// Discard the arrows events.
			scrollpane.setInputMap(WHEN_IN_FOCUSED_WINDOW, null); 
			scrollpane.setInputMap(WHEN_FOCUSED, null); 
			scrollpane.setInputMap(WHEN_ANCESTOR_OF_FOCUSED_COMPONENT , null); 
			scrollpane.getVerticalScrollBar().setUnitIncrement(10);
			scrollpane.getHorizontalScrollBar().setUnitIncrement(10);
			scrollpane.unregisterKeyboardAction(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0));
			_indexTable.put(scrollpane,graph);
			if (graph.isRoot()) {
				super.addTab(buildRootName(graph.getName()),scrollpane);
			} else {
				super.addTab(graph.getName(),scrollpane);
			}
			super.setSelectedComponent(scrollpane);
			_workspace.setCurrentGraph(graph);
		} else {
			super.setSelectedComponent(scrollpane);
			_workspace.setCurrentGraph(graph);
		}
	}

	/**
	 * Removes the specified graph from the editor.
	 * 
	 * @param graph the graph to be removed.
	 */
	public void remove( Graph graph ) {
		JScrollPane scrollpane = getScrollPane(graph);
		if (scrollpane!=null) {
			super.remove(scrollpane);
			_indexTable.remove(scrollpane);
			_workspace.setCurrentGraph(getGraph(getSelectedComponent()));
		}
	}

	/**
	 * Closes the specified graph from the editor.
	 * 
	 * @param index the index of the graph in the editor.
	 */
	protected void close( int index ) {
		Graph graph = getGraphAt(index);
		graph.close();
	}

	/**
	 * Sets the title to the given graph.
	 * 
	 * @param graph a graph.
	 * @param name the new name.
	 */
	public void setTitle( Graph graph, String name ) {
		if (graph == null) return;
		int index = indexOfComponent(getScrollPane(graph));
		graph.setName(name);
		if (index>=0) {
			if (graph.isRoot()) {
				setTitleAt(index,buildRootName(graph.getName()));
			} else {
				setTitleAt(index, graph.getName());
			}
		}
	}

	/**
	 * Closes all graphs.
	 * 
	 * @return true if all graph has been closed.
	 */
	public boolean closeAllGraphs( ) {
		// Closes only the root graph
		// because root graph deletes it.
		for (int i=getTabCount()-1; i>=0; i--) {
			Graph graph = getGraphAt(i);
			if (graph.isRoot()) {
				if (!graph.close()) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Invoked when the user select a new pane.
	 * 
	 * @param event a ChangeEvent object.
	 */ 
	@Override
	public void stateChanged( ChangeEvent event ) {
		JTabbedPane source = (JTabbedPane)event.getSource();
		Graph graph = getGraphAt(source.getSelectedIndex());
		_workspace.setCurrentGraph(graph); 
	}

	/**
	 * Repaints all graphs.
	 */
	public void repaintAllGraphs( ) {
		for (int i =0; i< getTabCount(); i++) {
			Graph graph = getGraphAt(i);
			graph.repaintAll();
		}
	}

	/**
	 *  Redefines the Basic L&F implementation of TabbedPaneUI
	 *  in order to create a JTabbedPane which has a close ('X')
	 *  icon on each tab.
	 *  Aligns the title from the left side of the tab.
	 */
	static class CloseableTabbedPaneUI extends BasicTabbedPaneUI {
		/**
		 * @return our tab layout manager.
		 */
		@Override
		protected LayoutManager createLayoutManager( ) {
			return new CloseableLayout();
		}

		/**
		 * Override the method in order to aligns the title
		 * from the left side of the tab.
		 */
		@Override
		protected void layoutLabel( int tabPlacement, FontMetrics metrics, int tabIndex, String title, Icon icon, Rectangle tabRect, Rectangle iconRect, Rectangle textRect, boolean isSelected ) {
			Insets defaultInsets = getTabInsets(tabPlacement,tabIndex);
			textRect.x=0;	textRect.y=0;
			iconRect.x=0;	iconRect.y=0;
			SwingUtilities.layoutCompoundLabel(tabPane,metrics,
					title, icon, SwingUtilities.CENTER, SwingUtilities.LEFT,
					SwingUtilities.CENTER,SwingUtilities.CENTER,tabRect,
					iconRect,textRect,textIconGap+2);
			// add insets
			textRect.x+=defaultInsets.left;
		}

		/**
		 * Adds enough space to insert the close box.
		 */
		@Override
		protected Insets getTabInsets(int tabPlacement,int tabIndex) {
			Insets defaultInsets = (Insets)super.getTabInsets(tabPlacement,tabIndex).clone();
			defaultInsets.left += 1;
			defaultInsets.right += 10;
			defaultInsets.top += 1;
			defaultInsets.bottom += 1;
			return defaultInsets;
		}

		/**
		 * The new layout.
		 */
		class CloseableLayout extends TabbedPaneLayout {
			/** The list of close buttons.*/
			protected final ArrayList<CloseButton> closeButtons = new ArrayList<CloseButton>();

			/**
			 * Lays out the container.
			 * 
			 * @parent the container.
			 */
			@Override
			public void layoutContainer( Container parent ) {
				super.layoutContainer(parent);
				// adds close button to new tabs.
				while (tabPane.getTabCount() > closeButtons.size()) {
					closeButtons.add(new CloseButton(closeButtons.size()));
				}
				Rectangle rect = new Rectangle();
				int i;
				for (i = 0; i < tabPane.getTabCount();i++) {
					rect = getTabBounds(i,rect);
					JButton closeButton = closeButtons.get(i);
					// sets location of the close buttons.
					closeButton.setLocation(rect.x+rect.width-18,rect.y+3);
					closeButton.setSize(15,15);
					tabPane.add(closeButton);
				}
				for (;i < closeButtons.size();i++) {
					//remove any extra close buttons
					tabPane.remove(closeButtons.get(i));
				}
			}

			/**
			 * implement UIResource so that when we add this button to the 
			 * tabbedpane, it doesn't try to make a tab for it!
			 */
			class CloseButton extends JButton implements javax.swing.plaf.UIResource {
				/**
				 * Creates a new close button.
				 * @param index the index of the button in the group.
				 */
				public CloseButton( int index ) {
					super(new CloseButtonAction(index));
					setToolTipText("Close this tab");
					setForeground(Color.BLACK);

					//remove the typical padding for the button
					setMargin(new Insets(0,0,0,0));
					addMouseListener(new MouseAdapter() {
						@Override
						public void mouseEntered(MouseEvent e) {
							setForeground(Color.RED);
						}
						@Override
						public void mouseExited(MouseEvent e) {
							setForeground(Color.BLACK);
						}
					});
				}
			}

			/**
			 * Creates the close action.
			 */
			class CloseButtonAction extends AbstractAction {
				/** The index of the button in the list of buttons.*/
				private final int _index;

				/**
				 * Closes the tab.
				 * 
				 * @param index the index of the button in the group.
				 */
				public CloseButtonAction( int index ) {
					super("x");
					_index = index;
				}

				/**
				 * Close the related tab.
				 * 
				 * @param e an event.
				 */
				@Override
				public void actionPerformed( ActionEvent e ) {
					CloseableLayout.this.closeButtons.get(_index).setForeground(Color.BLACK);
					((GraphEditor)tabPane).close(_index);
				}
			}
		}
	}
}