/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;

import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JWindow;
import javax.swing.border.EtchedBorder;

import fr.greyc.pantheon.ariane.Ariane;

/**
 * A <code>SplashScreen</code> displays an image during the initialization
 * of the software.
 * 
 * @author Régis CLOUARD - Oct 16, 2001
 * @version Régis CLOUARD - Nov 19, 2004
 * @version Régis CLOUARD - Jul 20, 2010 (add the release number)
 */
public final class SplashScreen extends JWindow {
	/** The current progressbar */
	protected ProgressBar _progressBar = new ProgressBar();

	/**
	 * Creates the <code>SplashScreen</code>.
	 * 
	 * @param iconName the image to be displayed.
	 */
	public SplashScreen( String iconName ) {
		ClassLoader classLoader= getClass().getClassLoader();
		ImageIcon splashImage = new ImageIcon(classLoader.getResource(iconName));

		JLabel splashButton = new JLabel("");
		if (splashImage != null) {
			int imgWidth = splashImage.getIconWidth();
			int imgHeight = splashImage.getIconHeight();
			Dimension scrSize = Toolkit.getDefaultToolkit().getScreenSize();
			setLocation(scrSize.width/2-imgWidth/2,scrSize.height/2-imgHeight/2);
			splashButton.setIcon(addReleaseNumberOnImage(splashImage));
		}
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		getContentPane().setLayout(new BorderLayout(0, 0));
		getContentPane().add(splashButton, BorderLayout.CENTER);
		getContentPane().add(_progressBar, BorderLayout.SOUTH);
		pack();
	}

	/**
	 * Adds the current release number onto the splash image.
	 * 
	 * @param imageIcon the splash image.
	 * @return the modified image.
	 */
	private ImageIcon addReleaseNumberOnImage ( ImageIcon imageIcon ) {
		Image image = imageIcon.getImage();
		int width = image.getWidth(null);
		int height = image.getHeight(null);
		BufferedImage bimg = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		bimg.createGraphics().drawImage(image, 0, 0, this);
		bimg.getGraphics().drawString("Version: "+Ariane.VERSION_ID,80,220);

		return new ImageIcon(bimg);
	}

	/**
	 * Displays a message in the status bar and fills the progress bar.
	 * 
	 * @param message a short message.
	 * @param percent the percentage value.
	 */
	public void showProgress( String message, int percent ) {
		_progressBar.showStatus(message);
		_progressBar.showProgress(percent);
	}

	/**
	 * A <code>ProgressBar </code> is used to reflect the
	 * progression of the initialization process.
	 */
	private final class ProgressBar extends JPanel {
		/** the current message. */
		private JLabel _msg = new JLabel();
		/** The current progress ratio. */
		private JProgressBar _progress = new JProgressBar();

		/**
		 * Creates a progress bar.
		 */
		public ProgressBar( ) {
			_progress.setMinimum(0);
			_progress.setMaximum(100);
			_progress.setMinimumSize(new Dimension(100, 20));
			_progress.setSize(new Dimension(100, 20));

			_msg.setMinimumSize(new Dimension(300, 20));
			_msg.setSize(new Dimension(300, 20));
			_msg.setFont(_msg.getFont().deriveFont(Font.BOLD, 9f));
			_msg.setForeground(Color.black);

			setLayout(new BorderLayout());
			setBorder(new EtchedBorder(EtchedBorder.LOWERED));
			add(_msg, BorderLayout.CENTER);
			add(_progress, BorderLayout.EAST);
		}

		/**
		 * Displays a message in the Progress Bar.
		 * 
		 * @param message a short string.
		 */
		public void showStatus( String message ) {
			_msg.setText(message);
			paintImmediately(getBounds());
		}

		/**
		 * Fills the progress bar from the specified percentage value.
		 * 
		 * @param percent the percentage value.
		 */
		public void showProgress( int percent ) {
			_progress.setValue(percent);
		}
	}
}