/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.ui;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.graph.item.CompositeItem;
import fr.greyc.pantheon.ariane.graph.slot.InSlotParameter;
import fr.greyc.pantheon.ariane.utils.GDialog;
import fr.greyc.pantheon.ariane.utils.MutableTreeItem;
import fr.greyc.pantheon.ariane.utils.SubroutineTreeItem;
import fr.greyc.pantheon.ariane.utils.TreeItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;
import fr.greyc.pantheon.ariane.xml.XMLFileOutputStream;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.tree.DefaultTreeModel;

/**
 * The class <code>SubroutineEditor</code> defines a dialog box
 * where the user can add new subroutine in the explorer.
 * 
 * @version Régis CLOUARD - Jul 20, 2010
 */
public final class SubroutineEditor extends GDialog {
	/** The operator to describe. */
	private SubroutineTreeItem _currentItem;
	/** the node that support the operator. */
	private MutableTreeItem _currentNode;
	/** The tree model. */
	private DefaultTreeModel _treeModel;

	/** The category of the new routine. */
	private JComboBox _categoryField=null;

	/** The name of the new routine. */
	private JTextField _nameField=null;

	/** The name of the new routine. */
	private JTextField _signatureField=null;

	/** The documentation of the new routine. */
	private JTextField _docURLField=null;

	/** True if the subroutine should be directly inserted. */
	private boolean _inserted=false;

	private boolean _validated;
	
	/**
	 * Creates a dialog box for setting new paths.
	 * 
	 * @param item the item to be stored.
	 */
	public SubroutineEditor( CompositeItem item ) {
		super((JDialog)null, Ariane.lang.getString("SubroutineEditor.title"), true);
		addWindowListener(new WindowAdapter() {
			@Override
			public final void windowClosing(WindowEvent e) {
				cancelAction();
			}
		});

		try {
			XMLElement e;
			e = new XMLElement("Subroutine");
			e.addChildren(new XMLElement("Category", ""));
			e.addChildren(new XMLElement("Name", item.getNickName()));
			e.addChildren(new XMLElement("Documentation" ,""));
			// Build the default parameter values list
			String list = "";
			for (int i=0; i< item.getInSlotParameters().size(); i++ ) {
				InSlotParameter parameter = item.getInSlotParameters().get(i);
				String name=null;
				if (parameter.hasName()) {
					name=parameter.getName();
				} else {
					name="";
				}
				String defaultValue=null;
				if (!parameter.isConnected() && parameter.hasValue(false)) {
					defaultValue="="+parameter.getValue();
				} else {
					defaultValue="";
				}
				if (name.length()>0 || defaultValue.length()>0) {
					list+="$"+i+":"+name+defaultValue+";";
				}
			}
			e.addChildren(new XMLElement("Signature" ,list));
			e.addChildren(item.save(null));
			_currentItem = new SubroutineTreeItem(e.getChildren());
			_inserted=true;
		} catch( Exception e) {
			e.printStackTrace();
		}
		displayContent();
		setLocationRelativeTo(null);
	}

	/**
	 * Creates and shows the dialog box.
	 * 
	 * @param parent the parent for this dialog box.
	 * @param currentNode the node which supports the operator.
	 * @param treeModel the tree model.
	 */
	public SubroutineEditor( JDialog parent, MutableTreeItem currentNode, DefaultTreeModel treeModel ) {
		super(parent, Ariane.lang.getString("SubroutineEditor.title"), true);
		_currentNode=currentNode;
		_currentItem = (SubroutineTreeItem)_currentNode.getUserObject();
		_treeModel=treeModel;
		_inserted=false;
		displayContent();
		setLocationRelativeTo(parent);
	}

	/**
	 * Displays the content of the dialog box.
	 */
	private final void displayContent( ) {
		setResizable(false);
		JPanel textPanel = new JPanel();
		textPanel.setLayout(new GridBagLayout());
		textPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Ariane.lang.getString("SubroutineEditor.borderTitle")));
		GridBagConstraints c = new GridBagConstraints();

		c.fill = GridBagConstraints.BOTH;
		textPanel.add(new JLabel(Ariane.lang.getString("SubroutineEditor.categoryText")),c);
		Vector<String> categories=Ariane.workspace.getTree().getSubroutineCategories();
		_categoryField = new JComboBox(categories);
		_categoryField.setEditable(true);
		_categoryField.setSelectedItem(_currentItem.getCategory());
		c.gridwidth = GridBagConstraints.REMAINDER; //end row
		textPanel.add(_categoryField,c);

		c.gridwidth =1;
		textPanel.add(new JLabel(Ariane.lang.getString("SubroutineEditor.nameText")),c);
		_nameField= new JTextField(30);
		_nameField.setText(_currentItem.getName());
		c.gridwidth = GridBagConstraints.REMAINDER; //end row
		textPanel.add(_nameField,c);

//		// parameters
//		c.gridwidth =1;
//		textPanel.add(new JLabel(Ariane.lang.getString("SubroutineEditor.parameters")),c);
//		JTextField t1 = new JTextField(""+_currentItem.getParameters());
//		t1.setEditable(false);
//		c.gridwidth = GridBagConstraints.REMAINDER; //end row
//		textPanel.add(t1,c);

		// Signature
		c.gridwidth =1;
		textPanel.add(new JLabel(Ariane.lang.getString("SubroutineEditor.signature")),c);
		_signatureField= new JTextField(_currentItem.getSignature());
		c.gridwidth = GridBagConstraints.REMAINDER; //end row
		textPanel.add(_signatureField,c);
		textPanel.add(new JLabel(Ariane.lang.getString("SubroutineEditor.signatureExample")),c);

//		// inputs
//		c.gridwidth =1;
//		textPanel.add(new JLabel(Ariane.lang.getString("SubroutineEditor.inputs")),c);
//		JTextField t2 = new JTextField(""+_currentItem.getInImages());
//		t2.setEditable(false);
//		c.gridwidth = GridBagConstraints.REMAINDER; //end row
//		textPanel.add(t2,c);

//		//outputs
//		c.gridwidth =1;
//		textPanel.add(new JLabel(Ariane.lang.getString("SubroutineEditor.outputs")),c);
//		JTextField t3 = new JTextField(""+_currentItem.getOutImages());
////		System.out.println("objet="+_currentItem);
//		t3.setEditable(false);
//		c.gridwidth = GridBagConstraints.REMAINDER; //end row
//		textPanel.add(t3,c);

		c.gridwidth =1;
		textPanel.add(new JLabel(Ariane.lang.getString("SubroutineEditor.documentationText")),c);
		_docURLField = new JTextField(30);
		_docURLField.setText(_currentItem.getDocURL());
		textPanel.add(_docURLField,c);
		JButton button2 = new JButton(Ariane.lang.getString("SubroutineEditor.load"));
		button2.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) {
				JFileChooser fc = new JFileChooser(".");
				fc.setDialogType(JFileChooser.OPEN_DIALOG);
				fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
				if (_docURLField.getText().length()>6) {
					fc.setSelectedFile(new File(_docURLField.getText().substring(7)));
				}
				int returnVal = fc.showOpenDialog(SubroutineEditor.this);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					_docURLField.setText("file://"+fc.getSelectedFile().getAbsolutePath());
					_docURLField.setCaretPosition(0);
				}
			}
		});
		JButton button3 = new JButton(Ariane.lang.getString("OperatorEditor.edit"));
		button3.addActionListener( new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				HTMLDocumentEditor editor = new HTMLDocumentEditor(SubroutineEditor.this);
				if (_docURLField.getText().length()>6) {
					editor.setSelectedFile(new File(_docURLField.getText().substring(7)));
				}
				if (editor.edit() == HTMLDocumentEditor.OK_OPTION) {
					_docURLField.setText("file://"+editor.getSelectedFile().getAbsolutePath());
					_docURLField.setCaretPosition(0);
				}
			}
		});
		textPanel.add(button2, c);
		c.gridwidth = GridBagConstraints.REMAINDER; //end row
		textPanel.add(button3, c);

		// cancel button
		JButton cancelButton = new JButton();
		cancelButton.setAction(new AbstractAction(Ariane.lang.getString("SubroutineEditor.cancel"),
				new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"gtk-cancel.png"))) {
			@Override
			public final void actionPerformed(ActionEvent e) {
				cancelAction();
			}
		});

		// Apply button
		JButton applyButton = new JButton(Ariane.lang.getString("SubroutineEditor.apply"),
				new ImageIcon(getClass().getClassLoader().getResource(Ariane.iconDir+"gtk-apply.png")));

		applyButton.addActionListener( new ActionListener() {
			@Override
			public final void actionPerformed(ActionEvent e) {
				applyAction();
			}
		});

		JPanel buttonPanel = new JPanel();
		buttonPanel.add(applyButton);
		buttonPanel.add(cancelButton);

		setDefaultButton(applyButton);
		setCancelButton(cancelButton);

		getContentPane().setLayout(new BorderLayout());
		getContentPane().add(textPanel, BorderLayout.NORTH);
		getContentPane().add(buttonPanel, BorderLayout.SOUTH);
		pack();
	}

	/**
	 * Validates information.
	 */
	private void cancelAction( ) {
		setVisible(false);
		_validated=false;
		dispose();
	}

	/**
	 * Applies and saves the chosen preferences,
	 * and eventually closes the dialog box.
	 */
	private final void applyAction( ) {
		String category= ""+_categoryField.getSelectedItem();
		if (category.trim().length()==0) {
			Ariane.workspace.printErrorMessage(Ariane.lang.getString("SubroutineEditor.error1"));
			return;
		}
		_currentItem.setCategory(category);
		_currentItem.setName(_nameField.getText());
		_currentItem.setSignature(_signatureField.getText());
		_currentItem.setDocURL(_docURLField.getText());
		if (_inserted) {
			try {
				Ariane.workspace.getTree().addSubroutine(_currentItem);
				save();
				_validated=true;
			} catch( Exception e) { }
		} else {
			_treeModel.nodeChanged(_currentNode);
			_validated=false;
		}
		setVisible(false);
		dispose();
	}

	/**
	 * Saves the current subroutine list in the convenient file.
	 * 
	 * @return true if all of the subroutines has been saved.
	 */
	private final boolean save() {
		File f = new File(Ariane.getSubroutineFilename());
		File d = f.getParentFile();
		try {
			d.mkdirs();
			XMLFileOutputStream out = new XMLFileOutputStream(f,null);
			out.write(saveSubroutine());
			out.close();
		} catch (Exception ex) {
			Ariane.workspace.printSystemMessage(Ariane.lang.getString("PreferenceChooser.message.saved")+" "+Ariane.getPreferencesFilename());
			return false;
		}
		return true;
	}

	/**
	 * Saves subroutines in an XML tree.
	 * 
	 * @return the XML tree.
	 * @throws IOException in case of IO problems.
	 */
	private static final XMLElement saveSubroutine() throws IOException {
		XMLElement e = null;
		LinkedList<TreeItem> subRoutineList=Ariane.workspace.getTree().getSubroutineList();
		if (subRoutineList.size() > 0) {
			e = new XMLElement("SubroutineLibrary");
			Iterator<TreeItem> it=subRoutineList.iterator();
			while (it.hasNext())
				e.addChildren(it.next().save());
		}
		return e;
	}
	
	public boolean isValidated() {
		return _validated;
	}
}