/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.utils;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.util.LinkedList;

import fr.greyc.pantheon.ariane.graph.Graph;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureForItem;
import fr.greyc.pantheon.ariane.graph.item.ControlStructureWhileItem;
import fr.greyc.pantheon.ariane.graph.item.GraphItem;
import fr.greyc.pantheon.ariane.graph.item.MacroOperatorItem;
import fr.greyc.pantheon.ariane.xml.XMLElement;

/**
 * A <code>SubroutineTreeItem</code> is an item that can be inserted in
 * the tree of the explorer.
 *
 * @version Régis CLOUARD - Jul 17, 2010
 */
public final class SubroutineTreeItem extends TreeItem implements Transferable {
	/** The list of items ion the subgraph. */
	private LinkedList<XMLElement> _body=null;

	/** The magic number for transferable item. */
	public static DataFlavor[] flavors = {new DataFlavor("application/x-java-serialized-subroutine-item-tree", "SubroutineTreeItem Flavor")};

	/**
	 * Creates a new TreeItem from values in the given XML tree.
	 * 
	 * @param xmlTree the XML tree of properties.
	 * @throws Exception any Exception.
	 */
	public SubroutineTreeItem( LinkedList<XMLElement> xmlTree ) throws Exception {
		super( null, null, null, null, null, 0, 0, 0,"" );
		load(xmlTree);
	}

	/**
	 * Gets the transfered object.
	 */
	@Override
	public final DataFlavor[] getTransferDataFlavors( ) {
		return flavors;
	}

	/**
	 * Gets the transfered data.
	 * 
	 * @param flavor the flavor of the transfered object.
	 * @return the data.
	 */
	@Override
	public final Object getTransferData( DataFlavor flavor ) throws UnsupportedFlavorException, IOException {
		if (isDataFlavorSupported(flavor)) return this;
		throw new UnsupportedFlavorException(flavor);
	}

	/**
	 * Is the object the transferable object?
	 * 
	 * @param flavor the flavor to be tested.
	 * @return true if the flavor is transferable.
	 */
	@Override
	public final boolean isDataFlavorSupported( DataFlavor flavor ) {
		return flavor.equals(flavors[0]);
	}

	/**
	 * Instantiates an item in the given graph.
	 * 
	 * @param graph the graph where to add the item.
	 * @return the new graph item.
	 */
	@Override
	public final GraphItem createGraphItem( Graph graph ) {
		try {
			GraphItem g=null;
			switch(getType()) {
			case MACROOPERATOR: g = new MacroOperatorItem(graph,_body, this); break;
			case CONTROLFOR: g= new ControlStructureForItem(graph,_body, this); break;
			case CONTROLWHILE: g = new ControlStructureWhileItem(graph,_body, this); break;
			default: return null;
			}
			g.updateName();
			return g;
		} catch ( IOException e) {
		}
		return null;
	}

	/**
	 * Loads properties of a tree item from an XML tree.
	 * 
	 * @param xmlelements the XML Tree of properties.
	 * @throws Exception any Exception.
	 */
	@Override
	protected final void load( LinkedList<XMLElement> xmlelements ) throws Exception {
		for (XMLElement item : xmlelements )
			if (item.getName().equalsIgnoreCase("Category")) {
				setCategory(item.getValue());
			} else if (item.getName().equalsIgnoreCase("Name")) {
				setName(item.getValue());
			} else if (item.getName().equalsIgnoreCase("Documentation")) {
				setDocURL(item.getValue());
			} else if (item.getName().equalsIgnoreCase("Signature")) {
				setSignature(item.getValue());
			} else if (item.getName().equalsIgnoreCase("MacroOperator")) {
				setType(Item.MACROOPERATOR);
				_body = item.getChildren();
			} else if (item.getName().equalsIgnoreCase("For")) {
				setType(Item.CONTROLFOR);
				_body = item.getChildren();
			} else if (item.getName().equalsIgnoreCase("While")) {
				setType(Item.CONTROLWHILE);
				_body = item.getChildren();
			}
	}

	/**
	 * Saves the subroutine tree item as an XML Element.
	 * 
	 * @return the resulted XML tree.
	 * @throws IOException any Exception.
	 */
	@Override
	public final XMLElement save( ) throws IOException {
		XMLElement e = null;
		e = new XMLElement("Subroutine");
		e.addChildren(new XMLElement("Category",getCategory()));
		e.addChildren(new XMLElement("Name" ,getName()));
		e.addChildren(new XMLElement("Documentation",getDocURL()));
		e.addChildren(new XMLElement("Signature",getSignature()));
		XMLElement i = null;
		switch(getType()) {
		case MACROOPERATOR: i=new XMLElement("MacroOperator"); break;
		case CONTROLFOR: i= new XMLElement("For"); break;
		case CONTROLWHILE: i=new XMLElement("While"); break;
		}
		i.addChildren(_body);
		e.addChildren(i);
		return e;
	}

	public LinkedList<XMLElement> getBody() {
		return _body;
	}
}