/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c) 2013, GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 * 
 *   
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

package fr.greyc.pantheon.ariane.xml;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.util.LinkedList;

import org.xml.sax.Attributes;
import org.xml.sax.helpers.AttributesImpl;

/**
 * An <code>XMLElement</code> defines a pair tag-value
 * that can be read or write in an XML file.
 *
 * @author Régis CLOUARD - Nov 25, 2002
 * @version Régis CLOUARD - Nov 30, 2004
 */
public final class XMLElement implements Transferable, java.io.Serializable {
	/** The name of the tag. */
	private String _tag = null;

	/** The list of related attributes. */
	private AttributesImpl _atts = null;

	/** The value of the tag. */
	private StringBuffer _value = new StringBuffer(100);

	/** The list of all children tags. */
	public LinkedList<XMLElement> _children =  null;

	/**
	 * Creates a new XMLElement with the specified tag name.
	 * 
	 * @param tag the tag name.
	 */
	public XMLElement( String tag ) {
		_tag = tag;
	}

	/**
	 * Creates a new XMLElement with the specified attributes list.
	 * 
	 * @param tag the tag name.
	 * @param atts the list of related attributes.
	 */
	public XMLElement( String tag, Attributes atts ) {
		_tag = tag;
		_atts = new AttributesImpl(atts);
	}

	/**
	 * Creates a new XMLElement with the specified value.
	 * 
	 * @param tag the tag name.
	 * @param value the value.
	 */
	public XMLElement( String tag, String value ) {
		_tag = tag;
		setValue(value);
	}

	/**
	 * Creates a new XMLElement tag-attribute/value-value with the
	 * specified value and the specified attribute value.
	 * 
	 * @param tag the tag name.
	 * @param value the value.
	 * @param attName the name of the attached attribute.
	 * @param attValue the value of the attached attribute.
	 */
	public XMLElement( String tag, String value, String attName, String attValue ) {
		_tag = tag;
		_atts=new AttributesImpl();
		_atts.addAttribute("","",attName, "NMTOKEN", attValue);
		setValue(value);
	}

	/**
	 * @return the name of the tag.
	 */
	public final String getName( ) { return _tag; }

	/**
	 * Sets the new value to the tag.
	 * 
	 * @param value the value.
	 */
	public final void setValue( String value ) {
		_value.setLength(0);
		if (value != null) {
			_value = _value.append(value);
		}
	}

	/**
	 * Returns the value as a String.
	 * 
	 * @return a String or "".
	 */
	public final String getValue( ) { return _value.toString(); }

	/**
	 * Returns the value of a given attribute.
	 * 
	 * @param attName the name of an attribute.
	 * @return a String or "0".
	 */
	public final String getAttributeValue( String attName ) {
		if (_atts==null || _atts.getLength()==0) return "0";
		return _atts.getValue(attName);
	}

	/**
	 * Returns the value of a given attribute.
	 * 
	 * @return an AttributesImpl or null.
	 */
	public final AttributesImpl getAttributes( ) {
		return _atts;
	}

	/**
	 * @return the list of all children.
	 */
	public final LinkedList<XMLElement> getChildren( ) {
		return _children;
	}

	/**
	 * Adds a subtree as the child of the 
	 * specified XML element.
	 * 
	 * @param element the XML element.
	 */
	public final void addChildren( XMLElement element ) {
		if (element != null) {
			if (_children == null)
				_children = new LinkedList<XMLElement>();
			_children.add(element);
		}
	}

	/**
	 * Adds a subtree as the child of the specified XML tree.
	 * 
	 * @param xmlTree the subtree to be added.
	 */
	public final void addChildren( LinkedList<XMLElement> xmlTree ) {
		if (xmlTree == null) {
			return;
		}
		if (_children == null) {
			_children = xmlTree;
		} else { // Catenate 2 linked-lists.
			while (! xmlTree.isEmpty())
				_children.addLast(xmlTree.removeFirst());
		}
	}

	// Transferable
	public static final DataFlavor flavor = new DataFlavor(fr.greyc.pantheon.ariane.xml.XMLElement.class, "Ariane graph");

	private static final DataFlavor[] flavors = { flavor, };

	@Override
	public DataFlavor[] getTransferDataFlavors() {
		return flavors.clone();
	}

	@Override
	public boolean isDataFlavorSupported( DataFlavor flavor ) {
		return (flavor.equals(flavors[0]));
	}

	@Override
	public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException {
		if (flavor.equals(flavors[0])) {
			return this;
		}
		throw new UnsupportedFlavorException(flavor);
	}
}